package edu.washington.cs.rfid.mstreams.data;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import markovianStream.ITimestep;
import storageManager.schema.ITuple;
import edu.washington.cs.rfid.mstreams.exception.MStreamSyntaxException;

public class ParticleFilterStreamCPTData {

   private static final String TOKEN_DELIMETER    = ",";
   
   //--//
   
   public long                               tagId;
   public long                               startTimestamp;
   public long                               endTimestamp;
   public Vector<NodeTransition>             nodeTransitions;
   public HashMap<Long,HashMap<Long,Double>> transitionMap;
   
   //--//
   
   /**
    * Parses line to create a new Marginal.
    * @param line
    */
   public ParticleFilterStreamCPTData(String line) throws MStreamSyntaxException {
      
      this.nodeTransitions = new Vector<NodeTransition>();
      this.transitionMap   = new HashMap<Long,HashMap<Long,Double>>();
      
      try {
         String[] tokens = line.split(ParticleFilterStreamCPTData.TOKEN_DELIMETER);
         
         // Parse the tag and times
         if(tokens.length > 2){
            this.tagId          = Long.parseLong(tokens[0]);
            this.startTimestamp = Long.parseLong(tokens[1]);
            this.endTimestamp   = Long.parseLong(tokens[2]);
         }
         
         for(int i = 3; i < tokens.length; i++){
        	NodeTransition transition = new NodeTransition(tokens[i]);
            this.nodeTransitions.add(transition);
            
            if(null == this.transitionMap.get(transition.node1)){
               HashMap<Long,Double> map = new HashMap<Long,Double>();
               map.put(transition.node2,transition.prob);
               this.transitionMap.put(transition.node1, map);
            } else {
               this.transitionMap.get(transition.node1).put(transition.node2, transition.prob);
            }
         }
      } catch(Exception e){
         throw new MStreamSyntaxException("Error parsing marginal: " + line);
      }
   }
   
   
   public ParticleFilterStreamCPTData(long tagId         , 
		          long startTimestamp, 
		          long endTimestamp  , 
		          Map<ITuple,Map<ITuple,Double>> map){
	   
	   this.tagId           = tagId;
	   this.startTimestamp  = startTimestamp;
	   this.endTimestamp    = endTimestamp;
	   this.nodeTransitions = new Vector<NodeTransition>();
	   this.transitionMap   = new HashMap<Long,HashMap<Long,Double>>();
	   
	   for(ITuple node1Tuple : map.keySet()){
		   Map<ITuple,Double> subMap = map.get(node1Tuple);
		   
		   for(ITuple node2Tuple : subMap.keySet()){
			   long           node1 = ((Long)node1Tuple.getAttributeValue(1)).longValue();
			   long           node2 = ((Long)node2Tuple.getAttributeValue(1)).longValue();
			   double         prob  = subMap.get(node2Tuple);
			   NodeTransition trans = new NodeTransition(node1,node2,prob);
			   
			   this.nodeTransitions.add(trans);
			   
			   if(null == this.transitionMap.get(node1)){
				   this.transitionMap.put(node1, new HashMap<Long,Double>());
			   }
			   this.transitionMap.get(node1).put(node2, prob);
		   }
	   }
   }
   
   /**
    * Creates the object from the timestep (whcih is assumed to conform to the OfficeRFIDStreamSchema)
    * @param timestep
    */
   public ParticleFilterStreamCPTData(ITimestep ts, long tagId) { 

	   this.tagId           = tagId;
	   this.startTimestamp  = (Long)ts.getExtraData();
	   this.endTimestamp    = startTimestamp+1000;
	   this.nodeTransitions = new Vector<NodeTransition>();
	   this.transitionMap   = new HashMap<Long,HashMap<Long,Double>>();

	   for(ITuple node1Tuple : ts.getCPT().getValues().keySet()) {
		   Map<ITuple,Double> subMap = ts.getCPT().getValues().get(node1Tuple);

		   for(ITuple node2Tuple : subMap.keySet()){
			   long           node1 = ((Integer)node1Tuple.getAttributeValue(0)).longValue();
			   long           node2 = ((Integer)node2Tuple.getAttributeValue(0)).longValue();
			   double         prob  = subMap.get(node2Tuple);
			   NodeTransition trans = new NodeTransition(node1,node2,prob);

			   this.nodeTransitions.add(trans);

			   if(null == this.transitionMap.get(node1)){
				   this.transitionMap.put(node1, new HashMap<Long,Double>());
			   }
			   this.transitionMap.get(node1).put(node2, prob);
		   }
	   }
}
   
   
   public Vector<Double> getTotalProbs(){
	   
	   Vector<Double> totalProbs = new Vector<Double>();
	   
	   for(Long node1 : this.transitionMap.keySet()){
		   HashMap<Long,Double> subMap = this.transitionMap.get(node1);
		   double               totalP = 0.0;
		   for(Double prob : subMap.values()){
			   totalP += prob.doubleValue();
		   }
		   totalProbs.add(totalP);
	   }
	   return totalProbs;
   }
   
   
   public boolean equals(ParticleFilterStreamCPTData c){
	   boolean result = false;
	   
	   if(this.tagId == c.tagId && this.startTimestamp == c.startTimestamp && this.endTimestamp == c.endTimestamp){
		   
		   // Compare list of transitions
		   if(this.nodeTransitions.size() == c.nodeTransitions.size()){
			   Collections.sort(this.nodeTransitions);
			   Collections.sort(c.nodeTransitions);
			   
			   result = true;
			   
			   for(int i = 0; i < this.nodeTransitions.size(); i++){
				   if(!this.nodeTransitions.get(i).equals(c.nodeTransitions.get(i))){
					   result = false;
					   break;
				   }
			   }
		   } else {
			   System.out.println("CPT: Node transitions are not the same size!");
			   System.out.println("file1 transitions " + this.nodeTransitions.size() + ":");
			   for(NodeTransition t : this.nodeTransitions){
				   System.out.print(t.toString() + " ");
			   }
			   System.out.println("\nfile2 transitions " + c.nodeTransitions.size() + ":");
			   for(NodeTransition t : c.nodeTransitions){
				   System.out.print(t.toString() + " ");
			   }
		   }
		   
		   // Now compare maps
		   Set<Long> myKeys    = this.transitionMap.keySet();
		   Set<Long> otherKeys = c.transitionMap.keySet();
		   
		   if(myKeys.size() == otherKeys.size()){
			   System.out.println("CPT: map keys sets same length!");
			   
			   if(myKeys.equals(otherKeys)){
				   System.out.println("CPT: map keys sets are equal!");
				   
				   // Now compare map contents
				   for(Long key : myKeys){
					   if(!this.transitionMap.get(key).keySet().equals(c.transitionMap.get(key).keySet())){
						   result = false;
						   System.out.println("CPT: map values not equal for key: " + key);
//						   System.out.println("CPT: 1) " + key + " --> " +  );
						   break;
					   }
				   }
			   } else {
				   System.out.println("CPT: map key sets not equal!");
				   result = false;
			   }
		   } else {
			   System.out.println("CPT: maps key set sizes not equal!");
			   result = false;
		   }
	   }
	   
	   return result;
   }
   
   
   public String toString(){
      String str = this.tagId          + ParticleFilterStreamCPTData.TOKEN_DELIMETER + 
                   this.startTimestamp + ParticleFilterStreamCPTData.TOKEN_DELIMETER + 
                   this.endTimestamp   + ParticleFilterStreamCPTData.TOKEN_DELIMETER;
      
      for(NodeTransition transition : this.nodeTransitions){
         str += transition.toString() + ParticleFilterStreamCPTData.TOKEN_DELIMETER;
      }
      str = str.substring(0, str.length()-2);
      return str;
   }
}
