package edu.washington.cs.rfid.mstreams.data;

import java.util.Collections;
import java.util.Map;
import java.util.Vector;

import markovianStream.ITimestep;

import storageManager.schema.ITuple;
import edu.washington.cs.rfid.mstreams.exception.MStreamSyntaxException;

public class ParticleFilterStreamMarginalData {

   private static final String TOKEN_DELIMETER    = ",";
   private static final String TIME_TAG_DELIMETER = "_";
   
   //--//
   
   public long                 timestamp;
   public long                 tagId;
   public Vector<NodeProbPair> nodeProbPairs;
   
   //--//
   
   /**
    * Parses line to create a new Marginal.
    * @param line
    */
   public ParticleFilterStreamMarginalData(String line) throws MStreamSyntaxException {
      
      this.nodeProbPairs = new Vector<NodeProbPair>();
      
      try {
         String[] tokens = line.split(ParticleFilterStreamMarginalData.TOKEN_DELIMETER);
         
         // Parse the time and tag
         if(tokens.length > 0){
            String[] timeTag = tokens[0].split(ParticleFilterStreamMarginalData.TIME_TAG_DELIMETER);
            if(timeTag.length == 2){
               this.timestamp = Long.parseLong(timeTag[0]);
               this.tagId     = Long.parseLong(timeTag[1]);
            }
         }
         if(tokens.length % 2 == 1){
            for(int i = 1; i < tokens.length; i+=2){
               long   node = Long  .parseLong  (tokens[i]  );
               double prob = Double.parseDouble(tokens[i+1]);
               
               this.nodeProbPairs.add(new NodeProbPair(node, prob));
            }
         }
      } catch(Exception e){
         throw new MStreamSyntaxException("Error parsing marginal: " + line);
      }
   }
   
   
   public ParticleFilterStreamMarginalData(long timestamp, long tagId, Map<ITuple,Double> map) {
	   this.timestamp     = timestamp;
	   this.tagId         = tagId;
	   this.nodeProbPairs = new Vector<NodeProbPair>();
	   
	   for(ITuple tuple : map.keySet()){
		   long   node = ((Long)tuple.getAttributeValue(1)).longValue();
		   double prob = map.get(tuple);
		   
		   this.nodeProbPairs.add(new NodeProbPair(node,prob));
	   }
   }
   
   /**
    * Creates the object from the timestep (whcih is assumed to conform to the OfficeRFIDStreamSchema)
    * @param timestep
    */
   public ParticleFilterStreamMarginalData(ITimestep timestep, long tagID) {
	   this.timestamp = (Long)timestep.getExtraData();
	   this.tagId = tagID;
	   this.nodeProbPairs = new Vector<NodeProbPair>();
	   
	   for(ITuple tuple : timestep.getMarginal().keySet()) {
		   int node = ((Integer)tuple.getAttributeValue(0)).intValue();
		   double prob = timestep.getMarginal().get(tuple);
		   this.nodeProbPairs.add(new NodeProbPair(node, prob));
	   }
   }
   
   
   public double getTotalProb(){
	   double totalP = 0.0;
	   for(NodeProbPair pair : this.nodeProbPairs){
		   totalP += pair.prob;
	   }
	   return totalP;
   }
   
   
   public boolean equals(ParticleFilterStreamMarginalData m){
	   boolean result = false;
	   
	   if(this.timestamp == m.timestamp & this.tagId == m.tagId){
		   if(this.nodeProbPairs.size() == m.nodeProbPairs.size()){
			   Collections.sort(this.nodeProbPairs);
			   Collections.sort(m.nodeProbPairs);
			   
			   result = true;
			   
			   for(int i = 0; i < this.nodeProbPairs.size(); i++){
				   NodeProbPair p1 = this.nodeProbPairs.get(i);
				   NodeProbPair p2 = this.nodeProbPairs.get(i);
				   
				   if(p1.node != p2.node || p1.prob != p2.prob){
					   result = false;
					   break;
				   }
			   }
		   }
	   }
	   
	   return result;
   }
   
   
   public String toString(){
      String str = this.timestamp + ParticleFilterStreamMarginalData.TIME_TAG_DELIMETER + this.tagId + ParticleFilterStreamMarginalData.TOKEN_DELIMETER;
      
      for(int i = 0; i < this.nodeProbPairs.size() - 1; i++){
         NodeProbPair pair  = this.nodeProbPairs.get(i);
         str               += pair.toString() + ParticleFilterStreamMarginalData.TOKEN_DELIMETER;
      }
      str += this.nodeProbPairs.get(this.nodeProbPairs.size() - 1).toString();
      return str;
   }
   
   
   //--//
   
   
   public class NodeProbPair implements Comparable<NodeProbPair>{
      
      public long   node;
      public double prob;
      
      public NodeProbPair(long node, double prob){
         this.node = node;
         this.prob = prob;
      }
      
      public int compareTo(NodeProbPair p){
    	  int result = 0;
    	  
    	  if(this.node > p.node){
    		  result = 1;
    	  } else if(this.node < p.node){
    		  result = -1;
    	  } else {
    		  if(this.prob > p.prob){
    			  result = 1;
    		  } else if(this.prob < p.prob){
    			  result = -1;
    		  } else {
    			  result = 0;
    		  }
    		  
    	  }
    	  
    	  return result;
      }
      
      public String toString(){
         return this.node + ParticleFilterStreamMarginalData.TOKEN_DELIMETER + this.prob;
      }
   }
   
}
