package edu.washington.cs.rfid.mstreams.data.db;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import markovianStream.ITimestep;
import storageManager.schema.IStreamSchema;
import storageManager.schema.IStreamSchema.ImprecisionType;
import accessPaths.IStorageLayer;
import accessPaths.IStreamCursor;
import accessPaths.IStreamWriter;
import accessPaths.bdb.BDBStorageLayer;
import edu.washington.cs.rfid.mstreams.schema.ParticleFilterStreamSchema;

/**
 * Class to assist in converting result streams into BDB format files.
 * @author evan
 */
public class BdbConnection {
   
   protected static String DEFAULT_BDB_ROOT_DIR = "/data/bdbRfidStreams";

   //--//
   
   protected String bdbRootDir;
   
   //--//
   
   /**
    * Creates a new BdbLoader for the database in the default directory;
    */
   public BdbConnection(){
      this.bdbRootDir = BdbConnection.DEFAULT_BDB_ROOT_DIR;
   }
   
   
   /**
    * Creates a new BdbLoader for the database in the specified directory.
    * @param bdbRootDir
    */
   public BdbConnection(String bdbRootDir){
      this.bdbRootDir = bdbRootDir;
   }
   
   
   /**
    * Stores the given MStream into a specified BDB file with the specified 
    * stream key and schema.
    * 
    * @param stream
    * @param bdbRootDir
    * @param streamKey
    * @param schema
    * @param overwriteIfExists
    * @return
    */
   public boolean storeMStream(List<ITimestep> stream    ,
                               String          streamKey ,
                               IStreamSchema   schema    ,
                               boolean         overwriteIfExists){
      boolean          success        = false;
      File             mStreamRootDir = null;
      File             bdbDir         = null;
      IStorageLayer    bdbLayer       = null;
      IStreamWriter    writer         = null;

      try {
         mStreamRootDir = new File(this.bdbRootDir);
         bdbDir         = new File(mStreamRootDir, streamKey);
         
         if(bdbDir.exists() && !overwriteIfExists){
            
            success = true;
            
         } else {
            
            bdbLayer = new BDBStorageLayer(bdbDir   , 
                                           streamKey, 
                                           schema   , 
                                           ImprecisionType.MARKOVIAN);
            bdbLayer.open(true);
   
            writer = bdbLayer.getStreamWriterHandle();
            writer.open();
            
            for(ITimestep ts : stream) {
            	writer.appendTimestep(ts, true);
            }
            
            success = true;
         }
      } finally {
         if(null != writer)   writer.close();
         if(null != bdbLayer) bdbLayer.close();
      }
      return success;
   }
   

   /**
    * Converts an MStream in BDB to an in-memory MStream.
    *  
    * @param bdbRootDir
    * @param streamKey
    * @return
    */
   public List<ITimestep> loadMStream(String streamKey) {
      
      List<ITimestep> stream     = new ArrayList<ITimestep>();
      File            bdbDirFile = new File(this.bdbRootDir, streamKey);
      IStorageLayer   bdbLayer   = new BDBStorageLayer(bdbDirFile                      ,
                                                       streamKey                       ,
                                                       new ParticleFilterStreamSchema(),
                                                       ImprecisionType.MARKOVIAN);
      bdbLayer.open(false);
      
      IStreamCursor<ITimestep> cursor = bdbLayer.getStreamCursorHandle();

      cursor.open();
      cursor.setToBeginning();

      while(cursor.hasNext()){
         ITimestep ts = cursor.getNext();
         stream.add(ts);
      }
      cursor.close();
      bdbLayer.close();
      
      return stream;
   }
}
