package edu.washington.cs.rfid.mstreams.data.db;

import java.io.File;
import java.util.List;

import markovianStream.ITimestep;
import optimizer.IQueryPlan.PostProcessingType;
import optimizer.IQueryPlan.RegType;
import parser.ParsedQuery;
import parser.UnsupportedQueryException;
import query.nfa.NFA;
import query.nfa.NFALink;
import query.nfa.NFALinks;
import query.nfa.NFAStateID;
import query.nfa.nfaPredicates.TriviallyTruePredicate;
import query.streamPredicates.GroupingPredicate_ByStreamKey;
import query.streamPredicates.StreamPredicate_EqualityCheck;
import storageManager.MarkovianStreamHandle;
import storageManager.schema.IStreamSchema;
import storageManager.schema.IStreamSchema.ImprecisionType;
import accessPaths.IStorageLayer;
import accessPaths.StorageLayerFactory;
import api.Lahar;
import api.queryResults.QueryResult;
import edu.washington.cs.rfid.mstreams.data.StreamKey;
import edu.washington.cs.rfid.mstreams.util.StreamConverter;

/**
 * Loads result streams into Lahar.
 * 
 * @author evan
 */
public class LaharConnection {

   protected static String DEFAULT_LAHAR_DB_DIR = "/scratch/data/db_loc/RfidDB";
   protected static String DEFAULT_BDB_DIR      = "/scratch/data/bdbRfidStreams";
   protected static String DEFAULT_TYPE         = "MARKOVIAN";

   //--//
   
   protected String bdbRootDir;
   protected String laharDbDir;
   protected Lahar  lahar;
   
   //--//
   
   /**
    * Creates a new ResultStreamLoader with the default Lahar location.
    */
   public LaharConnection(){
      this(LaharConnection.DEFAULT_BDB_DIR, LaharConnection.DEFAULT_LAHAR_DB_DIR);
   }
   
   
   /**
    * Creates a new ResultStreamLoader with the specified Lahar location.
    */
   public LaharConnection(String bdbRootDir, String laharDbDir){
      this.bdbRootDir = bdbRootDir;
      this.laharDbDir = laharDbDir;
      this.lahar      = Lahar.openExistingDatabase(new File(this.laharDbDir));
   }
   
   
   /**
    * Returns true if the stream with the specified key exists in the specified
    * schema.
    * 
    * @param sk
    * @param schema
    * @return
    */
   public boolean streamKeyExists(StreamKey sk, IStreamSchema schema){
      return this.lahar.streamKeyExists(sk.getKey(), schema.getSchemaName());
   }
   
   
   /**
    * Loads the stream with the given key from BDB into Lahar if it does not already
    * exist.
    * 
    * @param streamKey
    * @return
    */
   public void moveStream(StreamKey sk, IStreamSchema schema) throws IllegalStateException {
      this.moveStream(sk.getKey(), schema);
   }
   
   
   /**
    * Loads the stream with the given key from BDB into Lahar if it does not already
    * exist.
    * 
    * @param streamKey
    * @return
    */
   public void moveStream(String streamKey, IStreamSchema schema) throws IllegalStateException {
    
      if(!this.lahar.streamKeyExists(streamKey, schema.getSchemaName())){
         File            sourcePath         = new File(this.bdbRootDir + "/" + streamKey);
         ImprecisionType imprecisionType    = ImprecisionType.fromString(LaharConnection.DEFAULT_TYPE);
         boolean         loadApproximations = true;
         boolean		 createIndexes      = true;
         IStorageLayer   storageLayer       = StorageLayerFactory.createStorageLayer(this.lahar.getConfiguration().getStorageType(), 
                                                                                     sourcePath                                    , 
                                                                                     streamKey                                     , 
                                                                                     schema                                        , 
                                                                                     imprecisionType);
         storageLayer.open(true);
   
         MarkovianStreamHandle streamHandle = new MarkovianStreamHandle(schema, storageLayer, streamKey);
   
         lahar.loadStreamIntoLahar(streamHandle, loadApproximations, createIndexes);
         storageLayer.close();
   
   
         // TODO: Is this necessary??
         // Verify that all expected copies of the stream exist (one for each approximation)
//         String loadingStreamKey   = sourcePath.getName();
//       
//         File schemaDir = new File(new File(this.laharDbDir, "Streams"), schema.getSchemaName());
//         for(ImprecisionType it : ImprecisionType.values()) {
//            if ((!loadApproximations) && (it != ImprecisionType.MARKOVIAN)){
//               continue;
//            }
//            String dataDirName   = "Data_"+it.toString();
//            File   dataDir       = new File(schemaDir, dataDirName);
//            File   bdbLoadingDir = new File(dataDir, loadingStreamKey);
//   
//            if (!bdbLoadingDir.exists()){
//               throw new IllegalStateException("Error: bdb data directory does not exist; by now it should.");
//            }
//         }
      }
   }
   
   
   
   /**
    * Loads the stream with the specified stream and schema from Lahar.
    * @param sk
    * @param schema
    * @return
    */
   public List<ITimestep> loadMStream(StreamKey sk, IStreamSchema schema) {
     
	   QueryResult result = this.loadCachedResult(sk, schema);
      return StreamConverter.convertToMStream(result);
   }
   
   
   
   /**
    * Loads the result with the specified stream and schema from Lahar.
    * @param sk
    * @param schema
    * @return
    */
   public QueryResult loadCachedResult(StreamKey sk, IStreamSchema schema) {
      return this.loadTrivialResult(sk, schema);
   }

   
   /**
    * Executes a specified query without using indexing.
    * 
    * @param q
    * @param useIndexing
    * @return
    */
   public QueryResult executeQuery(ParsedQuery q) throws UnsupportedQueryException{
      return this.lahar.executeQuery(q, false, false, ImprecisionType.MARKOVIAN);   
   }
   
   
   /**
    * Executes a specified query, using indexing if specified.
    * 
    * @param q
    * @param useIndexing
    * @return
    */
   public QueryResult executeQuery(ParsedQuery q, boolean useIndexing) throws UnsupportedQueryException {
      return this.lahar.executeQuery(q, useIndexing, false, ImprecisionType.MARKOVIAN);   
   }
   
   
   //--//
   
   
   /**
    * Loads a specified stream from Lahar using a trivially true query.
    * 
    * @param sk
    * @param schema
    * @return
    */
   protected QueryResult loadTrivialResult(StreamKey sk, IStreamSchema schema) {
      
      NFALinks               links    = new NFALinks();
      TriviallyTruePredicate truePred = new TriviallyTruePredicate();      
      NFALink                trueLink = new NFALink(truePred, null, new NFAStateID(0), true);
   
      links.addLinkToEnd(trueLink);
      
      NFA         myQuery = new NFA(links);
      ParsedQuery query   = new ParsedQuery(schema.getSchemaName()                        ,
                                            new StreamPredicate_EqualityCheck(sk.getKey()),
                                            new GroupingPredicate_ByStreamKey()           ,
                                            myQuery                                       ,
                                            null,
                                            RegType.INSTANTS              ,
                                            PostProcessingType.STREAMEXISTS);
      
      QueryResult result = null;
      try {
    	  result = this.executeQuery(query, false);
      } catch (UnsupportedQueryException ex) {
    	  ex.printStackTrace();
      }
      return result;
   }
}