package edu.washington.cs.rfid.mstreams.data.db;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

import edu.washington.cs.rfid.data.db.RfidDataConnection;
import edu.washington.cs.rfid.data.db.SimpleDbConnection;
import edu.washington.cs.rfid.mstreams.data.StreamKey;

public class MStreamDataConnection extends RfidDataConnection {

   protected static final String STREAM_KEY    = "stream_key";
   protected static final String DESCRIPTION   = "description";
   
   //--//
   
   protected PreparedStatement selectStreamKeyByDescription;
   protected PreparedStatement selectStreamKeyByKey;
   protected PreparedStatement insertStreamKey;
   protected PreparedStatement deleteStreamKey;
   
   //--//
   
   /**
    * Creates a new <code>MStreamDataConnection</code> using the given <code>SimpleDBConnection</code>.
    * @param db
    * @throws IllegalArgumentException
    * @throws SQLException
    */
   public MStreamDataConnection(SimpleDbConnection db) throws IllegalArgumentException, SQLException {
      super(db);
      
      this.debug                        = false;
      this.selectStreamKeyByDescription = null;
   }
   
   
   /**
    * Creates a new <code>MStreamDataConnection</code> using the given <code>SimpleDBConnection</code> and debug switch.
    * 
    * @param db
    * @param debug
    * @throws IllegalArgumentException
    * @throws SQLException
    */
   public MStreamDataConnection(SimpleDbConnection db, boolean debug) throws IllegalArgumentException, SQLException {
      this(db);
      
      this.debug = debug;
   }
   
   
   /**
    * Retrieves a stream key by its description, if there is one.
    * @param description
    * @return
    * @throws SQLException
    */
   public StreamKey getStreamKey(String description) throws SQLException {
      StreamKey sk = null;
      
      if(this.db.isConnected()){
         if(null == this.selectStreamKeyByDescription){
            this.selectStreamKeyByDescription = this.db.prepareStatement("SELECT * FROM lahar_stream_key_map " +
            		                                                       "WHERE  description = ?");
         }
         
         this.selectStreamKeyByDescription.setString(1, description);
         
         ResultSet rs = this.selectStreamKeyByDescription.executeQuery();
         
         if(rs.next()){
            sk = new StreamKey(rs.getLong  (MStreamDataConnection.STREAM_KEY ),
                               rs.getString(MStreamDataConnection.DESCRIPTION));
         }
      }
      return sk;
   }
   
   
   /**
    * Retrieves a stream key by its key, if there is one.
    * @param key
    * @return
    * @throws SQLException
    */
   public StreamKey getStreamKey(long key) throws SQLException {
      StreamKey sk = null;
      
      if(this.db.isConnected()){
         if(null == this.selectStreamKeyByKey){
            this.selectStreamKeyByKey = this.db.prepareStatement("SELECT * FROM lahar_stream_key_map " +
                                                                 "WHERE  stream_key = ?");
         }
         
         this.selectStreamKeyByKey.setLong(1, key);
         
         ResultSet rs = this.selectStreamKeyByKey.executeQuery();
         
         if(rs.next()){
            sk = new StreamKey(rs.getLong  (MStreamDataConnection.STREAM_KEY ),
                               rs.getString(MStreamDataConnection.DESCRIPTION));
         }
      }
      return sk;
   }
   
   
   /**
    * Inserts the specified stream key if a stream key with the specified key value
    * does not already exist.
    * @param sk
    * @return
    * @throws SQLException
    * @throws IllegalArgumentException
    */
   public boolean addStreamKey(StreamKey sk) throws SQLException, IllegalArgumentException {
      
      boolean success = false;
      
      if(this.db.isConnected()){
         if(null == this.selectStreamKeyByKey){
            this.selectStreamKeyByKey = this.db.prepareStatement("SELECT * FROM lahar_stream_key_map " +
                                                                 "WHERE  stream_key = ?");
         }
         if(null == this.insertStreamKey){
            this.insertStreamKey = this.db.prepareStatement("INSERT INTO lahar_stream_key_map " +
                                                            "VALUES (?, ?)");
         }
         try {
            this.startTransaction();
            
            this.selectStreamKeyByKey.setLong(1, sk.key);
            
            ResultSet rs = this.selectStreamKeyByKey.executeQuery();
            
            if(!rs.next()){
               this.insertStreamKey.setLong  (1, sk.key        );
               this.insertStreamKey.setString(2, sk.description);
               
               success = (0 < this.insertStreamKey.executeUpdate());
            }
//            else {
//               throw new IllegalArgumentException("Stream with key " + sk.key + " already exists!");
//            }
         } finally {
            this.endTransaction();
         }
      }
      return success;
   }
   
   
   /**
    * Deletes a stream key from the stream key map.
    * @param sk
    * @return
    * @throws SQLException
    */
   public boolean removeStreamKey(StreamKey sk) throws SQLException {
      boolean success = false;
      
      if(this.db.isConnected()){
         if(null == this.deleteStreamKey){
            this.deleteStreamKey = this.db.prepareStatement("DELETE FROM lahar_stream_key_map " +
            		                                          "WHERE stream_key = ? AND description = ?");
         }
         this.deleteStreamKey.setLong  (1, sk.key        );
         this.deleteStreamKey.setString(2, sk.description);
         
         success = (0 < this.deleteStreamKey.executeUpdate());
      }
      return success;
   }
}