package edu.washington.cs.rfid.mstreams.data.db;

import java.io.IOException;
import java.sql.SQLException;

import edu.washington.cs.rfid.data.db.DbParameters;
import edu.washington.cs.rfid.mstreams.data.StreamKey;
import edu.washington.cs.rfid.mstreams.util.KeyHash;

/**
 * A helper class for managing stream keys.
 * @author magda
 *
 */
public class StreamKeyManager {
   
   protected static String DB_LOC                = "/scratch/rfid_data/db_loc/RfidDB";
   protected static long   INIT_HASH_VALUE       = 98102L;
   protected static String HASH_COLLISION_APPEND = "a";
   protected static int    MAX_COLLISIONS        = 10;

   //--//
   
   protected MStreamDataConnectionPool mstreamDbPool;
   protected MStreamFileConnection     mstreamFile;
   protected boolean                   useDb = false;


   public StreamKeyManager(){
      if(useDb){
         DbParameters dbParams = new DbParameters(DbParameters.TYPE_SQLSERVER,
                                                  "myserver" ,
                                                  1433                       ,
                                                  "mydb"          , 
                                                  "cascadia_user"              ,
                                                  "mypasswd".toCharArray());

         this.mstreamDbPool    = new MStreamDataConnectionPool(dbParams);
      } else {
         this.mstreamFile = new MStreamFileConnection();
      }
   }
   
   
   /**
    * Computes, checks and inserts a new stream key for the specified stream description.
    * @param description
    * @return
    */
   public long getStreamKey(String description) throws ClassNotFoundException, SQLException, IOException {
      long key = -1;
      
      if(this.useDb){
         key = this.getStreamKeyDb(description);
      } else {
         key = this.getStreamKeyFile(description);
      }
      
      return key;
   }
   
   
   /**
    * Clears the database of a stream key for the specified event description.
    * @param description
    * @return
    * @throws ClassNotFoundException
    * @throws SQLException
    */
   public int clearStreamKey(String description) throws ClassNotFoundException, SQLException, IOException {
      int numRemoved = 0;
      
      if(this.useDb){
         numRemoved = this.clearStreamKeyDb(description);
      } else {
         numRemoved = this.clearStreamKeyFile(description);
      }
      return numRemoved;
   }
   
   
   //--//
   
   
   private long getStreamKeyDb(String description) throws ClassNotFoundException, SQLException {
      String                curDesc   = new String(description);
      MStreamDataConnection mstreamDb = this.mstreamDbPool.getConnection();
      
      // Return the key if it already exists
      StreamKey sk = mstreamDb.getStreamKey(curDesc);
      
      if(null == sk){
         // Otherwise create and insert a new key
         long    keyTry = KeyHash.hash64(description, StreamKeyManager.INIT_HASH_VALUE);
         sk             = mstreamDb.getStreamKey(keyTry);
         boolean added  = false;
         int     tries  = 0;
         
         while(!added && tries < StreamKeyManager.MAX_COLLISIONS){         
            // Keep trying for a key until we find one that works (should almost never happen!) 
            while(sk != null && tries < StreamKeyManager.MAX_COLLISIONS){
               curDesc += StreamKeyManager.HASH_COLLISION_APPEND;
               keyTry   = KeyHash.hash64(description, StreamKeyManager.INIT_HASH_VALUE);
               sk       = mstreamDb.getStreamKey(keyTry);
               tries++;
            }
            if(tries < StreamKeyManager.MAX_COLLISIONS){
               sk = new StreamKey(keyTry,description);
               if(mstreamDb.addStreamKey(sk)){
                  added = true;
               } else {
                  sk = null;
               }
            }
         }
         if(StreamKeyManager.MAX_COLLISIONS <= tries){
            throw new IllegalArgumentException("No valid stream key found after " + StreamKeyManager.MAX_COLLISIONS + " tries!");
         }
      }
      return sk.key;
   }
   
   
   private long getStreamKeyFile(String description) throws IOException {
      String curDesc   = new String(description);
      
      // Return the key if it already exists
      StreamKey sk = this.mstreamFile.getStreamKey(curDesc);
      
      if(null == sk){
         // Otherwise create and insert a new key
         long    keyTry = KeyHash.hash64(description, StreamKeyManager.INIT_HASH_VALUE);
         sk             = this.mstreamFile.getStreamKey(keyTry);
         boolean added  = false;
         int     tries  = 0;
         
         while(!added && tries < StreamKeyManager.MAX_COLLISIONS){         
            // Keep trying for a key until we find one that works (should almost never happen!) 
            while(sk != null && tries < StreamKeyManager.MAX_COLLISIONS){
               curDesc += StreamKeyManager.HASH_COLLISION_APPEND;
               keyTry   = KeyHash.hash64(description, StreamKeyManager.INIT_HASH_VALUE);
               sk       = this.mstreamFile.getStreamKey(keyTry);
               tries++;
            }
            if(tries < StreamKeyManager.MAX_COLLISIONS){
               sk = new StreamKey(keyTry,description);
               if(this.mstreamFile.addStreamKey(sk)){
                  added = true;
               } else {
                  sk = null;
               }
            }
         }
         if(StreamKeyManager.MAX_COLLISIONS <= tries){
            throw new IllegalArgumentException("No valid stream key found after " + StreamKeyManager.MAX_COLLISIONS + " tries!");
         }
      }
      mstreamFile.writeTable();
      return sk.key;
   }
   
   
   private int clearStreamKeyDb(String description) throws ClassNotFoundException, SQLException {
      int                   numRemoved = 0;
      String                curDesc    = new String(description);
      MStreamDataConnection mstreamDb  = this.mstreamDbPool.getConnection();
      
      // Now try to find a match in the DB by description
      StreamKey sk = mstreamDb.getStreamKey(curDesc);
      
      if(null != sk){
         if(mstreamDb.removeStreamKey(sk)){
            numRemoved++;
         }
      }
      return numRemoved;
   }
   
   
   private int clearStreamKeyFile(String description) throws IOException {
      int                   numRemoved = 0;
      String                curDesc    = new String(description);
      
      // Now try to find a match in the DB by description
      StreamKey sk = mstreamFile.getStreamKey(curDesc);
      
      if(null != sk){
         if(mstreamFile.removeStreamKey(sk)){
            numRemoved++;
         }
      }
      mstreamFile.writeTable();
      return numRemoved;
   }
}
