package edu.washington.cs.rfid.mstreams.event;

import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.TreeSet;

import markovianStream.ITimestep;
import optimizer.IQueryPlan.PostProcessingType;
import optimizer.IQueryPlan.RegType;
import parser.ParsedQuery;
import parser.UnsupportedQueryException;
import query.nfa.NFA;
import query.nfa.NFALink;
import query.nfa.NFALinks;
import query.nfa.NFAStateID;
import query.nfa.nfaPredicates.IPredicate;
import query.nfa.nfaPredicates.OrPredicate;
import query.nfa.nfaPredicates.SingleAttributeEqualityPredicate;
import query.streamPredicates.GroupingPredicate_ByStreamKey;
import query.streamPredicates.StreamPredicate_EqualityCheck;
import storageManager.schema.Attribute;
import storageManager.schema.Attribute.AttributeType;
import api.queryResults.QueryResult;
import edu.washington.cs.rfid.mstreams.data.StreamKey;
import edu.washington.cs.rfid.mstreams.data.db.BdbConnection;
import edu.washington.cs.rfid.mstreams.data.db.LaharConnection;
import edu.washington.cs.rfid.mstreams.data.db.StreamKeyManager;
import edu.washington.cs.rfid.mstreams.util.StreamConverter;


/**
 * An inside event: true when a particular tag is at one or more specified nodes.
 * 
 * @author evan
 */
public class InsideEvent extends ScenicEvent {

   public AtEvent       atStream;
   public List<Integer> nodeList;
   
   //--//

   /**
    * Creates a new InsideEvent.
    * @param tag
    * @param nodeList
    */
   public InsideEvent(final AtEvent atStream, final List<Integer> nodeList){
      super();
      this.atStream = atStream;
      this.nodeList = nodeList;
   }
   
   
   /**
    * Gets the stream description for this event.
    */
   @Override
   public String getStreamDescription(){
      StringBuilder desc = new StringBuilder();
      
      desc.append(this.atStream.getStreamDescription());
      desc.append("_INSIDE");
      
      Collections.sort(this.nodeList);
      for(Integer node : this.nodeList){
         desc.append("_");
         desc.append(node);
      }
      return desc.toString();
   }
   
   
   /**
    * Executes the event using the given Lahar instance and returns a query result.
    * 
    * @param lahar
    * @return
    */
   @Override
   public QueryResult executeQuery(final BdbConnection    bdb,
                                   final LaharConnection  lahar, 
                                   final StreamKeyManager keyManager) throws IOException,
                                                                             ClassNotFoundException,
                                                                             SQLException,
                                                                             UnsupportedQueryException {
      QueryResult result = null;
      StreamKey   sk     = this.getStreamKey(keyManager);
      
      if(lahar.streamKeyExists(sk, this.bitmaskSchema)){
         result = lahar.loadCachedResult(sk, this.bitmaskSchema);
      } else {
         ParsedQuery     query        = this.getQuery(keyManager);
         
         try {
        	 result                       = lahar.executeQuery(query, true);
         } catch (UnsupportedQueryException ex) {
        	 ex.printStackTrace();
         }
         List<ITimestep> resultStream = StreamConverter.convertToMStream(result);
         
         // Store the result back into Lahar
         if(bdb.storeMStream(resultStream, sk.getKey(), this.bitmaskSchema, false)){
            lahar.moveStream(sk, this.bitmaskSchema);
         }
      }
      return result;
   }
   
   //--//
   
   /**
    * Gets a parsed query representing this event.
    */
   protected ParsedQuery getQuery(final StreamKeyManager keyManager) throws IOException,
                                                                            SQLException, 
                                                                            ClassNotFoundException,
                                                                            UnsupportedQueryException{
      StreamKey   atSk       = this.atStream.getStreamKey(keyManager);
      NFALinks    links      = new NFALinks();
      OrPredicate inRoomSet  = this.getDisjuncts();
      NFALink     inRoomLink = new NFALink(inRoomSet, null, new NFAStateID(0), true);

      links.addLinkToEnd(inRoomLink);
         
      NFA         myQuery = new NFA(links);
      ParsedQuery query   = new ParsedQuery(this.particleFilterSchema.getSchemaName()       ,
                                            new StreamPredicate_EqualityCheck(atSk.getKey()),
                                            new GroupingPredicate_ByStreamKey()             ,
                                            myQuery                                         ,
                                            null,
                                            RegType.INSTANTS                ,
                                            PostProcessingType.STREAMEXISTS);      
      return query;
   }
   

   /**
    * Gets the set of disjuncts.
    * @return
    */
   protected OrPredicate getDisjuncts() {
      List<IPredicate> inRoomPredicates = new ArrayList<IPredicate>();
      
      for(Integer node : this.nodeList){
         Attribute                        nodeId = new Attribute(AttributeType.LONG, (long)node.intValue());
         SingleAttributeEqualityPredicate inRoom = new SingleAttributeEqualityPredicate(1,nodeId);
         inRoomPredicates.add(inRoom);
      }
      TreeSet<IPredicate> disjunctPredicates = new TreeSet<IPredicate>(inRoomPredicates);
      OrPredicate         inRoomSet          = new OrPredicate(disjunctPredicates);
      
      return inRoomSet;
   }
}