package edu.washington.cs.rfid.mstreams.event;

import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import markovianStream.ITimestep;
import optimizer.IQueryPlan.PostProcessingType;
import optimizer.IQueryPlan.RegType;
import parser.ParsedQuery;
import parser.UnsupportedQueryException;
import query.nfa.NFA;
import query.nfa.NFALink;
import query.nfa.NFALinks;
import query.nfa.NFAStateID;
import query.nfa.nfaPredicates.SingleAttributeEqualityPredicate;
import query.streamPredicates.GroupingPredicate_ByStreamKey;
import query.streamPredicates.StreamPredicate_EqualityCheck;
import storageManager.schema.Attribute;
import storageManager.schema.Attribute.AttributeType;
import api.queryResults.QueryResult;
import edu.washington.cs.rfid.mstreams.data.StreamKey;
import edu.washington.cs.rfid.mstreams.data.db.BdbConnection;
import edu.washington.cs.rfid.mstreams.data.db.LaharConnection;
import edu.washington.cs.rfid.mstreams.data.db.StreamKeyManager;
import edu.washington.cs.rfid.mstreams.util.ResultStreamMerger;
import edu.washington.cs.rfid.mstreams.util.StreamConverter;

public class SceneEvent extends ScenicEvent {

   public List<ScenicEvent> subEvents;
   
   //--//
   
   /**
    * Creates a new SceneEvent 
    * @param subEvents
    */
   public SceneEvent(final List<ScenicEvent> subEvents){
      this.subEvents = subEvents;
   }
   
   
   /**
    * Gets the description for this scene, the description for the raw merge result
    * if specified by rawMerged, or the description for the final boolean result.
    *
    * @return
    */
   public String getStreamDescription(){
      StringBuilder keyBuilder = new StringBuilder();
      keyBuilder.append("_SCENE_");
      
      Collections.sort(this.subEvents);
      
      int len = this.subEvents.size();
      for(int i = 0; i < len - 1; i++){
         keyBuilder.append(this.subEvents.get(i).getStreamDescription());
         keyBuilder.append("_AND_");
      }
      keyBuilder.append(this.subEvents.get(len-1).getStreamDescription());
      
      return keyBuilder.toString();
   }
   
   
   /**
    * Executes the event using the given Lahar instance and returns a query result.
    * 
    * @param lahar
    * @return
    */
   @Override
   public QueryResult executeQuery(final BdbConnection    bdb,
                                   final LaharConnection  lahar, 
                                   final StreamKeyManager keyManager) throws IOException,
                                                                             ClassNotFoundException,
                                                                             SQLException,
                                                                             UnsupportedQueryException {
      QueryResult result = null;
      StreamKey   sk     = this.getStreamKey(keyManager);
      
      if(lahar.streamKeyExists(sk, this.bitmaskSchema)){
         result = lahar.loadCachedResult(sk, this.bitmaskSchema);
      } else {
      
         // 1) Execute all the sub-events
         List<QueryResult> subResults = new ArrayList<QueryResult>();
         
         for(ScenicEvent e : this.subEvents){
            QueryResult r = e.executeQuery(bdb, lahar, keyManager);
            subResults.add(r);
         }
         
         // 2) Merge the results and store as an intermediate stream in BDB and lahar
         List<ITimestep> merged  = ResultStreamMerger.mergeQueryResults(subResults);
         StreamKey       tempKey = this.getIntermediateStreamKey(keyManager);
         
         if(bdb.storeMStream(merged, tempKey.getKey(), this.bitmaskSchema, false)){
            lahar.moveStream(tempKey, this.bitmaskSchema);
            
            // 3) Issue the query for the scene and store the result in Lahar
            ParsedQuery     query        = this.getQuery(keyManager);
            result                       = lahar.executeQuery(query, true);
            List<ITimestep> resultStream = StreamConverter.convertToMStream(result);
            
            if(bdb.storeMStream(resultStream, sk.getKey(), this.bitmaskSchema, false)){
               lahar.moveStream(sk, this.bitmaskSchema);
            }
         }
      }
      return result;
   }
   
   
   /**
    * Gets a ParsedQuery that can be run on top of the stored merge of this
    * scene's sub event results.
    * @return
    */
   public ParsedQuery getQuery(final StreamKeyManager keyManager) throws IOException,
                                                                         SQLException, 
                                                                         ClassNotFoundException,
                                                                         UnsupportedQueryException {
      StreamKey                        tempKey    = this.getIntermediateStreamKey(keyManager);
      NFALinks                         links      = new NFALinks();
      Attribute                        combo      = new Attribute(AttributeType.INT, this.getOccurredBitMask());
      SingleAttributeEqualityPredicate occurs     = new SingleAttributeEqualityPredicate(0,combo);
      NFALink                          occursLink = new NFALink(occurs, null, new NFAStateID(0), true);

      links.addLinkToEnd(occursLink);

      NFA         myQuery = new NFA(links);
      ParsedQuery query   = new ParsedQuery(this.bitmaskSchema.getSchemaName()                 ,
                                            new StreamPredicate_EqualityCheck(tempKey.getKey()),
                                            new GroupingPredicate_ByStreamKey()                ,
                                            myQuery                                            ,
                                            null,
                                            RegType.INSTANTS                   ,
                                            PostProcessingType.STREAMEXISTS);
      
      return query;
   }
   
   //--//
   
   /**
    * Returns a bit mask representing the event that this scene occurred.
    * @return
    */
   public int getOccurredBitMask(){
      int bitMask      = 0;
      int numSubEvents = (null != this.subEvents)? this.subEvents.size() : 0;
      
      for(int i = 0; i < numSubEvents; i++){
         bitMask = bitMask | (1<<i);
      }
      return bitMask;
   }
}
