package edu.washington.cs.rfid.mstreams.event;

import java.io.IOException;
import java.sql.SQLException;

import parser.UnsupportedQueryException;
import api.queryResults.QueryResult;
import edu.washington.cs.rfid.mstreams.data.StreamKey;
import edu.washington.cs.rfid.mstreams.data.db.BdbConnection;
import edu.washington.cs.rfid.mstreams.data.db.LaharConnection;
import edu.washington.cs.rfid.mstreams.data.db.StreamKeyManager;
import edu.washington.cs.rfid.mstreams.schema.BitmaskStreamSchema;
import edu.washington.cs.rfid.mstreams.schema.ParticleFilterStreamSchema;


/**
 * A scenic event with data and methods for executing the Lahar query
 * corresponding to the event.
 * 
 * @author evan
 */
public abstract class ScenicEvent implements Comparable<ScenicEvent> {

   protected static final String DESC_INTERMEDIATE = "_INTERMEDIATE";
   protected static final String DESC_FINAL        = "_FINAL";
   
   //--//
   
   public ParticleFilterStreamSchema particleFilterSchema;
   public BitmaskStreamSchema        bitmaskSchema;
   
   //--//
   
   /**
    * Creates a new ScenicEvent.
    */
   public ScenicEvent(){
      this.particleFilterSchema = new ParticleFilterStreamSchema();
      this.bitmaskSchema        = new BitmaskStreamSchema();
   }
   
   
   /**
    * Executes the event using the given Lahar instance and returns a query result.
    * @param lahar
    * @return
    */
   public abstract QueryResult executeQuery(final BdbConnection    bdb  ,
                                            final LaharConnection  lahar, 
                                            final StreamKeyManager keyManager) throws IOException,
                                                                                      ClassNotFoundException,
                                                                                      SQLException,
                                                                                      UnsupportedQueryException;
   
   
   /**
    * Retrieves the final stream key for this event. 
    * 
    * @param keyManager
    * @return
    * @throws IOException
    * @throws ClassNotFoundException
    * @throws SQLException
    */
   public StreamKey getStreamKey(final StreamKeyManager keyManager) throws IOException,
                                                                           ClassNotFoundException, 
                                                                           SQLException {
      return this.getStreamKey(this.getFinalStreamDescription(this.getStreamDescription()),
                               keyManager);
   }
   
   
   /**
    * Gets the description of this event stream
    * @return
    */
   public abstract String getStreamDescription();
   
   
   /**
    * Compares this PrimitiveEvent to another event.
    */
   public int compareTo(ScenicEvent se) {
      String desc1 = this.getStreamDescription();
      String desc2 = se.getStreamDescription();
      
      return desc1.compareTo(desc2);
   }
   
   
   //--//
   
   
   /**
    * Retrieves the intermediate stream key for this event. 
    * 
    * @param keyManager
    * @return
    * @throws IOException
    * @throws ClassNotFoundException
    * @throws SQLException
    */
   protected StreamKey getIntermediateStreamKey(final StreamKeyManager keyManager) throws IOException,
                                                                                          ClassNotFoundException, 
                                                                                          SQLException {
      return this.getStreamKey(this.getIntermediateStreamDescription(this.getStreamDescription()),
                               keyManager);
   }
   
   
   /**
    * Retrieves the final stream key for this event. 
    * 
    * @param keyManager
    * @return
    * @throws IOException
    * @throws ClassNotFoundException
    * @throws SQLException
    */
   protected StreamKey getStreamKey(final String           desc,
                                    final StreamKeyManager keyManager) throws IOException,
                                                                              ClassNotFoundException, 
                                                                              SQLException {
      long      key  = keyManager.getStreamKey(desc);
      StreamKey sk   = new StreamKey(key, desc);
      
      return sk;
   }
   

   /**
    * Gets the intermediate result description of this event stream
    * @return
    */
   protected String getIntermediateStreamDescription(final String desc) {
      return desc + ScenicEvent.DESC_INTERMEDIATE;
   }
   
   
   /**
    * Gets the final description of this event stream
    * @return
    */
   protected String getFinalStreamDescription(final String desc) {
      return desc + ScenicEvent.DESC_FINAL;
   }
}
