package edu.washington.cs.rfid.mstreams.event;

import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.TreeSet;

import markovianStream.ITimestep;
import optimizer.IQueryPlan.PostProcessingType;
import optimizer.IQueryPlan.RegType;
import parser.ParsedQuery;
import parser.UnsupportedQueryException;
import query.nfa.NFA;
import query.nfa.NFALink;
import query.nfa.NFALinks;
import query.nfa.NFAStateID;
import query.nfa.nfaPredicates.IPredicate;
import query.nfa.nfaPredicates.OrPredicate;
import query.nfa.nfaPredicates.SingleAttributeEqualityPredicate;
import query.streamPredicates.GroupingPredicate_ByStreamKey;
import query.streamPredicates.StreamPredicate_EqualityCheck;
import storageManager.schema.Attribute;
import storageManager.schema.Attribute.AttributeType;
import api.queryResults.QueryResult;
import edu.washington.cs.rfid.mstreams.data.StreamKey;
import edu.washington.cs.rfid.mstreams.data.db.BdbConnection;
import edu.washington.cs.rfid.mstreams.data.db.LaharConnection;
import edu.washington.cs.rfid.mstreams.data.db.StreamKeyManager;
import edu.washington.cs.rfid.mstreams.util.ResultStreamMerger;
import edu.washington.cs.rfid.mstreams.util.StreamConverter;

public class SequenceEvent extends ScenicEvent {

   // TODO: Should also have some support for time between scenes!!
   public List<SceneEvent> sceneEvents;
   
   //--//
   
   /**
    * Creates a new SequenceEvent.
    * @param scenes
    */
   public SequenceEvent(final List<SceneEvent> scenes){
      this.sceneEvents = scenes;
   }
   
   
   /**
    * Gets the description for this scene, the description for the raw merge result
    * if specified by rawMerged, or the description for the final boolean result.
    *
    * @return
    */
   public String getStreamDescription(){
      return this.getSequenceDescription(this.sceneEvents);
   }
   
   
   
   
   /**
    * Executes the event using the given Lahar instance and returns a query result.
    * 
    * @param lahar
    * @return
    */
   @Override
   public QueryResult executeQuery(final BdbConnection    bdb,
                                   final LaharConnection  lahar, 
                                   final StreamKeyManager keyManager) throws IOException,
                                                                             ClassNotFoundException,
                                                                             SQLException,
                                                                             UnsupportedQueryException {
      QueryResult result = null;
      StreamKey   sk     = this.getStreamKey(keyManager);
      
      if(lahar.streamKeyExists(sk, this.bitmaskSchema)){
         result = lahar.loadCachedResult(sk, this.bitmaskSchema);
      } else {
         
         // 1) Execute each scene
         List<QueryResult> sceneResults = new ArrayList<QueryResult>();
         
         for(SceneEvent e : this.sceneEvents){
            QueryResult r = e.executeQuery(bdb, lahar, keyManager);
            sceneResults.add(r);
         }
         
         // 2) Merge the results and store as an intermediate stream in BDB and lahar
         List<ITimestep> merged  = ResultStreamMerger.mergeQueryResults(sceneResults);
         StreamKey       tempKey = this.getIntermediateStreamKey(keyManager);
         
         if(bdb.storeMStream(merged, tempKey.getKey(), this.bitmaskSchema, false)){
            lahar.moveStream(tempKey, this.bitmaskSchema);
            
            // 3) Issue the query for the sequence and store the result in Lahar
            ParsedQuery     query        = this.getQuery(keyManager);
            result                       = lahar.executeQuery(query, true);
            List<ITimestep> resultStream = StreamConverter.convertToMStream(result);
            
            if(bdb.storeMStream(resultStream, sk.getKey(), this.bitmaskSchema, false)){
               lahar.moveStream(sk, this.bitmaskSchema);
            }
         }
      }
      return result;
   }
   
   
   //--//
   
   
   /**
    * Gets a ParsedQuery that can be run on top of the stored merge of this
    * scene's sub event results.
    * TODO: Note that this query ignores non-instant transitions between 
    * scenes (i.e. self-loops) for now.
    * @return
    */
   protected ParsedQuery getQuery(final StreamKeyManager keyManager) throws IOException ,
                                                                            SQLException, 
                                                                            ClassNotFoundException,
                                                                            UnsupportedQueryException {
      StreamKey tempKey = this.getIntermediateStreamKey(keyManager);
      NFALinks  links   = new NFALinks();

      // Assemble the query as an NFA with consecutive links, one for each scene:
      // ex: 1**, *1*, **1
      for(int i = 0; i < this.sceneEvents.size(); i++){
         List<IPredicate> predicates = this.getSceneOccurredDisjuncts(i);
         if(0 == predicates.size()){
            continue;
         }
         TreeSet<IPredicate> predicateTree = new TreeSet<IPredicate>(this.getSceneOccurredDisjuncts(i));
         OrPredicate         occurences    = new OrPredicate(predicateTree);
         NFALink             occursLink    = new NFALink(occurences, null, new NFAStateID(0), true);
         links.addLinkToEnd(occursLink);
      }
      NFA         myQuery = new NFA(links);
      ParsedQuery query   = new ParsedQuery(this.bitmaskSchema.getSchemaName()                 ,
                                            new StreamPredicate_EqualityCheck(tempKey.getKey()),
                                            new GroupingPredicate_ByStreamKey()                ,
                                            myQuery                                            ,
                                            null,
                                            RegType.INSTANTS                   ,
                                            PostProcessingType.STREAMEXISTS);
      return query;
   }
   
   
   /**
    * Returns a list of predicates that cover all possible ways that a particular
    * scene could have occurred.
    * TODO: NOTE NOTE NOTE: HERE WE ASSUME THAT THE FIRST SCENE IS THE MSB IN THE
    *       OCCURRED BIT MASK -- IS IT REALLY SO IN THE MERGED STREAM WE ARE QUERYING??
    * @return
    */
   protected List<IPredicate> getSceneOccurredDisjuncts(int i){
      List<IPredicate> occursPredicates = new ArrayList<IPredicate>();
      boolean[][]      comboMatrix      = this.getComboMatrix(this.sceneEvents.size());
      int[]            combos           = this.getBitMasksHolding(i, comboMatrix);
      
      for(int j = 0; j < combos.length; j++){
         Attribute                        combo  = new Attribute(AttributeType.INT, combos[j]);         
         SingleAttributeEqualityPredicate occurs = new SingleAttributeEqualityPredicate(0,combo);
         occursPredicates.add(occurs);
      }
      return occursPredicates;
   }

   
   /**
    * Gets a matrix representing every possible combination of streams for a given number of streams.
    * @param numStreams
    * @return
    */
   protected boolean[][] getComboMatrix(int numStreams){
      int         combinations = (int)Math.pow(2.0, numStreams);
      int[]       modulus      = new int[numStreams];
      boolean[][] comboMatrix  = new boolean[combinations][numStreams]; // rows, cols
      
      for(int i = 0; i < numStreams; i++){
         modulus[i] = (0 == i)? 1 : modulus[i-1] + modulus[i-1];
      }
      
      boolean val = true;
      for(int j = 0; j < numStreams; j++){
         for(int i = 0; i < combinations; i++){
            int switchVal     = i % modulus[j];
            val               = (i == 0 || switchVal == 0)? !val : val;
            comboMatrix[i][(numStreams - 1 - j)] = val;
         }
      }
      return comboMatrix;
   }
   
   
   /**
    * Creates an array of bit masks, one for each row of the given combo matrix.
    * 
    * @precond the comboMatrix is not null
    * @param comboMatrix
    * @return
    */
   protected int[] getBitMasksHolding(int fixedIndex, boolean[][] comboMatrix){
      int[] masks     = new int[comboMatrix.length/2];
      int   maskIndex = 0;
      
      for(int i = 0; i < comboMatrix.length; i++){
         if(comboMatrix[i][fixedIndex]){
            masks[maskIndex] = this.convertToBitmask(comboMatrix[i]);
            maskIndex++;
         }
      }
      return masks;
   }
   
   
   /**
    * Converts an array of boolean flags into an integer bitmask.
    * @param flags
    * @return
    */
   protected int convertToBitmask(boolean[] flags){
      int mask          = 0x0000;
      int msbShiftIndex = flags.length-1;
      
      for(int i = 0; i  < flags.length; i++){
         mask = (flags[i])? (mask | (1<<(msbShiftIndex-i))) : (mask | (0<<(msbShiftIndex-i)));
      }
      return mask;
   }
   
   
   /**
    * Gets a list containing the stream key for each sub-sequence of this sequence.
    */
   protected List<String> getSubsequenceStreamKeys(){
      List<String> keys = new ArrayList<String>();
      
      for(int i = 0; i < this.sceneEvents.size(); i++){
         List<SceneEvent> subSequence = new ArrayList<SceneEvent>();
         for(int j = 0; j <= i; j++){
            subSequence.add(this.sceneEvents.get(j));
         }
         keys.add(this.getSequenceDescription(subSequence));
      }
      return keys;
   }
   
   
   /**
    * A helper method
    * @param scenes
    * @return
    */
   protected String getSequenceDescription(List<SceneEvent> scenes){
      StringBuilder keyBuilder = new StringBuilder();
      keyBuilder.append("_SEQUENCE_");
      
      int len = scenes.size();
      for(int i = 0; i < len - 1; i++){
         keyBuilder.append(scenes.get(i).getStreamDescription());
         keyBuilder.append("_THEN_");
      }
      keyBuilder.append(scenes.get(len-1).getStreamDescription());
      
      return keyBuilder.toString();
   }
}