package edu.washington.cs.rfid.mstreams.test;

import java.io.IOException;
import java.sql.SQLException;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import markovianStream.ITimestep;
import operators.post.IFinalOutput;
import optimizer.IQueryPlan.PostProcessingType;
import optimizer.IQueryPlan.RegType;
import parser.ParsedQuery;
import parser.UnsupportedQueryException;
import query.nfa.NFA;
import query.nfa.NFALink;
import query.nfa.NFALinks;
import query.nfa.NFAStateID;
import query.nfa.nfaPredicates.NegatedPredicate;
import query.nfa.nfaPredicates.SingleAttributeEqualityPredicate;
import query.streamPredicates.GroupingPredicate_ByStreamKey;
import query.streamPredicates.StreamPredicate_EqualityCheck;
import storageManager.schema.Attribute;
import storageManager.schema.IStreamSchema;
import storageManager.schema.ITuple;
import storageManager.schema.Attribute.AttributeType;
import api.queryResults.GroupResult;
import api.queryResults.QueryResult;
import edu.washington.cs.rfid.mstreams.data.ResultStreamCPTData;
import edu.washington.cs.rfid.mstreams.data.ResultStreamMarginalData;
import edu.washington.cs.rfid.mstreams.data.ScenicEventExtraData;
import edu.washington.cs.rfid.mstreams.data.db.BdbConnection;
import edu.washington.cs.rfid.mstreams.data.db.LaharConnection;
import edu.washington.cs.rfid.mstreams.data.db.StreamKeyManager;
import edu.washington.cs.rfid.mstreams.event.AtEvent;
import edu.washington.cs.rfid.mstreams.event.FarEvent;
import edu.washington.cs.rfid.mstreams.event.InsideEvent;
import edu.washington.cs.rfid.mstreams.event.NearEvent;
import edu.washington.cs.rfid.mstreams.event.OutsideEvent;
import edu.washington.cs.rfid.mstreams.event.SceneEvent;
import edu.washington.cs.rfid.mstreams.event.ScenicEvent;
import edu.washington.cs.rfid.mstreams.event.WithEvent;
import edu.washington.cs.rfid.mstreams.event.WithoutEvent;

public class QueryTestMain {

   public static final String DEFAULT_BRB_DIR      = "/data/bdbRfidStreams";
   public static final String DEFAULT_LAHAR_DB_DIR = "/data/db_loc/RfidDB";
   
	public static final  String TYPE               = "MARKOVIAN";
	
	public static final  String FIRST_STREAM_DESC  = "2254716205721255785_1226633102799_1226633815799";
	public static final  String SECOND_STREAM_DESC = "2254716205721255785_1226636093426_1226636743426";
	public static final  long   DEFAULT_TAG        = 2254716205721255785L;
	public static final  int    DEFAULT_NODE1      = 375787;
	public static final  int    DEFAULT_NODE2      = 316271;

	public static final  int    EVENT_INSIDE       = 1;
	public static final  int    EVENT_OUTSIDE      = 2;
	public static final  int    EVENT_NEAR         = 3;
	public static final  int    EVENT_FAR          = 4;
	public static final  int    EVENT_WITH         = 5;
	public static final  int    EVENT_WITHOUT      = 6;
	
	public static final  int    EVENT_SCENE_ONE    = 7;
	public static final  int    EVENT_SCENE_TWO    = 8;
	public static final  int    EVENT_SEQUENCE     = 9;
	
	//--//
	
	/**
	 * Tests Lahar by running simples queries.
	 * @param args
	 */
	public static void main(String[] args) {
	   
	   StreamKeyManager keyManager = new StreamKeyManager();
	   BdbConnection    bdb        = new BdbConnection(QueryTestMain.DEFAULT_BRB_DIR);
	   LaharConnection  lahar      = new LaharConnection(QueryTestMain.DEFAULT_BRB_DIR,QueryTestMain.DEFAULT_LAHAR_DB_DIR);
	   
	   AtEvent          atEvent1   = new AtEvent(QueryTestMain.FIRST_STREAM_DESC );
	   AtEvent          atEvent2   = new AtEvent(QueryTestMain.SECOND_STREAM_DESC);
	   
	   int              testNum    = QueryTestMain.EVENT_INSIDE;
	   
	   try {
	   switch(testNum){
   	   case QueryTestMain.EVENT_INSIDE:
   	   {
   	      List<Integer> nodes = new ArrayList<Integer>();
   	      nodes.add(QueryTestMain.DEFAULT_NODE1);
   	      
   	      InsideEvent inside = new InsideEvent(atEvent1, nodes);
   	      QueryResult result = inside.executeQuery(bdb, lahar, keyManager);
   	      
   	      QueryTestMain.printQueryResults(result);
   	      break;
   	   }
   	   case QueryTestMain.EVENT_OUTSIDE:
         {
            List<Integer> nodes = new ArrayList<Integer>();
            nodes.add(QueryTestMain.DEFAULT_NODE1);
            
            OutsideEvent outside = new OutsideEvent(atEvent1, nodes);
            QueryResult  result  = outside.executeQuery(bdb, lahar, keyManager);
            
            QueryTestMain.printQueryResults(result);
            break;
         }
   	   case QueryTestMain.EVENT_WITH:
         {
            List<AtEvent> events = new ArrayList<AtEvent>();
            events.add(atEvent1);
            events.add(atEvent2);
            
            WithEvent   with   = new WithEvent(events);
            QueryResult result = with.executeQuery(bdb, lahar, keyManager);
            
            QueryTestMain.printQueryResults(result);
            break;
         }
   	   case QueryTestMain.EVENT_WITHOUT:
         {
            List<AtEvent> events = new ArrayList<AtEvent>();
            events.add(atEvent1);
            events.add(atEvent2);
            
            WithoutEvent without = new WithoutEvent(events);
            QueryResult  result  = without.executeQuery(bdb, lahar, keyManager);
            
            QueryTestMain.printQueryResults(result);
            break;
         }
   	   case QueryTestMain.EVENT_NEAR:
         {
            List<AtEvent> events = new ArrayList<AtEvent>();
            events.add(atEvent1);
            events.add(atEvent2);
            
            NearEvent   near   = new NearEvent(events, 5.0);
            QueryResult result = near.executeQuery(bdb, lahar, keyManager);
            
            QueryTestMain.printQueryResults(result);
            break;
         }
   	   case QueryTestMain.EVENT_FAR:
         {
            List<AtEvent> events = new ArrayList<AtEvent>();
            events.add(atEvent1);
            events.add(atEvent2);
            
            FarEvent    far    = new FarEvent(events, 5.0);
            QueryResult result = far.executeQuery(bdb, lahar, keyManager);
            
            QueryTestMain.printQueryResults(result);
            break;
         }
   	   case QueryTestMain.EVENT_SCENE_ONE:
         {
            List<ScenicEvent> primitives  = new ArrayList<ScenicEvent>();
            List<Integer>     insideNodes = new ArrayList<Integer>();
            insideNodes.add(QueryTestMain.DEFAULT_NODE1);
            InsideEvent inside = new InsideEvent(atEvent1, insideNodes);
            
            List<Integer>     outsideNodes = new ArrayList<Integer>();
            outsideNodes.add(QueryTestMain.DEFAULT_NODE2);
            OutsideEvent outside = new OutsideEvent(atEvent1, outsideNodes);
            
            primitives.add(inside);
            primitives.add(outside);
            
            SceneEvent  scene  = new SceneEvent(primitives);
            QueryResult result = scene.executeQuery(bdb, lahar, keyManager);
            QueryTestMain.printQueryResults(result);
            break;
         }
   	   default:
	         break;
	   }
	   } catch(SQLException e){
	      e.printStackTrace();
	   } catch(ClassNotFoundException e){
	      e.printStackTrace();
	   } catch(IOException e){
	      e.printStackTrace();
	   } catch (UnsupportedQueryException e) {
		   e.printStackTrace();
	   }
	}
	
		
  /**
    * Print the results of a query.
    * @param result
    */
   public static void printQueryResults(QueryResult result){
      
      long startTime = 1226633102799L;
      long totalTime = 1226633815799L - startTime;
      
       List<GroupResult> grpResults = result.getResults();
       
      for(GroupResult grpResult : grpResults){
         System.out.println("Group Result for group " + grpResult.getGroupID());
         List<IFinalOutput> outputs = grpResult.getResults();
         for(IFinalOutput output : outputs){
            
            ScenicEventExtraData extra         = (ScenicEventExtraData) output.getMetaData();
            double               traceProgress = ((double)(extra.timestamp - startTime)/(double)totalTime)*100.0;
            DecimalFormat        df1           = new DecimalFormat("###");
            DecimalFormat        df2           = new DecimalFormat("#.##############");
            
            String               outputStr = "seqID: " + output.getIndex() +
                                             " time: " + extra.timestamp   +
                                             " prob: " + df2.format(output.getDistribution().getProbability(1)) +
                                             " loc:  " + df1.format(traceProgress) + "%";
            System.out.println("\t" + outputStr);
         }
      }
   }
   
   
   /**
    * Prints a list of timesteps.
    * @param timesteps
    */
   public static void printTimesteps(List<ITimestep> timesteps){
      for(ITimestep ts : timesteps){
         Map<ITuple,Double>             marginalMap = ts.getMarginal();
         Map<ITuple,Map<ITuple,Double>> cptMap      = ts.getCPT().getValues();
         ScenicEventExtraData          extra       = (ScenicEventExtraData)ts.getExtraData();

         ResultStreamMarginalData marginal = new ResultStreamMarginalData(extra.timestamp,marginalMap);
         ResultStreamCPTData      cpt      = new ResultStreamCPTData(extra.timestamp, extra.timestamp + 1000L, cptMap);

         System.out.println(marginal.toPrettyString());
         System.out.println(cpt.toPrettyString());
         System.out.println();
      }
   }
	
	
   //----------------------------- CURRENTLY UNUSED HELPER METHODS -----------------------------------
   
   
	/**
	 * Create a single link query for "in room 35787"
	 * @param schema
	 * @return
	 */
	public static ParsedQuery getSingleLinkQuery(IStreamSchema schema){
		NFALinks                         links      = new NFALinks();
		Attribute                        nodeId     = new Attribute(AttributeType.LONG, (long)375787L);
		SingleAttributeEqualityPredicate inRoom     = new SingleAttributeEqualityPredicate(1,nodeId);
		NFALink                          inRoomLink = new NFALink(inRoom, null, new NFAStateID(0), true);
		  
		links.addLinkToEnd(inRoomLink);
		   
		NFA         myQuery = new NFA(links);
		ParsedQuery query   = new ParsedQuery(schema.getSchemaName(),
											           new StreamPredicate_EqualityCheck(QueryTestMain.FIRST_STREAM_DESC),
											           new GroupingPredicate_ByStreamKey(),
											           myQuery,
											           null,
											           RegType.INSTANTS,
											           PostProcessingType.STREAMEXISTS);
		return query;
	}
	
	
	/**
    * Create a single link query for "in room 35787"
    * @param schema
    * @return
    */
   public static ParsedQuery getSingleLinkQuery(IStreamSchema schema, String streamKey){
      NFALinks                         links      = new NFALinks();
      Attribute                        nodeId     = new Attribute(AttributeType.LONG, (long)375787L);
      SingleAttributeEqualityPredicate inRoom     = new SingleAttributeEqualityPredicate(1,nodeId);
      NFALink                          inRoomLink = new NFALink(inRoom, null, new NFAStateID(0), true);
        
      links.addLinkToEnd(inRoomLink);
         
      NFA         myQuery = new NFA(links);
      ParsedQuery query   = new ParsedQuery(schema.getSchemaName(),
                                   new StreamPredicate_EqualityCheck(streamKey),
                                   new GroupingPredicate_ByStreamKey(),
                                   myQuery,
                                   null,
                                   RegType.INSTANTS,
                                   PostProcessingType.STREAMEXISTS);
      return query;
   }
	
	
	/**
    * Create a negated single link query for "outside room 35787"
    * @param schema
    * @return
    */
   public static ParsedQuery getNegatedSingleLinkQuery(IStreamSchema schema){
      NFALinks                         links      = new NFALinks();
      Attribute                        nodeId     = new Attribute(AttributeType.LONG, (long)375787L);
      SingleAttributeEqualityPredicate inRoom     = new SingleAttributeEqualityPredicate(1,nodeId);
      NegatedPredicate                 outRoom    = new NegatedPredicate(inRoom);
      NFALink                          outRoomLink = new NFALink(outRoom, null, new NFAStateID(0), true);
      
      links.addLinkToEnd(outRoomLink);
      
      NFA         myQuery = new NFA(links);
      ParsedQuery query   = new ParsedQuery(schema.getSchemaName(),
                                            new StreamPredicate_EqualityCheck(QueryTestMain.FIRST_STREAM_DESC),
                                            new GroupingPredicate_ByStreamKey(),
                                            myQuery,
                                            null,
                                            RegType.INSTANTS,
                                            PostProcessingType.STREAMEXISTS);
      return query;
   }
   
   
	/**
	 * Create a two-link query, for "entering room 313413 --> 367619.
	 * @param schema
	 * @return
	 */
	public static ParsedQuery getTwoLinkQuery(IStreamSchema schema){
		NFALinks                         links       = new NFALinks();
		Attribute                        firstNode   = new Attribute(AttributeType.LONG, (long)313413L);
		Attribute                        secondNode  = new Attribute(AttributeType.LONG, (long)367619L);		
		SingleAttributeEqualityPredicate outsideRoom = new SingleAttributeEqualityPredicate(1,firstNode );
		SingleAttributeEqualityPredicate insideRoom  = new SingleAttributeEqualityPredicate(1,secondNode);
		NFALink                          outRoomLink = new NFALink(outsideRoom, null, new NFAStateID(0), false);
		NFALink                          inRoomLink  = new NFALink(insideRoom,  null, new NFAStateID(1), true);
		  
		links.addLinkToEnd(outRoomLink);
		links.addLinkToEnd(inRoomLink );
		   
		NFA         myQuery = new NFA(links);
		ParsedQuery query   = new ParsedQuery(schema.getSchemaName(),
											  new StreamPredicate_EqualityCheck(QueryTestMain.FIRST_STREAM_DESC),
											  new GroupingPredicate_ByStreamKey(),
											  myQuery,
											  null,
											  RegType.INSTANTS,
											  PostProcessingType.STREAMEXISTS);
		return query;
	}
	
	
	/**
	 * Create a three-link query, for "entering room 313413 --> 367619 --> 316271"
	 * @param schema
	 * @return
	 */
	public static ParsedQuery getThreeLinkQuery(IStreamSchema schema){
		NFALinks                         links       = new NFALinks();
		Attribute                        firstNode   = new Attribute(AttributeType.LONG, (long)313413L);
		Attribute                        secondNode  = new Attribute(AttributeType.LONG, (long)367619L);
		Attribute                        thirdNode   = new Attribute(AttributeType.LONG, (long)316271L);
		SingleAttributeEqualityPredicate outsideRoom = new SingleAttributeEqualityPredicate(1,firstNode );
		SingleAttributeEqualityPredicate insideDoor  = new SingleAttributeEqualityPredicate(1,secondNode);
		SingleAttributeEqualityPredicate insideRoom  = new SingleAttributeEqualityPredicate(1,thirdNode);
		NFALink                          outRoomLink = new NFALink(outsideRoom, null, new NFAStateID(0), false);
		NFALink                          inDoorLink  = new NFALink(insideDoor,  null, new NFAStateID(1), false);
		NFALink                          inRoomLink  = new NFALink(insideRoom,  null, new NFAStateID(2), true);

		links.addLinkToEnd(outRoomLink);
		links.addLinkToEnd(inDoorLink );
		links.addLinkToEnd(inRoomLink );

		NFA         myQuery = new NFA(links);
		ParsedQuery query   = new ParsedQuery(schema.getSchemaName(),
											  new StreamPredicate_EqualityCheck(QueryTestMain.FIRST_STREAM_DESC),
											  new GroupingPredicate_ByStreamKey(),
											  myQuery,
											  null,
											  RegType.INSTANTS,
											  PostProcessingType.STREAMEXISTS);
		return query;
	}
	
	
	/**
	 * Create a two-link query with a self-loop.
	 * @param schema
	 * @return
	 */
	public static ParsedQuery getTwoLinkLoopQuery(IStreamSchema schema){
		NFALinks                         links        = new NFALinks();
		Attribute                        firstNode    = new Attribute(AttributeType.LONG, (long)313413L);
		Attribute                        secondNode   = new Attribute(AttributeType.LONG, (long)367619L);		
		SingleAttributeEqualityPredicate outsideRoom  = new SingleAttributeEqualityPredicate(1,firstNode );
		SingleAttributeEqualityPredicate insideRoom   = new SingleAttributeEqualityPredicate(1,secondNode);
		NegatedPredicate                 notInRoom    = new NegatedPredicate(insideRoom);
		NFALink                          outRoomLink  = new NFALink(outsideRoom, notInRoom, new NFAStateID(0), false);
		NFALink                          inRoomLink   = new NFALink(insideRoom,  null, new NFAStateID(1), true);

		links.addLinkToEnd(outRoomLink);
		links.addLinkToEnd(inRoomLink );
		   
		NFA         myQuery = new NFA(links);
		ParsedQuery query   = new ParsedQuery(schema.getSchemaName(),
											  new StreamPredicate_EqualityCheck(QueryTestMain.FIRST_STREAM_DESC),
											  new GroupingPredicate_ByStreamKey(),
											  myQuery,
											  null,
											  RegType.INSTANTS,
											  PostProcessingType.STREAMEXISTS);
		return query;
	}
}