package edu.washington.cs.rfid.mstreams.util;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import markovianStream.CPT;
import markovianStream.ITimestep;
import markovianStream.Timestep_Imprecise;
import operators.distributions.IResultDistribution;
import operators.post.IFinalOutput;
import storageManager.schema.Attribute;
import storageManager.schema.ITuple;
import storageManager.schema.Tuple;
import storageManager.schema.Attribute.AttributeType;
import api.queryResults.GroupResult;
import api.queryResults.QueryResult;
import edu.washington.cs.rfid.mstreams.data.ParticleFilterStream;
import edu.washington.cs.rfid.mstreams.data.ParticleFilterStreamCPTData;
import edu.washington.cs.rfid.mstreams.data.ParticleFilterStreamMarginalData;
import edu.washington.cs.rfid.mstreams.data.ScenicEventExtraData;
import edu.washington.cs.rfid.mstreams.data.ParticleFilterStream.MarginalCPTPair;
import edu.washington.cs.rfid.mstreams.data.ParticleFilterStreamMarginalData.NodeProbPair;
import edu.washington.cs.rfid.mstreams.exception.MStreamSyntaxException;


/**
 * Class to load particle filter stream files into timestamp lists.
 * 
 * @author evan
 */
public class StreamConverter {
   
   /**
    * Converts a particle filter stream object into an MStream.
    * 
    * @param stream
    * @return
    * @throws MStreamSyntaxException
    * @throws IOException
    */
   public static List<ITimestep> convertToMStream(ParticleFilterStream stream) throws MStreamSyntaxException, 
                                                                               IOException {
      List<ITimestep> timesteps = convertToConsistentMStream_helper(stream);

      return timesteps;
   }
   
   
   /**
    * Converts a particle filter stream file into an MStream.
    * 
    * @param filename
    * @return
    * @throws MStreamSyntaxException
    * @throws IOException
    */
   public static List<ITimestep> convertToMStream(String filename) throws MStreamSyntaxException, 
                                                                   IOException {
      List<ITimestep>            timesteps = new ArrayList<ITimestep>();
      ParticleFilterStreamParser parser    = new ParticleFilterStreamParser();
      ParticleFilterStream       stream    = parser.parseSmoothedPF(filename);
      
      timesteps = StreamConverter.convertToConsistentMStream_helper(stream);
      
      return timesteps;
   }
   
   
   /**
    * Converts a query result stream into an MStream.
    *  
    * @param result
    * @return
    */
   public static List<ITimestep> convertToMStream(QueryResult result){
      
      List<ITimestep>   timesteps  = new ArrayList<ITimestep>();
      List<GroupResult> grpResults = result.getResults();
       
      for(GroupResult grpResult : grpResults){
         List<IFinalOutput> outputs = grpResult.getResults();
         
         if(0 < outputs.size()){
            IFinalOutput           output       = outputs.get(0);
            ScenicEventExtraData extra        = new ScenicEventExtraData(((ScenicEventExtraData)output.getMetaData()).timestamp);
            Map<ITuple,Double>   lastMarginal = StreamConverter.getMarginal(output);
            int                  seqIndex     = 0;
            
            for(int i = 1; i < outputs.size(); i++){
               output                                       = outputs.get(i);
               Map<ITuple,Double>               curMarginal = StreamConverter.getMarginal(output);
               Map<ITuple, Map<ITuple, Double>> cpt         = MStreamCreatorUtil.getDefaultCPT(lastMarginal, curMarginal);
               ITimestep                        ts          = new Timestep_Imprecise(seqIndex, lastMarginal, new CPT(cpt));
               
               ts.setExtraData(extra);
               timesteps.add(ts);
               extra        = new ScenicEventExtraData(((ScenicEventExtraData)output.getMetaData()).timestamp);
               lastMarginal = curMarginal;
               seqIndex++;
            }
            Map<ITuple, Map<ITuple, Double>> cpt = MStreamCreatorUtil.getDefaultCPT(lastMarginal);
            ITimestep                        ts  = new Timestep_Imprecise(seqIndex, lastMarginal, new CPT(cpt));
            ts.setExtraData(extra);
            timesteps.add(ts);
         }
      }
      return timesteps;
   }
   
   
   /**
    * Takes a result stream and returns a negated version of it.
    * 
    * @return
    */
   public static List<ITimestep> negateResultStream(final List<ITimestep> stream){
      List<ITimestep> negated = new ArrayList<ITimestep>();
      
      if(0 < stream.size()){
         ITimestep            tuple        = stream.get(0);
         ScenicEventExtraData extra        = new ScenicEventExtraData(((ScenicEventExtraData)tuple.getExtraData()).timestamp);
         Map<ITuple,Double>   lastMarginal = StreamConverter.getNegatedMarginal(tuple.getMarginal());
         int                  seqIndex     = 0;
         
         for(int i = 1; i < stream.size(); i++){
            tuple                                      = stream.get(i);
            Map<ITuple,Double>             curMarginal = StreamConverter.getNegatedMarginal(tuple.getMarginal());
            Map<ITuple,Map<ITuple,Double>> cpt         = MStreamCreatorUtil.getDefaultCPT(lastMarginal, curMarginal);
            ITimestep                      ts          = new Timestep_Imprecise(seqIndex, lastMarginal, new CPT(cpt));
            
            ts.setExtraData(extra);
            negated.add(ts);
            extra        = new ScenicEventExtraData(((ScenicEventExtraData)tuple.getExtraData()).timestamp);
            lastMarginal = curMarginal;
            seqIndex++;
         }
         Map<ITuple, Map<ITuple, Double>> cpt = MStreamCreatorUtil.getDefaultCPT(lastMarginal);
         ITimestep                        ts  = new Timestep_Imprecise(seqIndex, lastMarginal, new CPT(cpt));
         ts.setExtraData(extra);
         negated.add(ts);
      }
      return negated;
   }
   
   
   //-------------------------- PARTICLE FILTER STREAM HELPER METHODS ------------------------//
   
   
   /**
    * Converts an old MStream to a new one with the specified location and stream key.
    *
    */
   protected static List<ITimestep> convertToMStream_helper(ParticleFilterStream file){
      List<ITimestep> ret   = new ArrayList<ITimestep>();
      int             index = 0;
      
      for(MarginalCPTPair pair : file.sequence){
         ITimestep ts = StreamConverter.generateITimestep(index, pair.marginal, pair.cpt);
         
         if(null != ts){
            ret.add(ts);
            index++;
         }
      }
      return ret;
   }
   
   
   /**
    * Converts an particle filter stream to a new one with the specified location and stream key.
    * Uses the first marginal and the "nextMarginal" method to create a stream that is
    * consistent - currently (8/09) the particle filter does not create consistent streams.
    * 
    * @param file
    * @return
    */
   protected static List<ITimestep> convertToConsistentMStream_helper(ParticleFilterStream file){
      List<ITimestep> ret   = new ArrayList<ITimestep>();
      int             index = 0;
      
      if(0 < file.sequence.size()) {
      
         Map<ITuple,Double> lastMarginalMap = StreamConverter.marginalToMap(file.sequence.firstElement().marginal);
         
         for(MarginalCPTPair pair : file.sequence){
            ITimestep ts = StreamConverter.generateITimestep(index, lastMarginalMap, pair.cpt);
            
            if(null != ts){
               lastMarginalMap = ((Timestep_Imprecise)ts).computeNextMarginal();
               ret.add(ts);
               index++;
            }
         }
      }
      return ret;
   }
   
   
   /**
    * Creates a new ITimestep.
    * 
    * @param index
    * @param marginal
    * @param cptData
    * @return
    */
   private static ITimestep generateITimestep(int                              index, 
                                       ParticleFilterStreamMarginalData marginal, 
                                       ParticleFilterStreamCPTData      cptData){
      ITimestep ts = null;
      
      Map<ITuple, Double>              marginalMap = StreamConverter.marginalToMap(marginal);
      Map<ITuple, Map<ITuple, Double>> cptMap      = StreamConverter.constructCPT(cptData);
      CPT                              cpt         = new CPT(cptMap);
      
      ts = new Timestep_Imprecise(index, marginalMap, cpt);
      
      ScenicEventExtraData extra = new ScenicEventExtraData(marginal.timestamp);
      ts.setExtraData(extra);
      
      return ts;
   }
   
   
   /**
    * Creates an ITimestep.
    * 
    * @param index
    * @param marginalMap
    * @param cptData
    * @return
    */
   protected static ITimestep generateITimestep(int                         index, 
                                         Map<ITuple,Double>          marginalMap, 
                                         ParticleFilterStreamCPTData cptData){
      
      Map<ITuple, Map<ITuple, Double>> cptMap = StreamConverter.constructCPT(cptData);
      CPT                              cpt    = new CPT(cptMap);
      ITimestep                        ts     = new Timestep_Imprecise(index, marginalMap, cpt);      
      ScenicEventExtraData             extra  = new ScenicEventExtraData(cptData.startTimestamp);
      ts.setExtraData(extra);
      
      return ts;
   }
   
   
   /**
    * Creates a Map from a marginal.
    * 
    * @param marginal
    * @return
    */
   protected static Map<ITuple, Double> marginalToMap(ParticleFilterStreamMarginalData marginal) {
      Map<ITuple, Double> marginalMap = new HashMap<ITuple, Double>();
      double              totalP      = 0.0;
      
      for(NodeProbPair pairs : marginal.nodeProbPairs){
         Attribute   tagIdAttribute = new Attribute(AttributeType.LONG, (long)marginal.tagId);
         Attribute   nodeAttribute  = new Attribute(AttributeType.LONG, (long)pairs.node    );
         Attribute[] attributeArr   = new Attribute[2];
         attributeArr[0]            = tagIdAttribute;
         attributeArr[1]            = nodeAttribute;
         ITuple      tuple          = new Tuple(attributeArr);
         
         marginalMap.put(tuple, pairs.prob);
         totalP += pairs.prob;
      }
      
      // Sanity check
      if (Math.abs(1.0-totalP) > 1e-3)
         throw new IllegalStateException("Error: marginal's probabilities do not sum to 1.0");
            
      return marginalMap;
   }
   
   
   /**
    * Creates a CPT Map.
    * 
    * @param cpt
    * @return
    */
   protected static Map<ITuple, Map<ITuple, Double>> constructCPT(ParticleFilterStreamCPTData cpt) {
      Map<ITuple, Map<ITuple, Double>> cptMap = new HashMap<ITuple, Map<ITuple, Double>>();
      
      for(Long node1 : cpt.transitionMap.keySet()){
         HashMap<Long,Double> map = cpt.transitionMap.get(node1);
         
         // First find the sum of all transition probs starting at node1
         double totalP = 0.0;
         for(Double prob : map.values()){
            totalP += prob;
         }
         
         // Now create the tuple map with normalized transition probabilities
         Attribute   tagIdAttribute = new Attribute(AttributeType.LONG, (long)cpt.tagId        );
         Attribute   node1Attribute = new Attribute(AttributeType.LONG, (long)node1.longValue());
         Attribute[] attribute1Arr  = new Attribute[2];
         attribute1Arr[0]           = tagIdAttribute;
         attribute1Arr[1]           = node1Attribute;
         ITuple      tuple1         = new Tuple(attribute1Arr);
         
         Map<ITuple,Double> tupleMap = new HashMap<ITuple,Double>();
         
         for(Long node2 : map.keySet()){
            Attribute   tagIdAttribute2 = new Attribute(AttributeType.LONG, (long)cpt.tagId        );
            Attribute   node2Attribute  = new Attribute(AttributeType.LONG, (long)node2.longValue());
            Attribute[] attribute2Arr   = new Attribute[2];
            attribute2Arr[0]            = tagIdAttribute2;
            attribute2Arr[1]            = node2Attribute;
            ITuple      tuple2          = new Tuple(attribute2Arr);
            Double      prob            = map.get(node2)/totalP;
            
            tupleMap.put(tuple2, prob);
         }
         cptMap.put(tuple1, tupleMap);
      }
      return cptMap;
    }
   
   
   //----------------------------- RESULT STREAM HELPER METHODS ---------------------------//
   
   
   /**
    * Helper method for converting a query result into a result stream.
    * Generates the marginal corresponding to a given output instance.
    * @param output
    * @return
    */
   protected static Map<ITuple,Double> getMarginal(IFinalOutput output){
      
      Map<ITuple,Double>  marginal = new HashMap<ITuple,Double>();
      IResultDistribution dist     = output.getDistribution();
      
      // Make one tuple for each boolean value
      double      trueProb       = dist.getProbability(1);
      Attribute   truthAttribute = new Attribute(AttributeType.INT, 1);      
      Attribute[] attributeArr   = new Attribute[]{truthAttribute};
      ITuple      trueTuple      = new Tuple(attributeArr);
      
      marginal.put(trueTuple, trueProb);
      
      double falseProb  = 1.0 - trueProb; // dist.getProbability(0); 
      // TODO: The above is a hack because result streams don't have false prob - ask Julie!
      truthAttribute    = new Attribute(AttributeType.INT, 0);
      attributeArr      = new Attribute[]{truthAttribute};
      ITuple falseTuple = new Tuple(attributeArr);
      
      marginal.put(falseTuple, falseProb);

      return marginal;
   }
   
   
   /**
    * Helper method that takes a result stream marginal and negates it.
    * @param marginal
    * @return
    */
   protected static Map<ITuple,Double> getNegatedMarginal(Map<ITuple,Double> marginal){
      
      Map<ITuple,Double> negated = new HashMap<ITuple,Double>();
      
      for(ITuple tuple : marginal.keySet()){
         double negatedProb = 1.0 - marginal.get(tuple).doubleValue();
         
         negated.put(tuple, negatedProb);
      }
      return negated;
   }
}
