package edu.washington.cs.rfid.mstreams.test;

import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import markovianStream.ITimestep;
import operators.distributions.BooleanDistribution;
import operators.post.FinalOutput;
import operators.post.IFinalOutput;
import storageManager.schema.ITuple;
import edu.washington.cs.rfid.mstreams.data.ResultStreamCPTData;
import edu.washington.cs.rfid.mstreams.data.ResultStreamMarginalData;
import edu.washington.cs.rfid.mstreams.data.ScenicEventExtraData;
import edu.washington.cs.rfid.mstreams.data.db.BdbConnection;
import edu.washington.cs.rfid.mstreams.util.ResultStreamMerger;

/**
 * Test driver for testing the result stream merging functions.
 * @author evan
 */
public class TestResultStreamMerger {

	/**
	 * Main method tests the creation and merging of multiple lists.
	 */
	public static void main(String[] args){

		BdbConnection converter = new BdbConnection();
		int                   n = 3;

		// First create n lists with 2^n entries, one for each possible combination of 1s and 0s
		boolean[][] comboMatrix = ResultStreamMerger.getComboMatrix_FOR_TEST(n);
		int[]       bitMasks    = ResultStreamMerger.getBitMasks_FOR_TEST(comboMatrix);
		double[]    probPerList = TestResultStreamMerger.getProbPerList(bitMasks.length);

		boolean useSync = false;

		Vector<List<IFinalOutput>> lists = null;
		if(useSync){
			lists = TestResultStreamMerger.getSyncOutputLists(comboMatrix, bitMasks, probPerList);
			TestResultStreamMerger.printLists(lists, comboMatrix, bitMasks, probPerList);

			// Now print the marginals for this set of streams
			boolean printMarginals = false;

			if(printMarginals){
				TestResultStreamMerger.printMarginals(converter, lists, comboMatrix, bitMasks);
			}

			// Now print the timesteps for the merged set of lock-step streams
			boolean printTimesteps = true;

			if(printTimesteps){
				TestResultStreamMerger.printTimesteps(lists, comboMatrix, bitMasks);         
			}   
		} else {
			lists = TestResultStreamMerger.getAsyncOutputLists();
			TestResultStreamMerger.printLists(lists);

			// Now print the timesteps for the merged set of async streams
			boolean printTimesteps = true;

			if(printTimesteps){
				TestResultStreamMerger.printTimesteps(lists);         
			}  
		}
	}


	//--------------------- HELPERS -------------------------//

	/**
	 * Creates the array that dictates the probability of an event occurring at a given
	 * timestep in the sync lists.
	 * @param numCombos
	 * @return
	 */
	public static double[] getProbPerList(int numCombos){
		double[] probPerList = new double[numCombos];
		double   curProb     = 1.0;
		double   increment   = curProb/(double)numCombos;
		for(int i = 0; i < probPerList.length; i++){
			probPerList[i] = Math.pow(curProb, 0.33333333);
			curProb       -= increment;
		}
		return probPerList;
	}


	/**
	 * Makes a set of lists that together step through each possible combination of events 
	 * occurring/not occurring in the given comboMatrix (and reflected in the bitMasks).
	 * There is one combination per timestep.  The probability of an event occurring at a 
	 * given timestep is given by the probPerList array.
	 *
	 * @param comboMatrix
	 * @param bitMasks
	 * @param probPerList
	 * @return
	 */
	public static Vector<List<IFinalOutput>> getSyncOutputLists(boolean[][] comboMatrix,
			int[]       bitMasks   ,
			double[]    probPerList){

		Vector<List<IFinalOutput>> lists = new Vector<List<IFinalOutput>>();

		// Step through entries for each stream
		for(int j = 0; j < comboMatrix[0].length; j++){
			List<IFinalOutput> list = new ArrayList<IFinalOutput>();
			lists.add(list);

			for(int i = 0; i < comboMatrix.length; i++){

				BooleanDistribution dist = null;

				if(comboMatrix[i][j]){
					dist = new BooleanDistribution(probPerList[i]);
					dist.addToPTrue (probPerList[i]);
				} else {
					dist = new BooleanDistribution(1.0 - probPerList[i]);
					dist.addToPTrue (1.0 - probPerList[i]);
				}
				IFinalOutput output = new FinalOutput(dist,(long)i);
				output.setMetaData(new ScenicEventExtraData(i*1000L + 1L));
				list.add(output);
			}
		}      
		return lists;
	}


	/**
	 * Makes three lists of length 30 all with the Boolean event occurring with probability
	 * .96549 (cubes to about .9).  One list has the event occurring every other timestep,
	 * the second list has the event occurring every third timestep, and the last has the
	 * event occurring every fifth timestep.
	 * 
	 * @return
	 */
	public static Vector<List<IFinalOutput>> getAsyncOutputLists(){

		Vector<List<IFinalOutput>> lists = new Vector<List<IFinalOutput>>();

		double probOccurs  = 0.96549;
		int    numLists    = 3;
		int[]  listModulus = new int[numLists];

		listModulus[0] = 2;
		listModulus[1] = listModulus[0] + 1;

		for(int i = 2; i < listModulus.length; i++){
			listModulus[i] = listModulus[i-1] + 2;
		}

		for(int i = 0; i < numLists; i++){
			lists.add(new ArrayList<IFinalOutput>());
		}

		// Step through the 30 timesteps
		for(int i = 0; i < 31; i++){
			long timestamp    = (i+1L)*1000L + 1L;
			long endTimestamp = (i+2L)*1000L;

			for(int j = 0; j < numLists; j++){
				List<IFinalOutput>    list = lists.get(j);

				if(i % listModulus[j] == 0){ // Then this event occurs
					BooleanDistribution dist = new BooleanDistribution(probOccurs);

					IFinalOutput output = new FinalOutput(dist,(long)i);

					output.setMetaData(new ScenicEventExtraData(timestamp));
					list.add(output);
				}
				//            else {
				//               dist.addToPTrue (0.0);
				//               dist.addToPFalse(1.0);
				//            }            
			}  
		}
		return lists;
	}


	/**
	 * Prints the sync lists to be merged
	 * @param lists
	 * @param comboMatrix
	 * @param bitMasks
	 * @param probPerList
	 */
	public static void printLists(Vector<List<IFinalOutput>> lists      ,
			boolean[][]              comboMatrix,
			int[]                    bitMasks   ,
			double[]                 probPerList){
		DecimalFormat df = new DecimalFormat("#.###");

		System.out.println("About to send through sync streams: ");
		for(int i = 0; i < comboMatrix.length; i++){

			System.out.print(bitMasks[i] + ": "); // Print the probability = 1.0 combo

			for(int j = 0; j < comboMatrix[i].length; j++){
				int num = (comboMatrix[i][j])? 1 : 0;
				System.out.print(num + " ");
			}
			System.out.print("  ");

			// Print out the distribution for each stream entry
			for(int j = 0; j < comboMatrix[i].length; j++){
				IFinalOutput output = lists.get(j).get(i);

				System.out.print(" | 1 = " + df.format(output.getDistribution().getProbability(1)));
				System.out.print(", 0 = " + df.format(output.getDistribution().getProbability(0)));
			}

			System.out.println(" -- prob per list = " + df.format(probPerList[i]));
		}
	}


	/**
	 * 
	 * @param lists
	 */
	public static void printLists(Vector<List<IFinalOutput>> lists){

		DecimalFormat df = new DecimalFormat("#.###");

		int    numLists    = lists.size();
		int[]  listIndex   = new int[numLists];
		int[]  listModulus = new int[numLists];

		listModulus[0] = 2;
		listModulus[1] = listModulus[0] + 1;

		for(int i = 2; i < listModulus.length; i++){
			listModulus[i] = listModulus[i-1] + 2;
		}

		System.out.println("About to send through async streams: ");

		for(int i = 0; i < 30; i++){
			long timestamp    = (i+1L)*1000L + 1L;
			long endTimestamp = (i+2L)*1000L + 1L;

			System.out.print(timestamp + ": ");
			for(int j = 0; j < lists.size(); j++){
				if(i % listModulus[j] == 0){
					System.out.print("1 ");
				} else {
					System.out.print("- ");
				}
			}

			System.out.print("  ");
			for(int j = 0; j < lists.size(); j++){
				if(i % listModulus[j] == 0){
					IFinalOutput output = lists.get(j).get(listIndex[j]);
					listIndex[j]++;

					System.out.print(" | 1 = " + df.format(output.getDistribution().getProbability(1)));
					System.out.print(", 0 = "  + df.format(output.getDistribution().getProbability(0)));
				} else {
					System.out.print(" | --------- ");
				}
			}
			System.out.println("|");
		}      
	}



	/**
	 * Prints the marginals resulting from a merge
	 * 
	 * @param lists
	 * @param comboMatrix
	 * @param bitMasks
	 */
	public static void printMarginals(BdbConnection            bdb        ,
			Vector<List<IFinalOutput>> lists      ,
			boolean[][]              comboMatrix,
			int[]                    bitMasks){
		int[] curIndex = new int[lists.size()];
		int[] listLen  = new int[lists.size()];

		for(int i = 0; i < listLen.length; i++){
			listLen[i] = bitMasks.length;
		}
		double[] initialComboProbs = new double[comboMatrix.length];
		for(int i = 0; i < initialComboProbs.length; i++){
			initialComboProbs[i] = 1.0;
		}

		// Print the marginal in human-readable form
		System.out.println("\n------------------------\n");

		for(int i = 0; i < bitMasks.length; i++){
			Map<ITuple,Double> marginalMap = ResultStreamMerger.getMarginal_FOR_TEST(lists, 
					(i+1)*1000L, 
					curIndex, 
					listLen, 
					comboMatrix, 
					bitMasks, 
					initialComboProbs);

			ResultStreamMarginalData marginal = new ResultStreamMarginalData(i*1000L, marginalMap);
			System.out.println("Marginal is: " + marginal.toPrettyString());
		}
	}


	/**
	 * Prints the timesteps resulting from the merge
	 * 
	 * @param lists
	 * @param comboMatrix
	 * @param bitMasks
	 */
	public static void printTimesteps(Vector<List<IFinalOutput>> lists      ,
			boolean[][]              comboMatrix,
			int[]                    bitMasks){

		int[] curIndex = new int[lists.size()];
		int[] listLen  = new int[lists.size()];

		for(int i = 0; i < listLen.length; i++){
			listLen[i] = bitMasks.length;
		}
		double[] initialComboProbs = new double[comboMatrix.length];
		for(int i = 0; i < initialComboProbs.length; i++){
			initialComboProbs[i] = 1.0;
		}


		// Print the timesteps in human-readable form
		System.out.println("\n------------------------\n");

		List<ITimestep> timesteps = ResultStreamMerger.mergeLists_FOR_TEST(lists);
		int i = 0;
		for(ITimestep ts : timesteps){

			Map<ITuple,Double>             marginalMap = ts.getMarginal();
			Map<ITuple,Map<ITuple,Double>> cptMap      = ts.getCPT().getValues();
			ScenicEventExtraData          extra       = (ScenicEventExtraData)ts.getExtraData();

			ResultStreamMarginalData marginal = new ResultStreamMarginalData(extra.timestamp,marginalMap);
			ResultStreamCPTData      cpt      = new ResultStreamCPTData(extra.timestamp, extra.timestamp + 1000L, cptMap);

			System.out.println("Mask: " + bitMasks[i]);
			System.out.println(marginal.toPrettyString());
			System.out.println(cpt.toPrettyString());
			System.out.println();
			i++;
		}
	}


	public static void printTimesteps(Vector<List<IFinalOutput>> lists){

		DecimalFormat df = new DecimalFormat("#.###");

		int    numLists    = lists.size();
		int[]  listModulus = new int[numLists];

		listModulus[0] = 2;
		listModulus[1] = listModulus[0] + 1;

		for(int i = 2; i < listModulus.length; i++){
			listModulus[i] = listModulus[i-1] + 2;
		}

		// Print the timesteps in human-readable form
		System.out.println("\n------------------------\n");

		List<ITimestep> timesteps = ResultStreamMerger.mergeLists_FOR_TEST(lists);
		int i = 0;
		for(ITimestep ts : timesteps){

			Map<ITuple,Double>             marginalMap = ts.getMarginal();
			Map<ITuple,Map<ITuple,Double>> cptMap      = ts.getCPT().getValues();
			ScenicEventExtraData          extra       = (ScenicEventExtraData)ts.getExtraData();

			ResultStreamMarginalData marginal = new ResultStreamMarginalData(extra.timestamp,marginalMap);
			ResultStreamCPTData      cpt      = new ResultStreamCPTData(extra.timestamp, extra.timestamp + 1000L, cptMap);

			String modulusStr = " ";

			for(int j = 0; j < listModulus.length; j++){
				modulusStr += (i % listModulus[j] == 0)? "1 " : "0 ";
			}

			System.out.println("Lists: " + modulusStr);
			System.out.println(marginal.toPrettyString());
			System.out.println(cpt.toPrettyString());
			System.out.println();
			i++;
		}
	}
}
