package edu.washington.cs.rfid.mstreams.tools;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Vector;

import markovianStream.ITimestep;
import edu.washington.cs.rfid.mstreams.data.ParticleFilterStream;
import edu.washington.cs.rfid.mstreams.data.ParticleFilterStreamCPTData;
import edu.washington.cs.rfid.mstreams.data.ParticleFilterStreamMarginalData;
import edu.washington.cs.rfid.mstreams.data.StreamKey;
import edu.washington.cs.rfid.mstreams.data.ParticleFilterStream.MarginalCPTPair;
import edu.washington.cs.rfid.mstreams.data.db.BdbConnection;
import edu.washington.cs.rfid.mstreams.data.db.LaharConnection;
import edu.washington.cs.rfid.mstreams.data.db.StreamKeyManager;
import edu.washington.cs.rfid.mstreams.event.AtEvent;
import edu.washington.cs.rfid.mstreams.exception.MStreamSyntaxException;
import edu.washington.cs.rfid.mstreams.schema.ParticleFilterStreamSchema;
import edu.washington.cs.rfid.mstreams.util.ParticleFilterStreamParser;
import edu.washington.cs.rfid.mstreams.util.StreamConverter;

public class LaharDbLoader {
   
   public static String DEFAULT_BDB_DIR      = "/data/bdbRfidStreams";
   public static String DEFAULT_LAHAR_DB_DIR = "/data/db_loc/RfidDB";
   public static String DEFAULT_TYPE         = "MARKOVIAN";
   
   //--//
   
   protected String          bdbDir;
   protected String          laharDbDir;
   
   protected BdbConnection   bdbManager;
   protected LaharConnection laharManager;
   
   //--//
   
   /**
    * Creates a new LaharDbLoader.
    * 
    * @param pfFileList
    * @param bdbDir
    * @param laharDbDir
    */
   public LaharDbLoader(String bdbDir, String laharDbDir) {
      this.bdbDir       = bdbDir;
      this.laharDbDir   = laharDbDir;
      
      this.bdbManager   = new BdbConnection(this.bdbDir);
      this.laharManager = new LaharConnection(this.bdbDir,this.laharDbDir);
   }
   
   
   /**
    * Loads a particle filter file into Lahar.
    * 
    * @param file
    * @return
    */
   public boolean loadParticleFilterFile(String file) {
      boolean success = false;
      
      try {
    	   ParticleFilterStream pfStream   = ParticleFilterStreamParser.parseSmoothedPF(file);
         StreamKeyManager     keyManager = new StreamKeyManager();
         String               streamDesc = pfStream.getStreamDescription();
         AtEvent              atEvent    = new AtEvent(streamDesc);
         StreamKey            sk         = atEvent.getStreamKey(keyManager);
         
         // 1) Convert the particle filter file into an MStream
         List<ITimestep>            stream = StreamConverter.convertToMStream(pfStream);
         ParticleFilterStreamSchema schema   = new ParticleFilterStreamSchema();
         
         // 2) Load the timesteps into BDB if they have not already been loaded
         if(this.bdbManager.storeMStream(stream,sk.getKey(),schema,false)){
            
            // 3) Now load the stream into Lahar
            this.laharManager.moveStream(sk, schema);
            
            success = true;
         }
      } catch(IOException e){
         e.printStackTrace();
      } catch(MStreamSyntaxException e){
         e.printStackTrace();
      } catch(SQLException e){
         e.printStackTrace();
      } catch(ClassNotFoundException e){
         e.printStackTrace();
      }
      return success;
   }
   
   
   /**
    * Compares a stream that has been loaded from a file to a stream that has been loaded
    * into BDB.
    * 
    * @param file
    * @param bdbDir
    * @param streamKey
    * @return
    */
   public boolean compareFileToBdb(String file, String streamKey) throws IOException, 
                                                                         MStreamSyntaxException {
      boolean result = false;
      
      // First read the file
      ParticleFilterStream stream       = ParticleFilterStreamParser.parseSmoothedPF(file);
      List<ITimestep>      loadedTsList = this.bdbManager.loadMStream(streamKey);
      ParticleFilterStream bdbStream    = this.convertFromMStream(loadedTsList);
      
      this.compareSmoothedFiles(stream, bdbStream);
      result = this.compareFilesExact(stream, bdbStream);
      
      return result;
   }
   
   
   /**
    * Converts a list of timesteps to an MStream.
    * 
    * @param timesteps
    * @return
    */
   protected ParticleFilterStream convertFromMStream(List<ITimestep> timesteps){
      ParticleFilterStream file   = ParticleFilterStreamParser.parseTimestepList(timesteps);      
      return file;
   }
   
   
   /**
    * Compares two particle filter stream files and prints out the results of the comparison
    * @param file1
    * @param file2
    */
   protected void compareSmoothedFiles(ParticleFilterStream file1, ParticleFilterStream file2){
      
      System.out.println("File 1 length: " + file1.sequence.size() + "  File 2 length: " + file2.sequence.size() + "\n");
      
      for(int i = 0; i < file1.sequence.size() && i < file2.sequence.size(); i++){
         MarginalCPTPair pair1 = file1.sequence.get(i);
         MarginalCPTPair pair2 = file2.sequence.get(i);
         
         ParticleFilterStreamMarginalData    marginal1 = pair1.marginal;
         ParticleFilterStreamMarginalData    marginal2 = pair2.marginal;
         
         ParticleFilterStreamCPTData         cpt1      = pair1.cpt;
         ParticleFilterStreamCPTData         cpt2      = pair2.cpt;
         
         double mProb1 = marginal1.getTotalProb();
         double mProb2 = marginal2.getTotalProb();
         
         System.out.println("Marginal probs = " + mProb1 + "  " + mProb2);
         
         Vector<Double> c1Probs = cpt1.getTotalProbs();
         Vector<Double> c2Probs = cpt2.getTotalProbs();
         
         System.out.println("");
         
         Collections.sort(c1Probs);
         Collections.sort(c2Probs);
         
         System.out.println("CPT 1 length = " + c1Probs.size() + "  CPT 2 length = " + c2Probs.size());
         
         for(int j = 0; j < c1Probs.size() && j < c2Probs.size(); j++){
            System.out.println("CPT probs = " + c1Probs.get(j) + "  " + c2Probs.get(j));
         }
         
         System.out.println("----------");
      }
   }
   
   
   /**
    * Compares two particle filter stream files and returns the result.
    * @param file1
    * @param file2
    * @return
    */
   protected boolean compareFilesExact(ParticleFilterStream file1, ParticleFilterStream file2) {
      
      boolean result = file1.equals(file2);
      return result;
   }
   
   
   //--//
   
   
   /**
    * Prints usage information.
    */
   public static void usage(){
      String usage = "Usage: LaharDbLoader  FILELIST_FILE  [BDB_DIR  LAHAR_DB_DIR]\n"          +
                     "Loads the particle filter files in FILELIST_FILE into a BDB database " +
                     "at BDB_DIR \n";
      System.out.println(usage);
   }
   
   
   /**
    * Runs a tool for loading particle filter streams into Lahar.
    * @param args
    */
   public static void main(String[] args) {
      
      String fileListFilename = null;
      String bdbDirFilename   = null;
      String laharDirFilename = null;
      
      if(args.length != 1 && args.length != 3) {
         LaharDbLoader.usage();
         System.exit(1);
      } else {
         fileListFilename = args[0];
         
         if(args.length == 3){
            bdbDirFilename   = args[1];
            laharDirFilename = args[2];
         } else {
            bdbDirFilename   = LaharDbLoader.DEFAULT_BDB_DIR;
            laharDirFilename = LaharDbLoader.DEFAULT_LAHAR_DB_DIR;
         }
      }
      
      // Load the files in the file list file
      List<String>   pfFileList = new ArrayList<String>();
      BufferedReader reader     = null;
      boolean        success    = false;
      try {
         reader      = new BufferedReader(new FileReader(fileListFilename));
         String line = null;
         
         while(null != (line = reader.readLine())) {
            pfFileList.add(line);
         }
         success = true;
      } catch(IOException e){
         e.printStackTrace();
      } finally {
         try {
            if(null != reader) reader.close();
         } catch(IOException e){
            e.printStackTrace();
         }
         if(!success){
            System.exit(1);
         }
      }
      
      // Now create the loader and load the files
      LaharDbLoader loader = new LaharDbLoader(bdbDirFilename, laharDirFilename);
      
      for(String file : pfFileList){
    	  try {
    	     loader.loadParticleFilterFile(file);
    	     System.out.println("Loaded file: " + file);
    	  } catch (Exception ex) {
    		  ex.printStackTrace();
    	  }
      }
   }

}
