package edu.washington.cs.rfid.mstreams.util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import markovianStream.ITimestep;
import storageManager.schema.Attribute;
import storageManager.schema.ITuple;
import storageManager.schema.Tuple;
import storageManager.schema.Attribute.AttributeType;
import edu.washington.cs.rfid.mstreams.data.ScenicEventExtraData;
import edu.washington.cs.rfid.mstreams.data.db.NodeManager;

public class ParticleFilterStreamMerger {
   
   protected static long MS_PER_SECOND   = 1000L;
   
   protected static int  MERGE_TYPE_WITH = 1;
   protected static int  MERGE_TYPE_NEAR = 2;
   
   //--//
   
   
   /**
    * Creates a result stream for the with event on the two given particle filter
    * streams.  Steps through the two streams timestep by timestep and creates the
    * boolean "with" stream as output.  
    * 
    * @param streams
    * @return
    */
   public static List<ITimestep> createWithStream(List<List<ITimestep>> streams) {
      return ParticleFilterStreamMerger.createMergedStream(streams, 0.0, ParticleFilterStreamMerger.MERGE_TYPE_WITH);
   }
   
   
   /**
    * Creates a result stream for the with event on the two given particle filter
    * streams.  Steps through the two streams timestep by timestep and creates the
    * boolean "with" stream as output.  
    * 
    * @param streams
    * @param thresh
    * @return
    */
   public static List<ITimestep> createNearStream(List<List<ITimestep>> streams, double thresh) {
      return ParticleFilterStreamMerger.createMergedStream(streams, thresh, ParticleFilterStreamMerger.MERGE_TYPE_NEAR);
   }
   
   
   //--//
   
   
   /**
    * Creates a result stream for a given type of event event on the two given particle filter
    * streams.  Steps through the two streams timestep by timestep and creates the
    * boolean stream as output.  
    * 
    * @param streams
    * @param thresh
    * @param mergeType
    * @return
    */
   protected static List<ITimestep> createMergedStream(List<List<ITimestep>> streams, 
                                                       double                thresh ,
                                                       int                   mergeType) {
      
      List<ITimestep> mergedStream = new ArrayList<ITimestep>();
      
      // Find the boundary times for the entire set of lists
      long                 startTime = -1;
      long                 endTime   = -1;
      ITimestep            output    = null;
      ScenicEventExtraData data      = null;
      int[]                streamLen = new int[streams.size()];
      
      // Get the earliest start and the latest end
      for(int i = 0; i < streams.size(); i++){
         List<ITimestep> stream = streams.get(i);
         streamLen[i]           = stream.size();
         if(0 < streamLen[i]){
            output    = stream.get(0);
            data      = (ScenicEventExtraData) output.getExtraData();
            startTime = (startTime > data.timestamp || startTime < 0)? data.timestamp : startTime;
            output    = stream.get(stream.size()-1);
            data      = (ScenicEventExtraData) output.getExtraData();
            endTime   = (endTime < data.timestamp || endTime < 0)? data.timestamp : endTime;
         }
      }
      
      
      // Now tick through the lists, timestep by timestep
      int         tsIndex        = 0;
      long        curTime        = startTime;
      long        timestep       = 1L * ParticleFilterStreamMerger.MS_PER_SECOND;
      long        curEndInterval = curTime + timestep;
      int[]       curIndex       = new int[streams.size()];
      
      endTime                   += timestep;    // For end boundary condition
      

      Map<ITuple, Double> lastMarginalMap = null;
      
      if(mergeType == ParticleFilterStreamMerger.MERGE_TYPE_WITH){
         lastMarginalMap = ParticleFilterStreamMerger.createWithMarginal(streams       , 
                                                   curEndInterval,
                                                   curIndex      ,
                                                   streamLen);
      } else if(mergeType == ParticleFilterStreamMerger.MERGE_TYPE_NEAR){
         lastMarginalMap = ParticleFilterStreamMerger.createNearMarginal(streams       , 
                                                   curEndInterval,
                                                   curIndex      ,
                                                   streamLen     ,
                                                   thresh);
      }
      
      
      curTime         = curEndInterval;
      curEndInterval += timestep;
      
      while(curEndInterval <= endTime){
         
         Map<ITuple, Double> curMarginalMap = null;
         
         if(mergeType == ParticleFilterStreamMerger.MERGE_TYPE_WITH){
            curMarginalMap = ParticleFilterStreamMerger.createWithMarginal(streams       , 
                                                     curEndInterval,
                                                     curIndex      ,
                                                     streamLen);
         } else if(mergeType == ParticleFilterStreamMerger.MERGE_TYPE_NEAR){
            curMarginalMap = ParticleFilterStreamMerger.createNearMarginal(streams       , 
                                                     curEndInterval,
                                                     curIndex      ,
                                                     streamLen     ,
                                                     thresh);
         }
         
         Map<ITuple,Map<ITuple,Double>> cptMap = MStreamCreatorUtil.getDefaultCPT(lastMarginalMap, curMarginalMap);
         
         ITimestep ts = MStreamCreatorUtil.getITimestep(tsIndex, curTime, lastMarginalMap, cptMap);
         mergedStream.add(ts);
         
         tsIndex++;
         curTime         = curEndInterval;
         curEndInterval += timestep;
         lastMarginalMap = curMarginalMap;
      }
      Map<ITuple,Map<ITuple,Double>> cptMap = MStreamCreatorUtil.getDefaultCPT(lastMarginalMap);
      ITimestep                      ts     = MStreamCreatorUtil.getITimestep(tsIndex, curTime, lastMarginalMap, cptMap);
      mergedStream.add(ts);
      
      return mergedStream;
   }
   
   
   /**
    * Generates the marginal for a With stream corresponding to a given output instance.
    * 
    * @param streams
    * @param curEndInterval
    * @param curIndex
    * @param streamLen
    * @return
    */
   protected static Map<ITuple, Double> createWithMarginal(final List<List<ITimestep>> streams       , 
                                                           final long                  curEndInterval,
                                                           int[]                       curIndex      ,
                                                           final int[]                 streamLen) {
      
      Map<ITuple, Double> marginalMap = new HashMap<ITuple,Double>();
      
      
      // 1) Get the marginal for the current timestep from each list
      List<Map<ITuple, Double>> marginals = new ArrayList<Map<ITuple, Double>>();
      
      for(int i = 0; i < streams.size(); i++){
         if(curIndex[i] < streamLen[i]){
            List<ITimestep>      stream   = streams.get(i);
            ITimestep            timestep = stream.get(curIndex[i]);
            ScenicEventExtraData data     = (ScenicEventExtraData)timestep.getExtraData();
            
            if(data.timestamp < curEndInterval){
               marginals.add(timestep.getMarginal());
               curIndex[i]++;
            }
         }
      }
      
      // If there is no marginal for the timestep in some list then the probability for "With" is 0.0
      double withProb = 0.0;
      
      if(0 > marginals.size() && marginals.size() == streams.size()){
         
         // 2) Now compute the probability that the tags for each stream are at the same node
         Map<ITuple, Double> firstMarginal = marginals.get(0);

         for(ITuple tuple : firstMarginal.keySet()) {
            double    probNode = 1.0;
            Attribute nodeAttr = tuple.getAttribute(1);
            long      node     = ((Long)nodeAttr.getValue()).longValue();
            
            for(int i = 1; i < marginals.size(); i++){
               Map<ITuple, Double> marginal           = marginals.get(i);
               Double              probNodeInMarginal = marginal.get(node);
               
               if(null != probNodeInMarginal){
                  probNode *= probNodeInMarginal.doubleValue();
               } else {
                  probNode = 0.0;
                  break;
               }
            }
            withProb += probNode;
         }
      }
      
      
      // 3) Create the "With" marginal map using withProb
      
      // Put 0 = 1 - probability true
      Attribute   truthAttribute = new Attribute(AttributeType.INT, 0);
      Attribute[] attributeArr   = new Attribute[1];
      attributeArr[0]            = truthAttribute;
      ITuple      tuple          = new Tuple(attributeArr);
      
      marginalMap.put(tuple, 1.0 - withProb);
      
      // Put 1 = probability true
      truthAttribute  = new Attribute(AttributeType.INT, 1);
      attributeArr    = new Attribute[1];
      attributeArr[0] = truthAttribute;
      tuple           = new Tuple(attributeArr);
      
      marginalMap.put(tuple, withProb);
      
      return marginalMap;
   }
   
   
   /**
    * Generates the marginal for a near stream corresponding to a given output instance.
    * 
    * @param streams
    * @param curEndInterval
    * @param curIndex
    * @param streamLen
    * @param thresh
    * @return
    */
   protected static Map<ITuple, Double> createNearMarginal(final List<List<ITimestep>> streams       , 
                                                           final long                  curEndInterval,
                                                           int[]                       curIndex      ,
                                                           final int[]                 streamLen     ,
                                                           double                      thresh) {
      
      Map<ITuple, Double> marginalMap = new HashMap<ITuple,Double>();
      
      
      // 1) Get the marginal for the current timestep from each list
      List<Map<ITuple, Double>> marginals = new ArrayList<Map<ITuple, Double>>();
      
      for(int i = 0; i < streams.size(); i++){
         if(curIndex[i] < streamLen[i]){
            List<ITimestep>      stream   = streams.get(i);
            ITimestep            timestep = stream.get(curIndex[i]);
            ScenicEventExtraData data     = (ScenicEventExtraData)timestep.getExtraData();
            
            if(data.timestamp < curEndInterval){
               marginals.add(timestep.getMarginal());
               curIndex[i]++;
            }
         }
      }
      
      // If there is no marginal for the timestep in some list then the probability for "Near" is 0.0
      double nearProb = 0.0;
      
      if(0 > marginals.size() && marginals.size() == streams.size()){
         
         // Prep: Get a list of nodes for each marginal and a corresponding list of probabilities for nodes
         List<List<Long>>   marginalNodeLists = new ArrayList<List<Long>>  ();
         List<List<Double>> marginalProbLists = new ArrayList<List<Double>>();
         
         for(Map<ITuple, Double> marginal : marginals){
            ArrayList<Long>   nodes = new ArrayList<Long>  ();
            ArrayList<Double> probs = new ArrayList<Double>();
            
            for(ITuple tuple : marginal.keySet()){
               Attribute nodeAttr = tuple.getAttribute(1);
               Long      node     = (Long)nodeAttr.getValue();
               Double    prob     = marginal.get(tuple);
               
               nodes.add(node);
               probs.add(prob);
            }
            marginalNodeLists.add(nodes);
            marginalProbLists.add(probs);
         }
         
         // 2) Now compute the probability that the tags for each stream are near each other
         nearProb = ParticleFilterStreamMerger.getNearProb(new Stack<Long>(), 
                                     1.0              , 
                                     marginalNodeLists, 
                                     marginalProbLists, 
                                     0                ,
                                     thresh);
      }
      
      // 3) Create the "Near" marginal map using nearProb
      
      // Put 0 = 1 - probability true
      Attribute   truthAttribute = new Attribute(AttributeType.INT, 0);
      Attribute[] attributeArr   = new Attribute[1];
      attributeArr[0]            = truthAttribute;
      ITuple      tuple          = new Tuple(attributeArr);
      
      marginalMap.put(tuple, 1.0 - nearProb);
      
      // Put 1 = probability true
      truthAttribute  = new Attribute(AttributeType.INT, 1);
      attributeArr    = new Attribute[1];
      attributeArr[0] = truthAttribute;
      tuple           = new Tuple(attributeArr);
      
      marginalMap.put(tuple, nearProb);
      
      return marginalMap;
   }
   
   
   /**
    * A recursive method to compute the probability that the given node is near any combination 
    * of nodes in the given marginal node lists (starting at the specified index).
    *   
    * @param nodeStack The list of nodes all other nodes must be near to
    * @param nodeStackProb The probability that nodes in the stack are near to each other
    * @param marginalNodeLists The list of node lists from all marginals 
    * @param marginalProbLists The list of node probabilities from all other marginals
    * @param listStartIndex The index at which to start in the lists
    * @param thresh The threshold in meters for "nearness"
    * @return
    */
   // TODO: Write an iterative version of this method
   protected static double getNearProb(Stack<Long>        nodeStack        ,
                                       double             nodeStackProb    ,
                                       List<List<Long>>   marginalNodeLists, 
                                       List<List<Double>> marginalProbLists,
                                       int                listStartIndex   , 
                                       double             thresh) {
      double      prob        = 0.0;
      NodeManager nodeManager = NodeManager.getInstance();
      
      // Process the first list separately
      List<Long>   nodes = marginalNodeLists.get(listStartIndex);
      List<Double> probs = marginalProbLists.get(listStartIndex);
      
      for(int i = 0; i < nodes.size(); i++){
         
         // First make sure the node is near to all other nodes in the stack
         long    otherNode = nodes.get(i);
         double  otherProb = probs.get(i);
         boolean near      = true;
         
         for(Long node : nodeStack){
            if(thresh < nodeManager.getShortestPathLength(node,otherNode)){
               near = false;
               break;
            }
         }
         
         
         if(near){
            
            // If this is the last list, then add the probability that all nodes are near each other 
            if(listStartIndex == marginalNodeLists.size() - 1) {
               
               prob += nodeStackProb * otherProb;
               
            } else {
            // If there are more lists, then then recurse through the rest of the lists
               
               nodeStack.push(otherNode);
               
               prob += ParticleFilterStreamMerger.getNearProb(nodeStack                , 
                                        nodeStackProb * otherProb, 
                                        marginalNodeLists        , 
                                        marginalProbLists        , 
                                        listStartIndex + 1       ,
                                        thresh);
               
               nodeStack.pop();
            }
         }
      }
      return prob;
   }
}
