package edu.washington.cs.rfid.mstreams.util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import markovianStream.ITimestep;
import operators.post.IFinalOutput;
import storageManager.schema.Attribute;
import storageManager.schema.ITuple;
import storageManager.schema.Tuple;
import storageManager.schema.Attribute.AttributeType;
import api.queryResults.GroupResult;
import api.queryResults.QueryResult;
import edu.washington.cs.rfid.mstreams.data.ScenicEventExtraData;

/**
 * Class that takes result streams and merges them into a single stream.
 * 
 * @author evan
 */
public class ResultStreamMerger {
    
   protected static long MS_PER_SECOND = 1000L;
   
   //--//
   
   /**
    * Given a set of result streams, merges them into a list of timesteps 
    * by AND-ing them together; then returns the list of timesteps.
    * 
    * @param result1
    * @param result2
    * @return
    */
   public static List<ITimestep> mergeQueryResults(List<QueryResult> results) {
      
      List<ITimestep> tsList = new ArrayList<ITimestep>();
      
      // Merge the streams and convert them into BDB format:
      // 1) Create a list of all output lists
      List<List<IFinalOutput>> listSet = new Vector<List<IFinalOutput>>();
      
      for(QueryResult result : results){
         List<GroupResult> grpResults = result.getResults();
         for(GroupResult grpResult : grpResults){
            List<IFinalOutput> outputs = grpResult.getResults();
            listSet.add(outputs);
         }
      }
      
      // 2) Merge all output lists into one single list of timesteps
      tsList = ResultStreamMerger.mergeResultStreams(listSet);
      
      return tsList;
   }
   
   
   
   //--------------------- HELPERS -------------------------//
   
   
   /**
    * Given a list of result streams, trims the lists to all-inclusive start and end timestamps.
    * @param streams
    */
   protected static void trimResultStreams(List<List<IFinalOutput>> streams){
      
      long                 maxStartTime = 0;
      long                 minEndTime   = 0;
      IFinalOutput           output       = null;
      ScenicEventExtraData data         = null;
      
      for(List<IFinalOutput> stream : streams){
         if(stream.size() > 0){
            output       = stream.get(0);
            data         = (ScenicEventExtraData) output.getMetaData();
            maxStartTime = (maxStartTime > data.timestamp)? maxStartTime : data.timestamp;
            output       = stream.get(stream.size()-1);
            data         = (ScenicEventExtraData) output.getMetaData();
            minEndTime   = (minEndTime < data.timestamp)? minEndTime : data.timestamp;
         }
      }
      
      // Now trim anything outside the interval
      for(List<IFinalOutput> list : streams){
         while(0 < list.size()){
            if(((ScenicEventExtraData)list.get(0).getMetaData()).timestamp < maxStartTime){
               list.remove(0);
            } else {
               break;
            }
         }
         while(0 < list.size()){
            if(((ScenicEventExtraData)list.get(list.size()-1).getMetaData()).timestamp > minEndTime){
               list.remove(list.size()-1);
            } else {
               break;
            }
         }
      }
   }
   
   
   /**
    * Given a trimmed list of lists, takes the earliest start time and uses it as the start
    * time for the trace.  Then identifies the earliest start time and uses it as a starting point 
    * for a global trace that all traces will be merged in to.  Ticks through each one second 
    * timestep, making each second a new sequence ID.  Merges streams by multiplying them together 
    * and getting the bit mask formed by their values at each timestep.  Pads "empty" sections
    * of the trace/list with blank, 0 probability tuples.
    * 
    * @param streams
    */
   protected static List<ITimestep> mergeResultStreams(List<List<IFinalOutput>> streams){
      
      List<ITimestep> timesteps = new ArrayList<ITimestep>();
      
      // Find the boundary times for the entire set of lists
      long                 startTime   = -1;
      long                 endTime     = -1;
      IFinalOutput           output      = null;
      ScenicEventExtraData data        = null;
      int[]                listLen     = new int[streams.size()];
      
      // Get the earliest start and the latest end
      for(int i = 0; i < streams.size(); i++){
         List<IFinalOutput> list = streams.get(i);
         listLen[i]            = list.size();
         if(0 < listLen[i]){
            output    = list.get(0);
            data      = (ScenicEventExtraData) output.getMetaData();
            startTime = (startTime > data.timestamp || startTime < 0)? data.timestamp : startTime;
            output    = list.get(list.size()-1);
            data      = (ScenicEventExtraData) output.getMetaData();
            endTime   = (endTime < data.timestamp || endTime < 0)? data.timestamp : endTime;
         }
      }
      
      
      // Now tick through the lists, timestep by timestep
      int         tsIndex        = 0;
      long        curTime        = startTime;
      long        timestep       = 1L * ResultStreamMerger.MS_PER_SECOND;
      long        curEndInterval = curTime + timestep;
      int[]       curIndex       = new int[streams.size()];
      
      boolean[][] comboMatrix    = ResultStreamMerger.getComboMatrix(streams.size());
      int[]       bitMasks       = ResultStreamMerger.getBitMasks(comboMatrix);
      
      endTime                   += timestep;    // For end boundary condition
      
      // Probability of each possible combination initialized to 1.0 
      double[] initialComboProbs = new double[comboMatrix.length];
      for(int i = 0; i < initialComboProbs.length; i++){
         initialComboProbs[i] = 1.0;
      }
      
      Map<ITuple, Double> lastMarginalMap = ResultStreamMerger.getMarginal(streams         , 
                                                             curEndInterval, 
                                                             curIndex      , 
                                                             listLen       ,
                                                             comboMatrix   , 
                                                             bitMasks      , 
                                                             initialComboProbs);
      
      curTime         = curEndInterval;
      curEndInterval += timestep;
      
      while(curEndInterval <= endTime){
         
         Map<ITuple, Double> curMarginalMap = ResultStreamMerger.getMarginal(streams         , 
                                                               curEndInterval, 
                                                               curIndex      , 
                                                               listLen       ,
                                                               comboMatrix   , 
                                                               bitMasks      , 
                                                               initialComboProbs);
         
         Map<ITuple,Map<ITuple,Double>> cptMap = MStreamCreatorUtil.getDefaultCPT(lastMarginalMap, curMarginalMap);
         
         ITimestep ts = MStreamCreatorUtil.getITimestep(tsIndex, curTime, lastMarginalMap, cptMap);
         timesteps.add(ts);
         
         tsIndex++;
         curTime         = curEndInterval;
         curEndInterval += timestep;
         lastMarginalMap = curMarginalMap;
      }
      Map<ITuple,Map<ITuple,Double>> cptMap = MStreamCreatorUtil.getDefaultCPT(lastMarginalMap);
      ITimestep                      ts     = MStreamCreatorUtil.getITimestep(tsIndex, curTime, lastMarginalMap, cptMap);
      timesteps.add(ts);
      
      return timesteps;
   }
   
   
   /**
    * Helper method for converting a query result into a result stream.
    * Generates the marginal corresponding to a given output instance.
    * 
    * @param lists
    * @param curEndInterval
    * @param curIndex
    * @param listLen
    * @param comboMatrix
    * @param bitMasks
    * @param initialComboProbs
    * @return
    */
   protected static Map<ITuple, Double> getMarginal(final List<List<IFinalOutput>>   lists         , 
                                             final long                     curEndInterval,
                                             int[]                          curIndex      ,
                                             final int[]                    listLen       ,
                                             final boolean[][]              comboMatrix   ,
                                             final int[]                    bitMasks      ,
                                             final double[]                 initialComboProbs){
      
      Map<ITuple, Double>  marginalMap = null;
      double[]             comboProb   = initialComboProbs.clone();
      IFinalOutput           output      = null;
      ScenicEventExtraData data        = null;
      
      // For each list, iterate through the list's column in the combination matrix
      // and multiply the probability in the comboProb array
      for(int i = 0; i < lists.size(); i++){
         List<IFinalOutput> list = lists.get(i);
         
         double probTrue  = 0.0;
         double probFalse = 1.0;
         
         // We make the assumption that all lists advance at least <timestep> ms between elements
         // Otherwise we have to step through and merge possibly multiple timesteps per interval
         // and the semantics are unclear.
         if(curIndex[i] < listLen[i]){
            output = list.get(curIndex[i]);
            data   = (ScenicEventExtraData) output.getMetaData();
            
            if(data.timestamp < curEndInterval){ // Then include in merge 
               probTrue     = output.getDistribution().getProbability(1);
               probFalse    = 1.0 - probTrue; //output.getDistribution().getProbability(0);
               curIndex[i]++;
            }
         }
        
         // TODO: possible optimization when multiplying by 0... cut it short
         // Now iterate through the list's column in the combination matrix
         // and multiply the probability in comboProb by the appropriate probability
         for(int j = 0; j < comboMatrix.length; j++){
            if(comboMatrix[j][i]){
               comboProb[j] *= probTrue;
            } else {
               comboProb[j] *= probFalse;
            }
         }
      }
      
      // Now create the marginal from the bitMasks and the probabilities
      marginalMap = new HashMap<ITuple, Double>();
      
      for(int i = 0; i < bitMasks.length; i++){
         Attribute   maskAttribute = new Attribute(AttributeType.INT, bitMasks[i]);
         Attribute[] attributeArr  = new Attribute[1];
         attributeArr[0]           = maskAttribute;
         ITuple      tuple         = new Tuple(attributeArr);
         
         marginalMap.put(tuple, comboProb[i]);
      }
      return marginalMap;
   }
   
   
   /**
    * Gets a matrix representing every possible combination of streams for a given number of streams.
    * @param numStreams
    * @return
    */
   protected static boolean[][] getComboMatrix(int numStreams){
      int         combinations = (int)Math.pow(2.0, numStreams);
      int[]       modulus      = new int[numStreams];
      boolean[][] comboMatrix  = new boolean[combinations][numStreams]; // rows, cols
      
      for(int i = 0; i < numStreams; i++){
         modulus[i] = (0 == i)? 1 : modulus[i-1] + modulus[i-1];
      }
      
      boolean val = true;
      for(int j = 0; j < numStreams; j++){
         for(int i = 0; i < combinations; i++){
            int switchVal     = i % modulus[j];
            val               = (i == 0 || switchVal == 0)? !val : val;
            comboMatrix[i][(numStreams - 1 - j)] = val;
         }
      }
      return comboMatrix;
   }
   
   
   /**
    * Creates an array of bit masks, one for each row of the given combo matrix.
    * @precond the comboMatrix is not null
    * @param comboMatrix
    * @return
    */
   protected static int[] getBitMasks(boolean[][] comboMatrix){
      int[]     masks     = new int[comboMatrix.length];
      boolean[] maskArray = new boolean[comboMatrix[0].length];
      
      for(int i = 0; i < comboMatrix.length; i++){
         for(int j = 0; j < comboMatrix[i].length; j++){
            maskArray[j] = comboMatrix[i][j];
         }
         masks[i] = ResultStreamMerger.convertToBitmask(maskArray);
      }
      return masks;
   }
   
   
   /**
    * Converts an array of boolean flags into an integer bitmask.
    * @param flags
    * @return
    */
   protected static int convertToBitmask(boolean[] flags){
      int mask          = 0x0000;
      int msbShiftIndex = flags.length-1;
      
      for(int i = 0; i  < flags.length; i++){
         mask = (flags[i])? (mask | (1<<(msbShiftIndex-i))) : (mask | (0<<(msbShiftIndex-i)));
      }
      return mask;
   }
   
   
   //------------------------------- FOR TEST ONLY ---------------------------------//
   
   public static List<ITimestep> mergeLists_FOR_TEST(List<List<IFinalOutput>> lists){
      return ResultStreamMerger.mergeResultStreams(lists);
   }
   
   public static boolean[][] getComboMatrix_FOR_TEST(int numStreams){
      return ResultStreamMerger.getComboMatrix(numStreams);
   }
   
   public static int[] getBitMasks_FOR_TEST(boolean[][] comboMatrix){
      return ResultStreamMerger.getBitMasks(comboMatrix);
   }
   
   public static Map<ITuple, Double> getMarginal_FOR_TEST(final List<List<IFinalOutput>>   lists         , 
                                                   final long                     curEndInterval,
                                                   int[]                          curIndex      ,
                                                   final int[]                    listLen       ,
                                                   final boolean[][]              comboMatrix   ,
                                                   final int[]                    bitMasks      ,
                                                   final double[]                 initialComboProbs){
      return ResultStreamMerger.getMarginal(lists, curEndInterval, curIndex, listLen, comboMatrix, bitMasks, initialComboProbs);
   }
}