package edu.washington.cs.rfid.cascadia.playback.client.ui;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.google.gwt.event.logical.shared.ValueChangeEvent;
import com.google.gwt.event.logical.shared.ValueChangeHandler;
import com.google.gwt.maps.client.geom.LatLng;
import com.google.gwt.user.client.ui.CheckBox;
import com.google.gwt.user.client.ui.HTML;
import com.google.gwt.user.client.ui.HorizontalPanel;
import com.google.gwt.user.client.ui.Image;
import com.google.gwt.user.client.ui.VerticalPanel;

import edu.washington.cs.rfid.cascadia.playback.client.data.PlaybackOptions;
import edu.washington.cs.rfid.map.client.data.LatLon;
import edu.washington.cs.rfid.map.client.data.LocationData;
import edu.washington.cs.rfid.map.client.data.ParticleCloud;
import edu.washington.cs.rfid.map.client.data.TrackedEntity;
import edu.washington.cs.rfid.map.client.data.TrackingData;
import edu.washington.cs.rfid.map.client.data.cache.DefaultMapFloors;
import edu.washington.cs.rfid.map.client.ui.MercatorMapWidget;
import edu.washington.cs.rfid.map.client.ui.overlay.MarkerFactory;
import edu.washington.cs.rfid.map.client.ui.overlay.OverlayFilter;
import edu.washington.cs.rfid.map.client.ui.overlay.OverlayWrapper;


/**
 * A panel containing an map widget that takes and displays location data.
 * @author evan
 */
public class PlaybackMap extends VerticalPanel {

   protected static final String STYLE_PLAYBACK_WIDGET = "playback-widget";
   protected static final String DEFAULT_ICON_URL      = "http://rfid.cs.washington.edu/map-icons/";
   protected static final String DEFAULT_FLOOR         = DefaultMapFloors.DEFAULT_FLOORS[3];
   
   protected static final int    DEFAULT_ZOOM          = 20;
   protected static final String DEFAULT_WIDTH         = "450px";
   protected static final String DEFAULT_HEIGHT        = "400px";
   protected static final int    PAN_THRESHOLD         = 10;
   
   //--//
   
   protected MarkerFactory            factory;
   protected Map<String,TrackingData> keyToTrackingData;
   
   protected PlaybackOptions          playbackOptions;
   protected String                   followedKey;
   
   protected HTML                     followingLabel;
   protected HorizontalPanel          alsoShowingPanel;
   protected Map<String,Boolean>      showingKeysMap;
   protected MercatorMapWidget        map;
   protected CheckBox                 cbData;
   protected CheckBox                 cbModel;
   
   //--//
   
   /**
    * Creates a new PlaybackMap using the playback options for the main item to
    * be tracked, as well as all the other items that the user may eventually want
    * to track in this map.
    * @param playbackOptions
    * @param playbackOptionList
    */
   public PlaybackMap(PlaybackOptions       playbackOptions,
                      List<PlaybackOptions> playbackOptionList){

      this.factory           = new MarkerFactory(PlaybackMap.DEFAULT_ICON_URL);
      this.keyToTrackingData = new HashMap<String,TrackingData>();
      
      // Initialize the main key (i.e., tracked entity) for this PlaybackWidget
      this.playbackOptions = playbackOptions;
      this.followedKey     = playbackOptions.followedKey;
      this.keyToTrackingData.put(this.followedKey, this.createTrackingData(playbackOptions));
      
      // Initialize the other keys and show them as checkbox options in the "also showing" panel
      this.alsoShowingPanel = new HorizontalPanel();
      this.alsoShowingPanel.add(new HTML("Also showing:&nbsp;&nbsp;"));
      this.showingKeysMap = new HashMap<String,Boolean>();
      this.showingKeysMap.put(this.followedKey, true);
      
      for(PlaybackOptions o : playbackOptionList){
         if(!o.followedKey.equals(this.followedKey)){
            CheckBox cb = new CheckBox(o.trackedEntityName);
            cb.addValueChangeHandler(new ShowingValueChangeHandler(o.followedKey));
            this.alsoShowingPanel.add(cb);
            this.alsoShowingPanel.add(new HTML("&nbsp;&nbsp;"));
            
            this.keyToTrackingData.put(o.followedKey, this.createTrackingData(o));
         }
      }
      
      // Create the label panel that shows info on the main key
      HorizontalPanel labelPanel = new HorizontalPanel();
      labelPanel.add(new HTML("<b>Following:</b>&nbsp;&nbsp;"));
      labelPanel.add(new Image(PlaybackMap.DEFAULT_ICON_URL + "/" + playbackOptions.trackedEntityImgUrl));
      this.followingLabel = new HTML(playbackOptions.trackedEntityName);
      labelPanel.add(this.followingLabel);
      
      
      // Create and initialize the map
      this.map = new MercatorMapWidget(PlaybackMap.DEFAULT_FLOOR ,
                                       PlaybackMap.DEFAULT_ZOOM  ,
                                       PlaybackMap.DEFAULT_WIDTH ,
                                       PlaybackMap.DEFAULT_HEIGHT,
                                       false);
      
      this.map.clearOverlays();
      this.map.checkResize();
      this.map.removeLargeMapControl();
      this.map.removeFloorControl();
      
      OverlayFilter filter = OverlayFilter.createDefaultFilter();
      filter.setAntVisible(false);
      filter.setRdrVisible(false);
      filter.setParticlesVisible(true);
      filter.setEntitiesVisible(true);
      filter.setConnectivityGraphVisible(true);
      
      this.map.setFilter(filter, true);
      
      
      // Create the bottom panel containing checkbox controls
      HorizontalPanel cbPanel = new HorizontalPanel();
      this.cbData             = new CheckBox("Show Data" );
      this.cbModel            = new CheckBox("Show Guesses");
      this.cbData.setValue(true);
      this.cbModel.setValue(true);
      cbPanel.add(this.cbData );
      cbPanel.add(this.cbModel);
      
      
      // Initialize the PlaybackMap panel itself
      this.setHorizontalAlignment(VerticalPanel.ALIGN_CENTER);
      this.add(labelPanel);
      this.add(this.alsoShowingPanel);
      HTML spacer = new HTML("&nbsp;");
      spacer.setStyleName("gwt-HTML-small");
      this.add(spacer);
      this.add(this.map);
      this.add(cbPanel);
      this.setStyleName(PlaybackMap.STYLE_PLAYBACK_WIDGET);
      
      
      // Initialize the value change handlers for the checkboxes
      this.cbData.addValueChangeHandler(new ValueChangeHandler<Boolean>(){
         public void onValueChange(ValueChangeEvent<Boolean> e){
            if(!e.getValue()){
               map.clearOverlayType(OverlayWrapper.TYPE_ENTITY);
            }
         }
      });
      this.cbModel.addValueChangeHandler(new ValueChangeHandler<Boolean>(){
         public void onValueChange(ValueChangeEvent<Boolean> e){
            if(!e.getValue()){
               map.clearOverlayType(OverlayWrapper.TYPE_PARTICLE);
            }
         }
      });
   }
   
   
   /**
    * Changes whether or not the location markers for the given key are shown
    * in this PlaybackWidget.
    * @author evan
    */
   private class ShowingValueChangeHandler implements ValueChangeHandler<Boolean> {
      
      String key;
      
      public ShowingValueChangeHandler(String key){
         this.key = key;
      }
      
      public void onValueChange(ValueChangeEvent<Boolean> e){
         if(e.getValue()){
            showingKeysMap.put(this.key, true);
         } else {
            showingKeysMap.put(this.key, false);
         }
      }
   }
   
   
   /**
    * Clears all map overlays
    */
   public void clearOverlays(){
      this.map.clearOverlays();
   }
      
   
   /**
    * Updates the location of each tracked item in locationMap.
    * @param locationMap
    */
   public void updateLocation(Map<String,LocationData> locationMap){
      
      for(String key : locationMap.keySet()){
         
         // Decide whether to show the location data for this item
         Boolean show = this.showingKeysMap.get(key);
         if(null != show && show){
            
            TrackingData data = this.keyToTrackingData.get(key);
            if(null != data){
               LocationData locData  = locationMap.get(key);
               
               String oldFloor = (null != data.entity.reading)? data.entity.reading.floor : null;
               LatLon oldLoc   = (null != data.entity.reading)? data.entity.reading.loc   : null;
                  
               // Do updates
               data.entity.updateLocation(locData.reading);
               data.cloud.updateParticles(locData.particles);
               this.map.processTrackedEntity(data.entity);
               this.map.processParticleCloud(data.cloud);
               
               // Check if we need to pan or switch floors
               if(key.equals(this.followedKey)){
                  String newFloor = (null != data.entity.reading)? data.entity.reading.floor : null;
                  LatLon newLoc   = (null != data.entity.reading)? data.entity.reading.loc : null;
                  
                  // Check floor
                  boolean oldIsNull   = (null == oldFloor && null != newFloor);
                  boolean bothNonNull = (null != oldFloor && null != newFloor);
                  
                  if(oldIsNull || (bothNonNull && !oldFloor.equals(newFloor))){
                     this.map.setFloor(newFloor);
                  }

                  // Check location
                  if(null == oldLoc && null != newLoc){
                     LatLng newCenter = LatLng.newInstance(newLoc.lat, newLoc.lon);
                     this.map.panTo(newCenter);
                  } else if(null != oldLoc && null != newLoc){
                     LatLng oldCoord = LatLng.newInstance(oldLoc.lat, oldLoc.lon);
                     LatLng newCoord = LatLng.newInstance(newLoc.lat, newLoc.lon);
                     if(newCoord.distanceFrom(oldCoord) > PlaybackMap.PAN_THRESHOLD){
                        this.map.panTo(newCoord);
                     }
                  }
               }
            }
         }
      }
   }
   
   
   //--//
   
   
   /**
    * Creates a new TrackingData object using the information in the provided PlaybackOptions.
    * @param o
    * @return
    */
   private TrackingData createTrackingData(PlaybackOptions o){
      
      TrackedEntity e = new TrackedEntity(o.followedKey,o.trackedEntityType,this.factory);
      ParticleCloud c = new ParticleCloud(o.followedKey,o.particleColor    ,this.factory);
      
      return new TrackingData(e,c);
   }
}
