package edu.washington.cs.rfid.cascadia.scenic.client.grammar;

import java.util.ArrayList;
import java.util.Vector;

import com.google.gwt.user.client.rpc.IsSerializable;

import edu.washington.cs.rfid.cascadia.scenic.client.xml.Document;
import edu.washington.cs.rfid.cascadia.scenic.client.xml.Node;

public class ScenicSequence implements IsSerializable {

   protected static final String SCENIC_SEQUENCE   = "scenicsequence";
   protected static final String TIME_FROM         = "timefrom";
   protected static final String TIME_TO           = "timeto";
   protected static final String TIME_DAYS         = "timedays";
   protected static final String SCENIC_ACTOR      = "scenicactor";
   protected static final String SCENIC_PERSON     = "scenicperson";
   protected static final String SCENIC_PERSON_GRP = "scenicpersongrp";
   protected static final String SCENIC_THING      = "scenicthing";
   protected static final String SCENIC_THING_GRP  = "scenicthinggrp";
   protected static final String SCENIC_PLACE      = "scenicplace";
   protected static final String SCENIC_SCENE      = "scenicscene";
   protected static final String SCENIC_TRANSITION = "scenictransition";
   
   //--//
   
   protected String                   fromTime;       // Start time of day for the event
   protected String                   toTime;         // End time of day for the event
   protected String                   days;           // Set of days for the event
   protected Vector<ScenicActor>      actors;         // Actors referenced in the specification
   protected Vector<ScenicScene>      scenes;         // Scenes in the specification
   protected Vector<ScenicTransition> transitions;    // Transitions in the specification
   
   //--//
   
   
   /**
    * Creates a new default ScenicSequence
    */
   public ScenicSequence(){
      this.fromTime    = null;
      this.toTime      = null;
      this.days        = null;
      this.actors      = new Vector<ScenicActor>();
      this.scenes      = new Vector<ScenicScene>();
      this.transitions = new Vector<ScenicTransition>();
   }
   
   
   /**
    * Creates a new ScenicSequence from XML.
    * @param xml
    */
   public ScenicSequence(String xml){
      this.fromTime           = null;
      this.toTime             = null;
      this.days               = null;
      this.actors             = new Vector<ScenicActor>();
      this.scenes             = new Vector<ScenicScene>();
      this.transitions        = new Vector<ScenicTransition>();
      Document        doc     = Document.xmlParse(xml);
      ArrayList<Node> topNode = doc.getChildren();
      ArrayList<Node> kids    = null;

      if(topNode.size() > 0 && topNode.get(0).getName().equals(ScenicSequence.SCENIC_SEQUENCE)){
         kids = topNode.get(0).getChildNodes();

         for(int i = 0; i < kids.size(); i++){
            Node   n    = kids.get(i);
            String name = n.getName();

            if(name.equals(ScenicSequence.TIME_FROM)){
               this.fromTime = n.getChildNodes().get(0).getValue();
            } else if(name.equals(ScenicSequence.TIME_TO)){
               this.toTime = n.getChildNodes().get(0).getValue();
            } else if(name.equals(ScenicSequence.TIME_DAYS)){
               this.days = n.getChildNodes().get(0).getValue();
            } else if(name.equals(ScenicSequence.SCENIC_ACTOR)){
               this.actors.add(new ScenicActor(n.getChildNodes()));
            } else if(name.equals(ScenicSequence.SCENIC_PERSON)){
               this.actors.add(new ScenicPerson(n.getChildNodes()));
            } else if(name.equals(ScenicSequence.SCENIC_PERSON_GRP)){
               this.actors.add(new ScenicPersonGrp(n.getChildNodes()));
            } else if(name.equals(ScenicSequence.SCENIC_THING)){
               this.actors.add(new ScenicThing(n.getChildNodes()));
            } else if(name.equals(ScenicSequence.SCENIC_THING_GRP)){
               this.actors.add(new ScenicThingGrp(n.getChildNodes()));
            } else if(name.equals(ScenicSequence.SCENIC_PLACE)){
               this.actors.add(new ScenicPlace(n.getChildNodes()));
            } else if(name.equals(ScenicSequence.SCENIC_SCENE)){
               this.scenes.add(new ScenicScene(n.getChildNodes()));
            } else if(name.equals(ScenicSequence.SCENIC_TRANSITION)){
               this.transitions.add(new ScenicTransition(n.getChildNodes()));
            }
         }
      } else {
         this.fromTime    = null;
         this.toTime      = null;
         this.days        = null;
         this.actors      = new Vector<ScenicActor>();
         this.scenes      = new Vector<ScenicScene>();
         this.transitions = new Vector<ScenicTransition>();
      }
   }
   
   
   /**
    * Creates a new ScenicSequence from parameters in the specification interface
    * @param fromTime
    * @param toTime
    * @param days
    * @param actors
    * @param scenes
    * @param transitions
    */
   public ScenicSequence(String                   fromTime   ,
                         String                   toTime     ,
                         String                   days       ,
                         Vector<ScenicActor>      actors     , 
                         Vector<ScenicScene>      scenes     ,
                         Vector<ScenicTransition> transitions){
      this.fromTime    = fromTime;
      this.toTime      = toTime;
      this.days        = days;
      this.actors      = actors;
      this.scenes      = scenes;
      this.transitions = transitions;
   }
   
   
   public String getFromTime(){
      return this.fromTime;
   }
   
   public String getToTime(){
      return this.toTime;
   }
   
   public String getDays(){
      return this.days;
   }
   
   public Vector<ScenicActor> getActors(){
      return this.actors;
   }
   
   public Vector<ScenicScene> getScenes(){
      return this.scenes;
   }
   
   public Vector<ScenicTransition> getTransitions(){
      return this.transitions;
   }
   
   public void setFromTime(String time){
      this.fromTime = time;
   }
   
   public void setToTime(String time){
      this.toTime = time;
   }
   
   public void setDays(String time){
      this.days = time;
   }
   
   public void addActor(ScenicActor actor){
      if (!this.actors.contains( actor )){
         this.actors.add( actor );
      }
   }
   
   public void addScene(ScenicScene scene){
      this.scenes.add(scene);
   }
   
   public void addTransition(ScenicTransition transition){
      this.transitions.add(transition);
   }
   
   public String toXml(){
       
       String xml = "<scenicsequence>\n";
       
       xml += "<timefrom>" + this.fromTime + "</timefrom>" +
       		  "<timeto>"   + this.toTime   + "</timeto>"   +
       		  "<timedays>" + this.days     + "</timedays>";
       
       if(null != this.actors) {
          for(int i = 0; i < this.actors.size(); i++){
             xml += this.actors.get(i).toXml(1);
          }
       }
       
       if(null != this.scenes && null != this.transitions) {
          
          for(int i = 0; i < this.scenes.size(); i++){
             if(i < this.transitions.size()){
                xml += this.scenes.get(i).toXml(1);
                xml += this.transitions.get(i).toXml(1);
             } else {
                xml += this.scenes.get(i).toXml(1);
             }
          }
       }
       
       xml += "</scenicsequence>\n";
       
       return xml;
   }
}