package edu.washington.cs.rfid.cascadia.scenic.client.grammar;

import java.util.ArrayList;
import java.util.List;
import java.util.Vector;

import edu.washington.cs.rfid.cascadia.scenic.client.data.Actor;
import edu.washington.cs.rfid.cascadia.scenic.client.data.Far;
import edu.washington.cs.rfid.cascadia.scenic.client.data.Inside;
import edu.washington.cs.rfid.cascadia.scenic.client.data.Near;
import edu.washington.cs.rfid.cascadia.scenic.client.data.Outside;
import edu.washington.cs.rfid.cascadia.scenic.client.data.Person;
import edu.washington.cs.rfid.cascadia.scenic.client.data.PersonGrp;
import edu.washington.cs.rfid.cascadia.scenic.client.data.Place;
import edu.washington.cs.rfid.cascadia.scenic.client.data.Primitive;
import edu.washington.cs.rfid.cascadia.scenic.client.data.Thing;
import edu.washington.cs.rfid.cascadia.scenic.client.data.ThingGrp;
import edu.washington.cs.rfid.cascadia.scenic.client.data.With;
import edu.washington.cs.rfid.cascadia.scenic.client.data.Without;
import edu.washington.cs.rfid.cascadia.scenic.client.exception.InvalidArgumentException;
import edu.washington.cs.rfid.cascadia.scenic.client.ui.Scene;
import edu.washington.cs.rfid.cascadia.scenic.client.ui.SequencePanel;
import edu.washington.cs.rfid.cascadia.scenic.client.ui.Transition;
import edu.washington.cs.rfid.cascadia.scenic.client.ui.widget.ActorWidget;
import edu.washington.cs.rfid.cascadia.scenic.client.ui.widget.ConnectorHandleWidget;
import edu.washington.cs.rfid.cascadia.scenic.client.ui.widget.PersonGrpWidget;
import edu.washington.cs.rfid.cascadia.scenic.client.ui.widget.PrimitiveWidget;
import edu.washington.cs.rfid.cascadia.scenic.client.ui.widget.ThingGrpWidget;

/**
 * A class to transform widgets into serializable grammar objects.
 * 
 * @author evan
 */
public class GrammarEncoder {
   
   /**
    * Gets a grammar object given a sequence widget.
    * @param sequencePanel
    * @return
    */
   public static ScenicSequence getScenicSequence(SequencePanel sequencePanel){

      ScenicSequence sequence = new ScenicSequence();
      Vector<Scene>  scenes   = sequencePanel.getScenes();

      // Add the time constraints
      sequence.setFromTime(sequencePanel.getFromTime());
      sequence.setToTime(sequencePanel.getToTime());
      sequence.setDays(sequencePanel.getDays());
      
      // Add the Scenes and Transitions
      for(int i = 0; i < scenes.size(); i++){
         Scene       s           = scenes.get(i);
         ScenicScene scenicScene = GrammarEncoder.getScenicScene(s, sequence);
         sequence.addScene(scenicScene);
         
         // Now check for a following transition...
         Vector<Transition> transitions = sequencePanel.getTransitions();
         if(i < transitions.size()){
            Transition       t                = transitions.get(i);
            ScenicTransition scenicTransition = new ScenicTransition(t.getType());
            sequence.addTransition(scenicTransition);
         }
      }
      return sequence;
   }
   
   
   /**
    * Gets a list of all sub-events for a given sequence event.
    * @param scenicSequence
    * @return
    */
   public static List<ScenicSequence> getSubEvents(ScenicSequence scenicSequence){
      
      List<ScenicSequence> subEvents = new ArrayList<ScenicSequence>();
      
      // Get timing constraints
      String fromTime = scenicSequence.getFromTime();
      String toTime   = scenicSequence.getToTime();
      String days     = scenicSequence.getDays();
      
      // Get all sub-sequences
      for(int i = 0; i < scenicSequence.scenes.size(); i++){
         List<ScenicScene>      scenes              = new ArrayList<ScenicScene>();
         List<ScenicTransition> subEventTransitions = new ArrayList<ScenicTransition>();
         for(int j = 0; j <= i; j++){
            scenes.add(scenicSequence.scenes.get(j));
            
            // Now add any transitions
            if(j < i){
               subEventTransitions.add(scenicSequence.transitions.get(j));
            }
         }
         ScenicSequence subEvent = GrammarEncoder.getSubEventScenicSequence(fromTime,toTime,days,scenes);
         for(ScenicTransition transition : subEventTransitions){
            subEvent.addTransition(transition);
         }
         subEvents.add(subEvent);
      }
      
      // Get all single scene and single primitive sequences
      for(int i = 0; i < scenicSequence.scenes.size(); i++){
         ScenicScene    scene    = scenicSequence.scenes.get(i);
         ScenicSequence subEvent = GrammarEncoder.getSubEventScenicSequence(fromTime,toTime,days,scene);
         subEvents.add(subEvent);
         
         for(ScenicPrimitive primitive : scene.primitives){
            subEvent = GrammarEncoder.getSubEventScenicSequence(fromTime,toTime,days,primitive);
            subEvents.add(subEvent);
         }
      }
      return subEvents;
   }
   
   
   /**
    * Gets a grammar object for a sequence consisting of a sequence containing 
    * only the given Scenes.
    * @param sequencePanel
    * @return
    */
   public static ScenicSequence getSubEventScenicSequence(String            fromTime,
                                                          String            toTime  ,
                                                          String            days    ,
                                                          List<ScenicScene> scenicScenes){
      ScenicSequence scenicSequence = new ScenicSequence();
      
      // Add the time constraints
      scenicSequence.setFromTime(fromTime);
      scenicSequence.setToTime(toTime);
      scenicSequence.setDays(days);
      
      for(int i = 0; i < scenicScenes.size(); i++){
         ScenicScene scenicScene = scenicScenes.get(i);
         
         // Add the actors to the sequence
         Vector<ScenicActor> actors = scenicScene.getActors();
         for(ScenicActor actor : actors){
            scenicSequence.addActor(actor);
         }
         
         // Add the scene
         scenicSequence.addScene(scenicScene);
      }
      return scenicSequence;
   }
   
   
   /**
    * Gets a grammar object for a sequence consisting of only the given Scene.
    * @param sequencePanel
    * @return
    */
   public static ScenicSequence getSubEventScenicSequence(String      fromTime,
                                                          String      toTime  ,
                                                          String      days    ,
                                                          ScenicScene scenicScene){
      ScenicSequence scenicSequence = new ScenicSequence();
      
      // Add the time constraints
      scenicSequence.setFromTime(fromTime);
      scenicSequence.setToTime(toTime);
      scenicSequence.setDays(days);
      
      // Add the actors to the sequence
      Vector<ScenicActor> actors = scenicScene.getActors();
      for(ScenicActor actor : actors){
         scenicSequence.addActor(actor);
      }
      
      // Add the scene
      scenicSequence.addScene(scenicScene);
      
      return scenicSequence;
   }

   
   /**
    * Gets a grammar object for a sequence consisting of only the given primitive.
    * @param sequencePanel
    * @return
    */
   public static ScenicSequence getSubEventScenicSequence(String          fromTime,
                                                          String          toTime  ,
                                                          String          days    ,
                                                          ScenicPrimitive scenicPrimitive){
      ScenicSequence scenicSequence = new ScenicSequence();
      
      scenicSequence.setFromTime(fromTime);
      scenicSequence.setToTime(toTime);
      scenicSequence.setDays(days);
      
      // Add the actors to the sequence and copy the actor vector
      Vector<ScenicActor> actorCopy = new Vector<ScenicActor>();
      Vector<ScenicActor> actors    = scenicPrimitive.actors;
      for(ScenicActor actor : actors){
         scenicSequence.addActor(actor);
         actorCopy.add(actor);
      }
      
      // Create a new primitive Vector
      Vector<ScenicPrimitive> primitives = new Vector<ScenicPrimitive>();
      primitives.add(scenicPrimitive);
      
      // Create the scene
      ScenicScene scene = new ScenicScene(1L, actorCopy, primitives);
            
      // Add the scene
      scenicSequence.addScene(scene);
      
      return scenicSequence;
   }
   
   
   
   //--//
   
   private static ScenicScene getScenicScene(Scene s, ScenicSequence sequence){
      Vector<ActorWidget>           actorWidgets     = s.getActorWidgets    ();
      Vector<PrimitiveWidget>       primitiveWidgets = s.getPrimitiveWidgets();
      
      Vector<ScenicActor>           scenicActors     = new Vector<ScenicActor>    ();
      Vector<ScenicPrimitive>       scenicPrimitives = new Vector<ScenicPrimitive>();
                   
      // Distill Actors
      for(int j = 0; j < actorWidgets.size(); j++){
         
         ActorWidget aw          = actorWidgets.get(j);
         Actor       a           = aw.getActor();
         ScenicActor scenicActor = GrammarEncoder.getScenicActor(s, aw, a);
         
         if(null != scenicActor){
            scenicActors.add(scenicActor);
            sequence.addActor(scenicActor);
         }
      }

      // Distill Primitives
      for(int j = 0; j < primitiveWidgets.size(); j++){
         
         PrimitiveWidget pw              = primitiveWidgets.get(j);
         Primitive       p               = pw.getPrimitive();
         ScenicPrimitive scenicPrimitive = GrammarEncoder.getScenicPrimitive(s, pw, p);

         if(null != scenicPrimitive){
            scenicPrimitives.add(scenicPrimitive);
         }
      }
      
      return new ScenicScene(1L, scenicActors, scenicPrimitives);
   }
   
   
   /**
    * Distills a ScenicActor from a Scene, ActorWidget and Actor.
    * @param s
    * @param aw
    * @param a
    * @return
    */
   private static ScenicActor getScenicActor(Scene s, ActorWidget aw, Actor a){

      ScenicActor scenicActor = null;
      int         id          = a.getNum();
      String      name        = a.getName();
      boolean     type        = a.getIsVbl();
      int         left        = s.getWidgetLeft(aw);
      int         top         = s.getWidgetTop(aw);
      String      imgUrl      = aw.getImgUrl();
      
      if(a instanceof Person){
         scenicActor = new ScenicPerson(id, name, type, left, top, imgUrl);
      } else if(a instanceof PersonGrp){
         PersonGrpWidget pgw                = (PersonGrpWidget)aw;
         scenicActor                        = new ScenicPersonGrp(id, name, type, left, top, imgUrl);
        ((ScenicPersonGrp)scenicActor).size = pgw.getSize();
      } else if(a instanceof Thing){
         scenicActor = new ScenicThing(id, name, type, left, top, imgUrl);
      } else if(a instanceof ThingGrp){
         ThingGrpWidget tgw                 = (ThingGrpWidget)aw;
         scenicActor                        = new ScenicThingGrp(id, name, type, left, top, imgUrl);
         ((ScenicThingGrp)scenicActor).size = tgw.getSize();
      } else if(a instanceof Place){
         scenicActor = new ScenicPlace(id, name, type, left, top, imgUrl);
      }
      return scenicActor;
   }
   
   
   /**
    * Distills a ScenicPrimitive from a Scene, PrimitiveWidget and Primitive.
    * @param s
    * @param pw
    * @param p
    * @return
    */
   public static ScenicPrimitive getScenicPrimitive(Scene s, PrimitiveWidget pw, Primitive p){
      
      ScenicPrimitive scenicPrimitive = null;
      int             left            = s.getWidgetLeft(pw);
      int             top             = s.getWidgetTop (pw);
      
      // Find which actors this primitive connects to
      Vector<ConnectorHandleWidget> connectorWidgets = pw.getConnectorHandles();
      Vector<ScenicActor>           scenicActors     = new Vector<ScenicActor>();
      
      for(ConnectorHandleWidget connector : connectorWidgets){
         ActorWidget aw = connector.actorWidget;
         if(null != aw){
            Actor a = aw.getActor();
            scenicActors.add(GrammarEncoder.getScenicActor(s, aw, a));
         }
      }
      
      if(p instanceof With){
         scenicPrimitive = new ScenicWith(scenicActors, left, top);
      } else if(p instanceof Without){
         scenicPrimitive = new ScenicWithout(scenicActors, left, top);
      } else if(p instanceof Near){
         scenicPrimitive = new ScenicNear(scenicActors, left, top);
      } else if(p instanceof Far){
         scenicPrimitive = new ScenicFar(scenicActors, left, top);
      } else if(p instanceof Inside){
         scenicPrimitive = new ScenicInside(scenicActors, left, top);
      } else if(p instanceof Outside){
         scenicPrimitive = new ScenicOutside(scenicActors, left, top);
      }

      for(int k = 0; k < scenicActors.size(); k++){
         ScenicActor a1 = scenicActors.get(k);
                                 
         for(int l = k+1; l < scenicActors.size(); l++){
            ScenicActor a2 = scenicActors.get(l);
                                          
            try {
               scenicPrimitive.addRelation(a1, a2);
            } catch(InvalidArgumentException e){
            }
         }
      }
      return scenicPrimitive;
   }
}
