package edu.washington.cs.rfid.cascadia.scenic.client.grammar;

import java.util.ArrayList;
import java.util.Vector;

import com.google.gwt.user.client.rpc.IsSerializable;

import edu.washington.cs.rfid.cascadia.scenic.client.data.Pair;
import edu.washington.cs.rfid.cascadia.scenic.client.exception.InvalidArgumentException;
import edu.washington.cs.rfid.cascadia.scenic.client.xml.Node;

public class ScenicPrimitive implements IsSerializable {
   
   protected static final String SCENIC_ACTOR       = "scenicactor";
   protected static final String SCENIC_PERSON      = "scenicperson";
   protected static final String SCENIC_PERSON_GRP  = "scenicpersongrp";
   protected static final String SCENIC_THING       = "scenicthing";
   protected static final String SCENIC_THING_GRP   = "scenicthinggrp";
   protected static final String SCENIC_PLACE       = "scenicplace";
   
   protected static final String LEFT               = "left";
   protected static final String TOP                = "top";
   protected static final String RELATION           = "relation";
   protected static final String RELATION_DELIMETER = ",";
   
   //--//
   
   protected Vector<ScenicActor> actors;
   protected boolean[][]         relations;
   protected int                 left;
   protected int                 top;
   
   
   public ScenicPrimitive(){
      this.actors    = null;
      this.relations = null;
      this.left      = -1;
      this.top       = -1;
   }
   
   public ScenicPrimitive(ArrayList<Node> kids){
       
      this.actors = new Vector<ScenicActor>();
       
      // Extract actors, left, and top first
      for(int i = 0; i < kids.size(); i++){
         Node   n    = kids.get(i);
         String name = n.getName();
          
         if(name.equals(ScenicPrimitive.SCENIC_ACTOR)){
            this.actors.add(new ScenicActor(n.getChildNodes()));
         } else if(name.equals(ScenicPrimitive.SCENIC_PERSON)){
            this.actors.add(new ScenicPerson(n.getChildNodes()));
         } else if(name.equals(ScenicPrimitive.SCENIC_PERSON_GRP)){
            this.actors.add(new ScenicPersonGrp(n.getChildNodes()));
         } else if(name.equals(ScenicPrimitive.SCENIC_THING)){
            this.actors.add(new ScenicThing(n.getChildNodes()));
         } else if(name.equals(ScenicPrimitive.SCENIC_THING_GRP)){
            this.actors.add(new ScenicThingGrp(n.getChildNodes()));
         } else if(name.equals(ScenicPrimitive.SCENIC_PLACE)){
            this.actors.add(new ScenicPlace(n.getChildNodes()));
         } else if(name.equals(ScenicPrimitive.LEFT)){
            try {
               this.left = Integer.parseInt(((Node)n.getChildNodes().get(0)).getValue());
            } catch(NumberFormatException e){
               this.left = 0;
            }
         } else if(name.equals(ScenicPrimitive.TOP)){
            try {
               this.top = Integer.parseInt(((Node)n.getChildNodes().get(0)).getValue());
            } catch(NumberFormatException e){
               this.top = 0;
            }
         }
      }
       
      // Assign relations on second pass
      this.relations = new boolean[this.actors.size()][this.actors.size()];
       
      for(int i = 0; i < kids.size(); i++){
   
         Node   n    = kids.get(i);
         String name = n.getName();
          
         if(name.equals(ScenicPrimitive.RELATION)){
            ArrayList<Node> relationKids = n.getChildNodes();
            
            if(null != relationKids){
               Node     pair  = relationKids.get(0);
               String[] names = pair.getValue().split(ScenicPrimitive.RELATION_DELIMETER);

               if(2 == names.length){
                  try {
//                     long firstId  = Long.parseLong(ids[0]);
//                     long secondId = Long.parseLong(ids[1]);
                     String firstName  = names[0];
                     String secondName = names[1];
                     int    first      = this.indexOfActor(firstName );
                     int    second     = this.indexOfActor(secondName);

                     if(first >= 0 && second >= 0){
                        this.relations[first][second] = true;
                        this.relations[second][first] = true;
                     }
                      
                  } catch(NumberFormatException e){}
               }
            }      
         }
      }
   }
   
   public ScenicPrimitive(Vector<ScenicActor> actors, int left, int top){
      this.actors    = actors;
      this.relations = new boolean[this.actors.size()][this.actors.size()];
      this.left      = left;
      this.top       = top;
   }
   
   protected int indexOfActor(long id){
      int idx = -1;
      for(int i = 0; i < this.actors.size(); i++){
         if(id == this.actors.get(i).id){
             idx = i;
             break;
         }
      }
      return idx;
   }
   
   protected int indexOfActor(String name){
      int idx = -1;
      for(int i = 0; i < this.actors.size(); i++){
         if(name.equals(this.actors.get(i).name)){
             idx = i;
             break;
         }
      }
      return idx;
   }
   
   public void addActor(ScenicActor actor){
      this.actors.add(actor);
      
      boolean[][] newRelations = new boolean[this.actors.size()][this.actors.size()];

      for(int i=0; i<relations.length; i++){
         for(int j=0; j<relations.length; j++){
            newRelations[i][j] = this.relations[i][j]; 
         }
      }
      this.relations = newRelations; 
   }
   
   /**
    * 
    * @precondition a1 and a2 must be contained by this primitive already
    * @param a1
    * @param a2
    */
   public void addRelation(ScenicActor a1, ScenicActor a2) throws InvalidArgumentException {
      int a1Index = this.actors.indexOf(a1);
      int a2Index = this.actors.indexOf(a2);
      
      if(-1 == a1Index || -1 == a2Index){
         throw new InvalidArgumentException();
      }
      this.relations[a1Index][a2Index] = true;
      this.relations[a2Index][a1Index] = true;
   }
   
   
   public void removeRelation(ScenicActor a1, ScenicActor a2) throws InvalidArgumentException {
      int a1Index = this.actors.indexOf(a1);
      int a2Index = this.actors.indexOf(a2);
      
      if(-1 == a1Index || -1 == a2Index){
         throw new InvalidArgumentException();
      }
      this.relations[a1Index][a2Index] = false;
      this.relations[a2Index][a1Index] = false;
   }
   
   
   /**
    * 
    * @param a1
    * @param a2
    * @return
    * @throws InvalidArgumentException
    */
   public boolean inRelation(ScenicActor a1, ScenicActor a2) throws InvalidArgumentException {
      int a1Index = this.actors.indexOf(a1);
      int a2Index = this.actors.indexOf(a2);
      
      if(-1 == a1Index || -1 == a2Index){
         throw new InvalidArgumentException();
      }
      return this.relations[a1Index][a2Index];
   }
   
   
   /**
    * 
    * @return
    */
   public Vector<Pair<ScenicActor,ScenicActor>> getRelation(){
      Vector<Pair<ScenicActor,ScenicActor>> relations = new Vector<Pair<ScenicActor,ScenicActor>>();
      
      for(int i = 0; i < this.relations.length; i++){
         for(int j =  i; j < this.relations[i].length; j++){
            if(this.relations[i][j]){
               relations.add(new Pair(this.actors.get(i),this.actors.get(j)));
            }
         }
      }
      return relations;
   }
   
   public int getLeft(){
      return this.left;
   }
    
   public int getTop(){
      return this.top;
   }
   
   public String toXml(){
       return this.toXml(0);
   }
   
   /**
    * Generates an XML representation of this object.
    * @return
    */
   public String toXml(int level){
      
      String indent = "";
       
      for(int i = 0; i < level; i++){
         indent += "\t";
      }
       
      String xml = indent + "<scenicprimitive>\n";
       
      if(null != this.actors){
         for(int i = 0; i < this.actors.size(); i++){
            xml += indent + this.actors.get(i).toXml(level + 1) + "\n";
         }
      }
       
      if(null != this.relations){
         for(int i = 0; i < this.relations.length; i++){
            for(int j = 0; j < this.relations[i].length; j++){
               if(this.relations[i][j]){
                   xml += indent + "\t<relation>\n" + 
                          this.actors.get(i).getName() + "," + this.actors.get(j).getName() +
                          "</relation>\n";
               }
            }
         }
      }
       
      xml += indent + "\t<left>"   + this.left   + "</left>\n"  + 
             indent + "\t<top>"    + this.top    + "</top>\n"   +
             indent + "</scenicprimitive>\n";
      return xml;
   }
}