package edu.washington.cs.rfid.cascadia.scenic.client.ui;

import java.util.ArrayList;
import java.util.List;

import com.allen_sauer.gwt.dnd.client.PickupDragController;
import com.google.gwt.event.logical.shared.ResizeEvent;
import com.google.gwt.event.logical.shared.ResizeHandler;
import com.google.gwt.user.client.DOM;
import com.google.gwt.user.client.Window;
import com.google.gwt.user.client.ui.HTML;
import com.google.gwt.user.client.ui.HasHorizontalAlignment;
import com.google.gwt.user.client.ui.HorizontalPanel;
import com.google.gwt.user.client.ui.RootPanel;
import com.google.gwt.user.client.ui.VerticalPanel;

import edu.washington.cs.rfid.cascadia.scenic.client.data.EventTask;
import edu.washington.cs.rfid.cascadia.scenic.client.data.EventTemplate;
import edu.washington.cs.rfid.cascadia.scenic.client.grammar.GrammarEncoder;
import edu.washington.cs.rfid.cascadia.scenic.client.grammar.ScenicSequence;
import edu.washington.cs.rfid.cascadia.scenic.client.ui.factory.ActorFactory;
import edu.washington.cs.rfid.cascadia.scenic.client.ui.factory.ActorLibrary;
import edu.washington.cs.rfid.cascadia.scenic.client.ui.factory.PrimitiveFactory;
import edu.washington.cs.rfid.cascadia.scenic.client.ui.factory.SequenceFactory;
import edu.washington.cs.rfid.cascadia.scenic.client.ui.widget.ActorWidget;
import edu.washington.cs.rfid.cascadia.scenic.client.ui.widget.PrimitiveWidget;
import edu.washington.cs.rfid.cascadia.scenic.client.util.JavaScriptHelper;


public class ScenicToolbar extends VerticalPanel implements ResizeHandler {

   private static final String TOOLBAR_WIDTH          = "100%";
   private static final String TOOLBAR_HEIGHT         = "300px";
   private static final String TOOLBAR_STYLE          = "toolbarPanel";
        
   private static final int    ICON_PANEL_SPACING     = 15;

   private static final String ACTOR_PANEL_WIDTH      = "230px";
   private static final String ACTOR_PANEL_HEIGHT     = "80px";
   private static final String PRIMITIVE_PANEL_WIDTH  = "420px";
   private static final String PRIMITIVE_PANEL_HEIGHT = "80px";
   private static final String TIMING_PANEL_WIDTH     = "220px";
   private static final String TIMING_PANEL_HEIGHT    = "80px";

   //--//
   
   private static ScenicToolbar        singleton;
   private static ActorLibrary         actorLibrary;
   private static PickupDragController dragController;

   //--//
   
   // DnD controllers
   private DefaultDropController dropController;
   private DefaultDropController rootDropController;   

   // Top strip with icons
   private HorizontalPanel       iconPanel;
   private IconPanel             actorPanel;
   private IconPanel             primitivePanel;
   private TimeConstraintsPanel  timingPanel;
   
   // Middle sequence strip
   private   SequenceViewer      sequenceViewer;
   protected SequencePanel       sequencePanel;
   
   // Buttons beneath the sequence
   private HorizontalPanel       buttonPanel;
   
   // Explains the currently displayed template (if there is one)
   private VerticalPanel         templatePanel;
   private int                   curTemplateNum;
   private HTML                  templateTitle;
   private HTML                  templateDescription;
   private HTML                  templateExampleTitle;
   private HTML                  templateApplication;
   private HTML                  templateScenario;

   // Task info displayed at the bottom - needed only for the formative study !
   private VerticalPanel         taskPanel;
   private int                   curTaskNum;
   private HTML                  taskNumber;
   private HTML                  taskApplication;
   private HTML                  taskScenario;
   
   
   //--//

   
   /**
    * Gets (or creates) the singleton instance of <code>ScenicToolbar</code>. 
    */
   public static ScenicToolbar getInstance(){
      if(null == ScenicToolbar.singleton){
         ScenicToolbar.singleton = new ScenicToolbar();
      }
      return ScenicToolbar.singleton;
   }
   
   
   /**
    * Gets (or creates) the singleton instance of <code>ActorLibrary</code>.
    * @return
    */
   public static ActorLibrary getActorLibrary(){
      if(null == ScenicToolbar.actorLibrary){
         ScenicToolbar.actorLibrary = new ActorLibrary();
      }
      return ScenicToolbar.actorLibrary;
   }
   
   
   /**
    * Gets (or creates) the singleton instance of <code>PickupDragController</code>.
    * @return
    */
   public static PickupDragController getDragController(){
      if(null == ScenicToolbar.dragController){
         ScenicToolbar.dragController = new PickupDragController( RootPanel.get(), true );
      }
      return ScenicToolbar.dragController;
   }
   
   
   //--//
   
   
   public String getFromTime(){
      return this.timingPanel.getFromTime();
   }
   
   
   public String getToTime(){
      return this.timingPanel.getToTime();
   }
   
   
   public String getDays(){
      return this.timingPanel.getDays();
   }
   
   
   /**
    * Gets a specification in the form of a <code>ScenicSequence</code> object.
    * @return
    */
   public ScenicSequence getSpecification(){
      return GrammarEncoder.getScenicSequence(this.sequencePanel);
   }
   
   
   public void clearSequence(){
      this.sequencePanel.clearScenes();
   }
   
   
   /**
    * Load a template from a given XML representation of a specification.
    * @param xml
    */
   public void loadTemplateFromXml(String xml){
      ScenicSequence sequence = new ScenicSequence(xml);
      this.sequencePanel.clearScenes();
      SequenceFactory.createSequencePanel(sequence, this.sequencePanel);
      this.timingPanel.setFromTime(sequence.getFromTime());
      this.timingPanel.setToTime(sequence.getToTime());
      this.timingPanel.setDays(sequence.getDays());
      
      this.displayThumbnails(sequence);
   }

   
   /**
    * Loads a template
    * @param template
    */
   public void loadTemplate(EventTemplate template){
      if(null != template){
         this.templateTitle       .setHTML("<b>Template Event #" + this.curTemplateNum + ":</b>&nbsp;&nbsp;" + template.title + "<br>&nbsp;<br>");
         this.templateDescription .setHTML("<b>Description:</b><br>" + template.description + "<br><br>");
         this.templateExampleTitle.setHTML("<b>Example:</b><br>----------<br>");
         this.templateApplication .setHTML("<b>Application:</b><br>" + template.exampleApp + "<br><br>");
         this.templateScenario    .setHTML("<b>Usage Scenario:</b><br>" + template.exampleScenario + "<br>&nbsp;<br>");
         
         this.curTemplateNum = (this.curTemplateNum % 4 == 0)? 1 : (this.curTemplateNum + 1);
         
         // Add the Sequence to the toolbar
         ScenicSequence sequence = new ScenicSequence(template.xml);
         sequencePanel.clearScenes();
         SequenceFactory.createSequencePanel(sequence, sequencePanel);
         this.timingPanel.setFromTime(sequence.getFromTime());
         this.timingPanel.setToTime(sequence.getToTime());
         this.timingPanel.setDays(sequence.getDays());
      } else {
         this.templateTitle.setHTML("Sorry, invalid template request!");
         this.templateDescription .setHTML("");
         this.templateExampleTitle.setHTML("");
         this.templateApplication .setHTML("");
         this.templateScenario    .setHTML("");
      }
   }
   
   
   /**
    * Adds a template thumbnail to the ScenicToolbar - for debugging thumbnails.
    * @param thumbnail
    */
   public void addThumbnail(SequencePanel thumbnail){
      this.add(thumbnail);
   }
   
   
   /**
    * Loads a task - useful for studies of the interface.
    * @param task
    */
   public void loadTask(EventTask task){
      if(null != task){
         this.taskNumber     .setHTML("<b>Task Number:&nbsp;&nbsp;" + task.eventNum + "</b><br>&nbsp;<br>");
         this.taskApplication.setHTML("<b>Application:</b><br>"     + task.application + "<br>&nbsp;<br>");
         this.taskScenario   .setHTML("<b>Usage Scenario:</b><br>"  + task.usageScenario + "<br>&nbsp;<br>");
         this.curTaskNum++;
      } else {
         System.out.println("result was null!");
      }
   }
   
   
   /**
    * Prints the result of a translated specification.
    * @param result
    */
   public void printTranslation(Object result){
      System.out.println("As XML:\n" + result.toString());
   }
   
   
   //------------------------------- HANDLER METHODS -----------------------------------//
   
   public void onResize(ResizeEvent e){
      int w = e.getWidth();
      this.buttonPanel.setWidth(w + "px");
   }
   
   
   //--//
   
   
   /**
    * Creates a new <code>ScenicToolbar</code>.
    */
   private ScenicToolbar(){
      super();

      this.setHeight   ( ScenicToolbar.TOOLBAR_HEIGHT );
      this.setWidth    ( ScenicToolbar.TOOLBAR_WIDTH  );
      this.setStyleName( ScenicToolbar.TOOLBAR_STYLE  );

      ScenicToolbar.actorLibrary = ScenicToolbar.getActorLibrary();

      this.initDndControls();
      this.initIconPanel  ();
      this.initSequenceViewer(SequenceFactory.createSequencePanel());
      this.initButtonPanel();
      this.initTemplatePanel();
      this.initTaskPanel();
      
      JavaScriptHelper.disableContextMenu( this.getElement()               , true );
      JavaScriptHelper.disableContextMenu( this.sequencePanel.getElement() , true );
      JavaScriptHelper.disableContextMenu( this.sequenceViewer.getElement(), true );
   }
   
   
   /**
    * Initializes drag and drop
    */
   private void initDndControls(){
      ScenicToolbar.dragController = ScenicToolbar.getDragController();
      ScenicToolbar.dragController.setBehaviorConstrainedToBoundaryPanel( true  );
      ScenicToolbar.dragController.setBehaviorMultipleSelection         ( false );
      ScenicToolbar.dragController.setBehaviorDragProxy                 ( false );
                
      this.dropController = new DefaultDropController( this );
      ScenicToolbar.dragController.registerDropController( this.dropController );
            
      this.rootDropController = new DefaultDropController( RootPanel.get() );
      ScenicToolbar.dragController.registerDropController( this.rootDropController );
   }
   
   
   /**
    * Initializes icon panel
    */
   private void initIconPanel(){
      this.iconPanel = new HorizontalPanel();
      this.iconPanel.setSpacing( ScenicToolbar.ICON_PANEL_SPACING );

      this.initActorPanel    ();
      this.initPrimitivePanel();
      this.initTimeConstraintsPanel();
      
      this.iconPanel.add( this.actorPanel     );
      this.iconPanel.add( this.primitivePanel );
      this.iconPanel.add( this.timingPanel    );
      
      JavaScriptHelper.disableTextSelection(this.actorPanel.getElement()    , true);
      JavaScriptHelper.disableTextSelection(this.primitivePanel.getElement(), true);
      JavaScriptHelper.disableTextSelection(this.timingPanel.getElement()   , true);
      
      JavaScriptHelper.disableContextMenu(this.actorPanel.getElement()    , true );
      JavaScriptHelper.disableContextMenu(this.primitivePanel.getElement(), true );
      JavaScriptHelper.disableContextMenu(this.timingPanel.getElement()   , true );
      
      this.add(this.iconPanel);
   }
   
   
   /**
    * Initializes actor panel
    */
   private void initActorPanel(){
      this.actorPanel       = new IconPanel( ScenicToolbar.dragController );

      ActorWidget person    = ActorFactory.createPersonWidget   ();
      ActorWidget personGrp = ActorFactory.createPersonGrpWidget();
      ActorWidget thing     = ActorFactory.createThingWidget    ();
      ActorWidget thingGrp  = ActorFactory.createThingGrpWidget ();
      ActorWidget place     = ActorFactory.createPlaceWidget    ();
      
      person   .setLabelVisible(false);
      personGrp.setLabelVisible(false);
      thing    .setLabelVisible(false);
      thingGrp .setLabelVisible(false);
      place    .setLabelVisible(false);
      
      this.actorPanel.add( person    );
      this.actorPanel.add( thing     );
      this.actorPanel.add( place     );
      //this.actorPanel.add( thingGrp  );
      //this.actorPanel.add( personGrp );
      
      this.actorPanel.setSize( ScenicToolbar.ACTOR_PANEL_WIDTH, ScenicToolbar.ACTOR_PANEL_HEIGHT );
   }
   
   
   /**
    * Initializes primitive panel
    */
   private void initPrimitivePanel(){
      this.primitivePanel   = new IconPanel( dragController );
      
      PrimitiveWidget with    = PrimitiveFactory.createWithWidget   ();
      PrimitiveWidget without = PrimitiveFactory.createWithoutWidget();
      PrimitiveWidget near    = PrimitiveFactory.createInsideWidget ();
      PrimitiveWidget far     = PrimitiveFactory.createOutsideWidget();
      PrimitiveWidget inside  = PrimitiveFactory.createNearWidget   ();
      PrimitiveWidget outside = PrimitiveFactory.createFarWidget    ();
      PrimitiveWidget lasts   = PrimitiveFactory.createLastsWidget  ();

      with   .setLabelVisible( false );
      without.setLabelVisible( false );
      inside .setLabelVisible( false );
      outside.setLabelVisible( false );
      near   .setLabelVisible( false );
      far    .setLabelVisible( false );
      lasts  .setLabelVisible( false );
      
      this.primitivePanel.add( with    );
      this.primitivePanel.add( without );
      this.primitivePanel.add( inside  );
      this.primitivePanel.add( outside );
      this.primitivePanel.add( near    );
      this.primitivePanel.add( far     );
      //this.primitivePanel.add( lasts   );
      
      this.primitivePanel.setSize( ScenicToolbar.PRIMITIVE_PANEL_WIDTH, ScenicToolbar.PRIMITIVE_PANEL_HEIGHT );
   }
   
   
   /**
    * Initializes time constraints panel
    */
   private void initTimeConstraintsPanel(){
      this.timingPanel = new TimeConstraintsPanel();
      this.timingPanel.setSize( ScenicToolbar.TIMING_PANEL_WIDTH, ScenicToolbar.TIMING_PANEL_HEIGHT );
      
   }
   
   
   /**
    * Initializes sequence viewer panel
    * @param sequencePanel
    */
   private void initSequenceViewer(SequencePanel sequencePanel){
      if(null != this.sequenceViewer){
         this.remove(this.sequenceViewer);
      }
      this.sequencePanel  = sequencePanel;
      this.sequenceViewer = new SequenceViewer(this.sequencePanel);
      Window.addResizeHandler(this.sequencePanel);
      //ScenicToolbar.disableTextSelectInternal( this.sequencePanel.getElement() , true );
      //ScenicToolbar.disableTextSelectInternal( this.sequenceViewer.getElement(), true );
      this.insert(this.sequenceViewer,1);
   }
   
   
   /**
    * Initializes button panel - a place holder for now.
    */
   private void initButtonPanel(){
      this.buttonPanel = new HorizontalPanel();
      this.buttonPanel.setWidth("100%");
      
      HorizontalPanel buttonHolder = new HorizontalPanel();
      
      this.buttonPanel.add(buttonHolder);
      this.buttonPanel.setCellHorizontalAlignment(buttonHolder, HasHorizontalAlignment.ALIGN_CENTER);
      
      this.add(this.buttonPanel);
   }
   
   
   /**
    * Initializes template panel - used only in the formative study.
    */
   private void initTemplatePanel(){
      this.curTemplateNum       = 1;
      this.templatePanel        = new VerticalPanel();
      
      VerticalPanel templateHolderPanel = new VerticalPanel();
      templateHolderPanel.setWidth("75%");
      
      this.templateTitle        = new HTML();
      this.templateDescription  = new HTML();
      this.templateExampleTitle = new HTML();
      this.templateApplication  = new HTML();
      this.templateScenario     = new HTML();
      
      DOM.setStyleAttribute(this.templatePanel.getElement(), "backgroundColor", "#FFFFFF");
      this.templatePanel.setWidth( "100%" );
      
      templateHolderPanel.add( new HTML("&nbsp;")        );
      templateHolderPanel.add( this.templateTitle        );
      templateHolderPanel.add( this.templateDescription  );
      templateHolderPanel.add( this.templateExampleTitle );
      templateHolderPanel.add( this.templateApplication  );
      templateHolderPanel.add( this.templateScenario     );
      
      this.templatePanel.add(templateHolderPanel);
      this.templatePanel.setCellHorizontalAlignment(templateHolderPanel, HorizontalPanel.ALIGN_CENTER);
   }
   
   
   /**
    * Initializes task panel - used only in the formative study
    */
   private void initTaskPanel(){
      this.curTaskNum = 1;
      this.taskPanel  = new VerticalPanel();
      
      VerticalPanel taskHolderPanel = new VerticalPanel();
      taskHolderPanel.setWidth("75%");
      
      this.taskNumber      = new HTML();
      this.taskApplication = new HTML();
      this.taskScenario    = new HTML();
      
      DOM.setStyleAttribute( this.taskPanel.getElement(), "backgroundColor", "#FFFFFF");
      this.taskPanel.setWidth( "100%" );
      
      taskHolderPanel.add( new HTML("&nbsp;")   );
      taskHolderPanel.add( this.taskNumber      );
      taskHolderPanel.add( this.taskApplication );
      taskHolderPanel.add( this.taskScenario    );
      
      this.taskPanel.add(taskHolderPanel);
      this.taskPanel.setCellHorizontalAlignment(taskHolderPanel, HorizontalPanel.ALIGN_CENTER);
   }
   
   
   /**
    * A debug function that displays the thumbnails for all sub-sequences of an event.
    * @param sequence
    */
   private void displayThumbnails(ScenicSequence sequence){
      // Now generate the thumbnail for each sub-event
      List<SequencePanel>  thumbnails = new ArrayList<SequencePanel>(); 
      SequencePanel        thumbnail  = SequenceFactory.createThumbnailSequencePanel(sequence);
      thumbnails.add(thumbnail);
      List<ScenicSequence> subEvents  = GrammarEncoder.getSubEvents(sequence);
      for(ScenicSequence subEvent : subEvents){
         thumbnail = SequenceFactory.createThumbnailSequencePanel(subEvent);
         thumbnails.add(thumbnail);
      }
      for(SequencePanel panel : thumbnails){
         this.addThumbnail(panel);
      }
   }
}