package edu.washington.cs.rfid.cascadia.timeline.client.event;

import com.google.gwt.user.client.rpc.IsSerializable;

/**
 * An event object for the ScenicTimeline.
 * @author evan
 *
 * Documentation from simile-widgets wiki on events: 
 * (http://code.google.com/p/simile-widgets/wiki/Timeline_EventSources)
 * 
 * Basic Event Attributes:
 *----------------------------------------------------------------------------------
 * start         - in full date format (e.g. "May 20 1961 00:00:00 GMT-0600"). 
 * latestStart   - for imprecise beginnings - same date format as start
 * earliestEnd   - for imprecise ends - same date format as start
 * end           - same date format as start
 * durationEvent - XML and SPARQL: "true" or "false". JSON: true or false. 
 *                 Only applies to events with start/end times.
 *   o true -- the event occurs over a time duration. No icon. The event 
 *             will be drawn as a dark blue tape. The tape color is set 
 *             with the color attribute. Default color is #58A0DC
 *   o false -- the event is focused on a specific "instant" (shown with the icon). 
 *              The event will be drawn as a blue dot icon (default) with a pale 
 *              blue tape. The tape is the default color (or color attribute color), 
 *              with opacity set to 20. To change the opacity, change the theme's 
 *              instant: {impreciseOpacity: 20} value. Maximum 100. 
 * title - text title that goes next to the tape in the Timeline. Also shown in the 
 *         bubble. The title attribute is optional. Leave it out if you want just an 
 *         icon or icon and tape. The description will be shown in the bubble when 
 *         the icon or tape is clicked. 
 *----------------------------------------------------------------------------------
 * Additional Event Attributes:
 * 
 * icon        - url. This image will appear next to the title text in the timeline 
 *               if (no end date) or (durationEvent = false). If a start and end 
 *               date are supplied, and durationEvent is true, the icon is not shown. 
 *               If icon attribute is not set, a default icon from the theme is used.
 * image       - url to an image that will be displayed in the bubble
 * link        - url. The bubble's title text be a hyper-link to this address.
 * color       - color of the text and tape (duration events) to display in the timeline. 
 *               If the event has durationEvent = false, then the bar's opacity will be 
 *               applied (default 20%). See durationEvent, above.
 * textColor   - color of the label text on the timeline. If not set, then the color 
 *               attribute will be used.
 * tapeImage and tapeRepeat - Sets the background image and repeat style for the event's 
 *                            tape (or 'bar') on the Timeline. Overrides the color setting 
 *                            for the tape. Repeat style should be one of 
 *                            {repeat | repeat-x | repeat-y}, repeat is the default. See 
 *                            the Cubism example for a demonstration. Only applies to duration events.
 * caption     - additional event information shown when mouse is hovered over the Timeline 
 *               tape or label. Uses the html title property. Looks like a tooltip. Plain 
 *               text only. See the cubism example.
 * classname   - added to the HTML classnames for the event's label and tape divs. Eg 
 *               classname attribute 'hot_event' will result in div classes of 
 *               'timeline-event-label hot_event' and 'timeline-event-tape hot_event' for the 
 *               event's Timeline label and tape, respectively.
 * description - will be displayed inside the bubble with the event's title and image.
 *   o XML Format: the description is stored as the text content of the event element 
 *                 (see below). Note: the XML standard requires that an element's text content 
 *                 must be escaped/formatted HTML.
 *   o JSON Format: the description key of the event hash 
 *   
 *----------------------------------------------------------------------------------
 * Event Attributes for Developers:
 * 
 * trackNum - used to override the automatic layout of events on the Timeline.
 * eventID  - a cookie attribute that is stored, not used by the Timeline library. 
 *            If you write a custom labeller or event bubble filler, the attribute can 
 *            be obtained using the getEventID() method on the event object. 
 */
public class ScenicTimelineEvent implements IsSerializable {

   public String  start;
   public String  latestStart;
   public String  end;
   public String  earliestEnd;
   public String  durationEvent;
   public String  color;
   public String  textColor;
   public String  caption;
   public String  title;
   public String  link;
   public String  description;
   public String  icon;
   public String  image;
   public String  html;
   
   public int     trackNum = -1;
   public String  eventId;
   
   //--//
   
   public ScenicTimelineEvent(){
      
   }
   
   /**
    * Converts this ScenicTimelineEvent to XML.
    * @return
    */
   public String toXml(){
      return this.toXmlBuilder().toString();
   }
   
   
   /**
    * Converts this ScenicTimelineEvent to XML.
    * @return
    */
   public StringBuilder toXmlBuilder(){
      StringBuilder xmlDuration = this.toXmlBuilder(true );
      StringBuilder xmlInstant  = this.toXmlBuilder(false);
      
      return xmlDuration.append(xmlInstant.toString());
   }
   
   
   /**
    * Helper function to generate XML for an event with the specified duration status.
    * @param duration
    * @return
    */
   protected StringBuilder toXmlBuilder(boolean duration){
      StringBuilder xml = new StringBuilder();
      
      xml.append("<event ");
      
      if(null != start){
         xml.append("start=\"");
         xml.append(start);
         xml.append("\" ");
      }
      if(null != latestStart){
         xml.append("latestStart=\"");
         xml.append(latestStart);
         xml.append("\" ");
      }
      if(null != end){
         xml.append("end=\"");
         xml.append(end);
         xml.append("\" ");
      }
      if(null != earliestEnd){
         xml.append("earliestEnd=\"");
         xml.append(earliestEnd);
         xml.append("\" ");
      }
      if(duration){
         xml.append("durationEvent=\"true\" ");
      } else {
         xml.append("durationEvent=\"false\" ");
      }
      if(null != color){
         xml.append("color=\"");
         xml.append(color);
         xml.append("\" ");
      }
      if(null != textColor){
         xml.append("textColor=\"");
         xml.append(textColor);
         xml.append("\" ");
      }
      if(null != caption){
         xml.append("caption=\"");
         xml.append(caption);
         xml.append("\" ");
      }
      if(null != title && !duration){
         xml.append("title=\"");
         xml.append(title);
         xml.append("\" ");
      }
      if(null != link){
         xml.append("link=\"");
         xml.append(link);
         xml.append("\" ");
      }
      if(null != description){
         xml.append("description=\"");
         xml.append(description);
         xml.append("\" ");
      }
      if(null != icon){
         xml.append("icon=\"../images/event-marker-sm.png\" ");
//         xml.append("icon=\"");
//         xml.append(icon);
//         xml.append("\" ");
      }
      if(null != image){
         xml.append("image=\"");
         xml.append(image);
         xml.append("\" ");
      }
      if(0 < trackNum){
         xml.append("trackNum=\"");
         xml.append(trackNum);
         xml.append("\" ");
      }
      if(null != eventId){
         xml.append("eventID=\"");
         xml.append(eventId);
         xml.append("\" ");
      }
      xml.append(">");
      if(null != html){
         xml.append(html);
      }
      xml.append("</event>");
      
      return xml;
   }
}
