package edu.washington.cs.rfid.cascadia.timeline.client.ui;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import com.google.gwt.user.client.Command;
import com.google.gwt.user.client.DeferredCommand;
import com.google.gwt.user.client.ui.AbsolutePanel;
import com.google.gwt.user.client.ui.HTML;
import com.google.gwt.user.client.ui.HorizontalPanel;
import com.google.gwt.user.client.ui.Image;
import com.google.gwt.user.client.ui.VerticalPanel;

import edu.washington.cs.rfid.cascadia.scenic.client.ui.SequencePanel;
import edu.washington.cs.rfid.cascadia.timeline.client.event.ScenicEventResult;
import edu.washington.cs.rfid.cascadia.timeline.client.event.ScenicEventResultBundle;
import edu.washington.cs.rfid.cascadia.timeline.client.event.ScenicTimelineEvent;
import edu.washington.cs.rfid.cascadia.timeline.client.filter.IScenicFilterMatcher;
import edu.washington.cs.rfid.cascadia.timeline.client.filter.NoStreamFilter;


public class ScenicTimelinePanel extends VerticalPanel {

   protected static String STYLE_OVERLAY_LABEL       = "gwt-HTML-big-label";
   protected static String STYLE_TRANSPARENT_OVERLAY = "timeline-overlay";
   protected static String DEFAULT_WIDTH_TIMELINE    = "700px";//"800px";
   protected static String DEFAULT_HEIGHT_TIMELINE   = "540px";
   
   //--//
   
   protected static ScenicTimelinePanel instance;
   
   //--//
   
   
   /**
    * Retrieves the static singleton instance, whether it has been initialized or not.
    */
   public static ScenicTimelinePanel getInstance(){
      return ScenicTimelinePanel.instance;
   }
   
   
   /**
    * Retrieves and possibly initializes the static singleton instance.
    */
   public static ScenicTimelinePanel getInstance(ScenicEventResultBundle bundle){
      if(null == ScenicTimelinePanel.instance){
         ScenicTimelinePanel.instance = 
            new ScenicTimelinePanel(ScenicTimelinePanel.DEFAULT_WIDTH_TIMELINE,
                                      ScenicTimelinePanel.DEFAULT_HEIGHT_TIMELINE,
                                      bundle);
      } else {
         ScenicTimelinePanel.instance = 
            new ScenicTimelinePanel(ScenicTimelinePanel.DEFAULT_WIDTH_TIMELINE,
                                      ScenicTimelinePanel.DEFAULT_HEIGHT_TIMELINE,
                                      bundle);
      }
      return ScenicTimelinePanel.instance;
   }
   
   
   //--//
   
   
   protected ScenicTimeline                                 timeline;
   protected AbsolutePanel                                  timelineThumb;
   protected AbsolutePanel                                  timelineBar;
   protected AbsolutePanel                                  timelinePanel;
   
   protected ScenicEventResultBundle                        eventBundle;
   protected HashMap<String,ArrayList<ScenicTimelineEvent>> keyToStreamMap;
   protected List<ScenicTimelineEventLabel>                 eventLabels;
   
   
   //--//
   
   
   /**
    * Creates a new ScenicTimelinePanel to hold a timeline and all its accompanying
    * interface components (e.g., playback thumb, highlighting and filtering controls)
    * as well as to initialize its contents (e.g., band labels, events).
    */
   public ScenicTimelinePanel(String width, String height, ScenicEventResultBundle eventBundle){

      this.initEvents(eventBundle);
      
      HTML timelineTitle = new HTML("Event Timelime");
      timelineTitle.setStyleName("gwt-HTML-big-label");
      
      this.timeline      = ScenicTimeline.getInstance(height,width,eventBundle);
      
      this.timelineThumb = new AbsolutePanel();
      this.timelineThumb.setStyleName("timeline-thumb");
      this.timelineThumb.add(new Image("../images/timeline-thumb.png"));
      this.timelineThumb.setPixelSize(19, 10);
      
      this.timelineBar   = new AbsolutePanel();
      this.timelineBar.setStyleName("timeline-bar");
      this.timelineBar.setPixelSize(1, this.timeline.height);
      
      this.timelinePanel = new AbsolutePanel();
      this.timelinePanel.setPixelSize(this.timeline.width, this.timeline.height);
      this.timelinePanel.add(this.timeline);
      this.timelinePanel.add(this.timelineThumb,(this.timeline.width/2) - 9,0);
      this.timelinePanel.add(this.timelineBar,this.timeline.width/2,0);
      
      this.add(timelineTitle);
      this.setCellHorizontalAlignment(timelineTitle, HorizontalPanel.ALIGN_CENTER);
      this.add(this.timelinePanel);
      
      this.hidePlaybackThumb();
      this.timeline.addBubbleFiller();
      
      DeferredCommand.addCommand(new AddBandLabelsCmd());
   }
   
   
   private void initEvents(ScenicEventResultBundle eventBundle){
      this.eventBundle    = eventBundle;
      this.keyToStreamMap = eventBundle.keyToStreamMap;
      this.eventLabels    = new ArrayList<ScenicTimelineEventLabel>();
   }
   
   
   /**
    * Loads all events into the timeline.
    */
   public void loadTimeline(){
      this.loadAllEventStreams();
   }
   
   
   /**
    * Gets the absolute X offset of the timeline in the timeline panel.
    * @return
    */
   public int getTimelineX(){
      return this.timelinePanel.getWidgetLeft(this.timeline);
   }
   
   
   /**
    * Gets the absolute Y offset of the timeline in the timeline panel.
    * @return
    */
   public int getTimelineY(){
      return this.timelinePanel.getWidgetTop(this.timeline);
   }
   
   
   /**
    * Gets an event by that event's ID.
    * @param eventId
    * @return
    */
   public ScenicTimelineEventLabel getEventLabelById(String eventId){
      ScenicTimelineEventLabel label = null;
      
      for(ScenicTimelineEventLabel l : this.eventLabels){
         List<ScenicTimelineEvent> events = l.keyToStreamMap.get(l.streamKey);
         for(ScenicTimelineEvent e : events){
            if(e.eventId.equals(eventId)){
               label = l;
               break;
            }
         }
      }
      return label;
   }
   
   
   /**
    * Sets the track parameter for all events in a given stream.
    * @param stream
    * @param track
    */
   private void setEventTrack(List<ScenicTimelineEvent> stream, int track){
      for(ScenicTimelineEvent e : stream){
         e.trackNum = track;
      }
   }
   
   
   /**
    * Adds a new label overlay to the timeline at the specified x and y coordinates.
    * @param text
    * @param x
    * @param y
    */
   public void addBandLabel(String text, int x, int y, String color){
      HTML eventLabel = new HTML(text);
      eventLabel.addStyleName(ScenicTimelinePanel.STYLE_OVERLAY_LABEL);
      eventLabel.addStyleName(ScenicTimelinePanel.STYLE_TRANSPARENT_OVERLAY);
      eventLabel.addStyleName(color);
      
      this.timelinePanel.add(eventLabel,x,y);    
   }
   
   
   public void addEventLabel(ScenicTimelineEventLabel label, int x, int y){
      this.eventLabels.add(label);
      this.timelinePanel.add(label,x,y);
   }
   
   
   /**
    * Cues playback to the given date
    * @param start
    */
   public void cuePlayback(String start){
//      this.clientProxy.getTracePack(this.curTracePackNum, tracePackCallback);
//      this.clientProxy.getTracePack(this.curTracePackNum, start, tracePackCallback);
   }
   
   
   /**
    * Shows the playback thumb
    */
   public void showPlaybackThumb(){
      this.timelineBar.setVisible(true);
      this.timelineThumb.setVisible(true);
   }
   
   
   /**
    * Hides playback thumb
    */
   public void hidePlaybackThumb(){
      this.timelineBar.setVisible(false);
      this.timelineThumb.setVisible(false);
   }
   
   
   
   //------------------------------ COMMANDS --------------------------------//
   
   
   /**
    * A command to add the timeline band labels, all start as selected with
    * filters added.
    * @author evan
    */
   private class AddBandLabelsCmd implements Command {
      public void execute(){
         List<ScenicEventResult> events      = timeline.getBandEvents();
         List<SequencePanel>     bandThumbs  = timeline.getBandThumbnails();
         int[]                   bandOffsets = timeline.getBandOffsets();
         String[]                bandLabels  = timeline.getBandLabels();
         String[]                bandColors  = timeline.getBandColors();
         int                     xBase       = getTimelineX();
         int                     yBase       = getTimelineY();
         int                     xBaseOffset = 10;
         int                     yBaseOffset = 10;
         
         for(int i = 0; i < bandOffsets.length; i++){
            
            addBandLabel(bandLabels[i]                       ,
                         xBase + xBaseOffset                 ,
                         yBase + yBaseOffset + bandOffsets[i],
                         bandColors[i]);
            
            if(i < bandOffsets.length - 2){
               int                      track      = 1;
               ScenicEventResult        event      = events.get(i);
               ScenicTimelineEventLabel eventLabel = new ScenicTimelineEventLabel(timeline         ,
                                                                                  bandThumbs.get(i),
                                                                                  event.streamKey  ,
                                                                                  keyToStreamMap   ,
                                                                                  event.title      ,
                                                                                  event.description,
                                                                                  i                ,
                                                                                  bandColors[i]);
               eventLabels.add(eventLabel); // Add the label to the list but not the panel
               
               // THUMBNAIL DEBUG BELOW
               timelinePanel.add(bandThumbs.get(i),-1000,-1000);
               
               setEventTrack(keyToStreamMap.get(event.streamKey), track++);
               
               List<ScenicEventResult> subevents       = timeline.getSubEvents(i);
               List<SequencePanel>     subThumbs       = timeline.getSubEventThumbnails(i);
               int[]                   subEventOffsets = timeline.getSubEventOffsets(i);
               SequencePanel subThumb = null;
               
               for(int j = 0; j < subevents.size(); j++){
                  ScenicEventResult        subEvent   = subevents.get(j);
                  if(j < subThumbs.size()){
                     subThumb = subThumbs.get(j);
                  }
                  eventLabel = new ScenicTimelineEventLabel(timeline            ,
                                                            subThumb            ,
                                                            subEvent.streamKey  ,
                                                            keyToStreamMap      ,
                                                            subEvent.title      ,
                                                            subEvent.description,
                                                            i                   ,
                                                            bandColors[i]);
                  
                  // DEBUG THUMBNAIL BELOW
                  timelinePanel.add(subThumb,-1000,-1000);
                  
                  setEventTrack(keyToStreamMap.get(subEvent.streamKey), track++);
                  addEventLabel(eventLabel, xBase + xBaseOffset, yBase + yBaseOffset + subEventOffsets[j]);                  
               }
            }
         }
         loadTimeline();
      }
   }
   
   
   /**
    * Loads all event streams in the keyToStreamMap into the timeline.
    */
   public void loadAllEventStreams(){
      for(String key : this.keyToStreamMap.keySet()){
         List<ScenicTimelineEvent> stream = this.keyToStreamMap.get(key);
         for(ScenicTimelineEvent e : stream){
            this.timeline.addEvent(e);
         }
      }
      this.timeline.loadAllEvents();
      
      // Also load filters with a deferred command
      DeferredCommand.addCommand(new AddDefaultFilterCmd());
      
      // TODO: Set Center Time here, dynamically!
   }
   
   
   /**
    * A command class to add the default filters and check the first tree item.
    */
   private class AddDefaultFilterCmd implements Command {
      
      public void execute(){
         List<ScenicEventResult> scenes = eventBundle.result.subEvents;
         
         if(null != scenes){
            int numEventBands = 1 + scenes.size();

            // Add default filter to all event bands
            IScenicFilterMatcher noStreamFilter = new NoStreamFilter();
            for(int i = 0; i < numEventBands; i++){
               timeline.addMatchingFilter(noStreamFilter,i);
            }
         }
         // Now add the filter for each event label and "check" that label
         for(ScenicTimelineEventLabel label : eventLabels){
            label.setValueNoRepaint(true);
         }
         timeline.repaintTimeline();
      }
   }
}
