package edu.washington.cs.rfid.cascadia.timeline.client.ui;

import java.util.ArrayList;
import java.util.List;

import com.google.gwt.core.client.JavaScriptObject;
import com.netthreads.gwt.simile.timeline.client.BandInfo;
import com.netthreads.gwt.simile.timeline.client.BandOptions;
import com.netthreads.gwt.simile.timeline.client.DateTime;
import com.netthreads.gwt.simile.timeline.client.EventSource;
import com.netthreads.gwt.simile.timeline.client.ITimeLineRender;
import com.netthreads.gwt.simile.timeline.client.Theme;
import com.netthreads.gwt.simile.timeline.client.TimeLineWidget;

import edu.washington.cs.rfid.cascadia.timeline.client.event.ScenicEventResult;
import edu.washington.cs.rfid.cascadia.timeline.client.event.ScenicEventResultBundle;

/**
 * A Scenic timeline renderer.
 * @author evan
 */
public class ScenicTimelineRender implements ITimeLineRender {

   protected static final int TRACK_HEIGHT_PX             = 32;
   protected static final int BAND_BASE_HEIGHT_PX         = 25;
   protected static final int BAND_LABEL_HEIGHT_PX        = 32;
   protected static final int BAND_LABEL_SPACER_HEIGHT_PX = 5;
   protected static final int EVENT_LABEL_HEIGHT_PX       = 15;
   protected static final int HOUR_SUMMARY_HEIGHT_PX      = 65;
   protected static final int DAY_SUMMARY_HEIGHT_PX       = 65;
   
   
   // Note: these styles can be found in the js/api/styles/ethers.js file
   protected static final String[] BAND_COLORS = new String[] {
      "timeline-band-color-0",
      "timeline-band-color-1",
      "timeline-band-color-2",
      "timeline-band-color-3"
   };
   
   //--//
   
   protected ScenicEventResultBundle bundle;
   protected int                     width;
   protected int                     height;
   protected int                     numSceneBands;
   protected int[]                   bandHeightsPx;
   protected int[]                   bandHeightsPercent;
   
   //--//
   
   /**
    * Creates a new ScenicTimelineRender.
    * @param width
    * @param height
    */
   public ScenicTimelineRender(int width, int height, ScenicEventResultBundle bundle){
      this.bundle = bundle;
      this.width  = width;
      this.initHeights();
   }
   
   
   /**
    * Renders the timeline.
    * @param widget
    */
   public void render(TimeLineWidget widget) {
      
      // Fix timeline size with dynamically set height
      widget.setPixelSize(this.width, this.height);
      
      // Set Theme parameters
      Theme theme = widget.getTheme();
      theme.setEventLabelWidth(400);
      ScenicTimelineRender.setMouseWheelZoom(theme);
      ScenicTimelineRender.setImpreciseOpacity(theme,100);
      
      // Get the event band options
      EventSource eventSource    = widget.getEventSource();
      ArrayList   bandHotZones   = widget.getBandHotZones();
      ArrayList   bandDecorators = widget.getBandDecorators();
      ArrayList   bandInfos      = widget.getBandInfos();

      this.setSceneBandOptions(theme, 
                               eventSource, 
                               bandHotZones, 
                               bandDecorators, 
                               bandInfos,
                               this.bundle.startDate);
   }
   
   
   /**
    * Returns a copy of the array of band colors.
    * @return
    */
   public String[] getBandColors(){
      
      int      numBands = 1 + this.numSceneBands + 1 + 1;
      String[] colors   = new String[numBands];
      int      colorIdx = 0;
      
      for(int i = 0; i < numBands; i++){
         colors[i] = ScenicTimelineRender.BAND_COLORS[colorIdx];
         colorIdx  = (colorIdx + 1) % ScenicTimelineRender.BAND_COLORS.length;
      }
      return colors;
   }
   
   
   /**
    * Gets the offsets for the top of each band in pixels.
    * @return
    */
   public int[] getBandOffsets(){
      int[] offsets = new int[this.bandHeightsPx.length + 1 + 1];
      offsets[0]    = 0;
      for(int i = 1; i < offsets.length - 1; i++){
         offsets[i] = this.bandHeightsPx[i-1] + offsets[i - 1];
      }
      offsets[offsets.length - 1] = ScenicTimelineRender.HOUR_SUMMARY_HEIGHT_PX + offsets[offsets.length - 2];
      return offsets;
   }
   
   
   /**
    * Gets the label for each event band.
    * @return
    */
   public String[] getBandLabels(){
      String[] bandLabels = new String[1 + this.numSceneBands + 1 + 1];
      
      bandLabels[0] = "EVENT";
      for(int i = 1; i < 1 + this.numSceneBands; i++){
         bandLabels[i] = "SCENE " + i;
      }
//      bandLabels[1 + this.numSceneBands] = "SUMMARY";
      bandLabels[this.numSceneBands + 1] = "HOURS";
      bandLabels[this.numSceneBands + 1 + 1] = "DAYS";
      
      return bandLabels;
   }
   
   
   /**
    * Gets the main band events for the timeline.
    * @param bandNum
    * @return
    */
   public List<ScenicEventResult> getBandEvents(){
      List<ScenicEventResult> subEvents = new ArrayList<ScenicEventResult>();      
      subEvents.add(this.bundle.result);
      subEvents.addAll(this.bundle.result.subEvents);
      return subEvents;
   }
   
   
   /**
    * Gets the offsets for the top of each sub-event checkbox in pixels for a
    * given band.
    * @param bandNum
    * @return
    */
   public int[] getSubEventOffsets(int bandNum){
      int[]                   offsets = null;
      List<ScenicEventResult> scenes  = this.bundle.result.getScenes();
      if(null != scenes){
         int[] bandOffsets = this.getBandOffsets();
         int   offsetBase  = bandOffsets[bandNum]                      + 
                             ScenicTimelineRender.BAND_LABEL_HEIGHT_PX + 
                             ScenicTimelineRender.BAND_LABEL_SPACER_HEIGHT_PX;
         if(0 == bandNum){
            offsets = new int[scenes.size()];
         } else if(0 < bandNum && bandNum < scenes.size() + 1){
            ScenicEventResult       r          = scenes.get(bandNum-1);
            List<ScenicEventResult> primitives = r.subEvents;
            if(null != primitives){
               offsets = new int[primitives.size()];
            }
         }
         if(null != offsets){
            for(int i = 0; i < offsets.length; i++){
               offsets[i]  = offsetBase;
               offsetBase += ScenicTimelineRender.BAND_LABEL_SPACER_HEIGHT_PX + 
                             ScenicTimelineRender.EVENT_LABEL_HEIGHT_PX;
            }
         }
      }
      return offsets;
   }
   
   
   /**
    * Gets labels for the sub events in the specified band
    * @param bandNum
    * @return
    */
   public String[] getSubEventLabels(int bandNum){
      String[]                labels = null;
      List<ScenicEventResult> scenes = this.bundle.result.getScenes();
      if(null != scenes){
         if(0 == bandNum){
            labels          = new String[scenes.size()];
            String sceneStr = "1";
            for(int i = 0; i < labels.length; i++){
               labels[i] = "Scenes " + sceneStr;
               sceneStr += "," + (i+2);
            }
         } else if(0 < bandNum && bandNum < scenes.size() + 1){
            ScenicEventResult       r          = scenes.get(bandNum-1);
            List<ScenicEventResult> primitives = r.subEvents;
            if(null != primitives){
               labels = new String[primitives.size()];
               for(int i = 0; i < labels.length; i++){
                  labels[i] = primitives.get(i).title;
               }
            }
         }
      }
      return labels;
   }
   
   
   /**
    * Gets the sub events for the specified band.
    * @param bandNum
    * @return
    */
   public List<ScenicEventResult> getSubEvents(int bandNum){
      List<ScenicEventResult> subEvents = new ArrayList<ScenicEventResult>();
      
      if(0 == bandNum){
         subEvents = this.bundle.result.getSubSequences();
      } else {
         List<ScenicEventResult> scenes = this.bundle.result.getScenes();
         if(null != scenes && 0 < bandNum && bandNum < scenes.size() + 1){
            subEvents = scenes.get(bandNum-1).subEvents;
         }
      }
      return subEvents;
   }
   
   
   //--//
   
   
   /**
    * Sets the options for the timeline using a separate band for each
    * scene.
    * @param theme
    * @param eventSource
    * @param bandHotZones
    * @param bandDecorators
    * @param bandInfos
    */
   private void setSceneBandOptions(Theme       theme         ,
                                    EventSource eventSource   ,
                                    ArrayList   bandHotZones  ,
                                    ArrayList   bandDecorators,
                                    ArrayList   bandInfos     ,
                                    String      centerDate){
      
      // 1) Event & Sub-sequence band
      BandOptions opts = BandOptions.create();
      opts.setWidth(this.bandHeightsPercent[0] + "%");
      opts.setTrackHeight(1.3f);
      opts.setTrackGap(0.2f);
      opts.setIntervalUnit(DateTime.MINUTE());//DateTime.HOUR());//DateTime.MINUTE());//
      opts.setIntervalPixels(100);//50);//
      opts.setShowEventText(false);
      opts.setTheme(theme);
      opts.setEventSource(eventSource);
      opts.setDate(centerDate);
      this.setBandZoomIndex(opts, 2);
      this.setBandZoomSteps(opts, this.getZoomStepArray());

      BandInfo fineBand = BandInfo.create(opts);
      fineBand.setDecorators(bandDecorators);
      bandInfos.add(fineBand);
      
      // 2) Scene bands
      for(int i = 0; i < this.numSceneBands; i++){
         opts = BandOptions.create();
         opts.setWidth(this.bandHeightsPercent[i + 1] + "%");
         opts.setTrackHeight(1.3f); 
         opts.setTrackGap(0.2f);
         opts.setIntervalUnit(DateTime.MINUTE()); // DateTime.HOUR());//DateTime.MINUTE());//
         opts.setIntervalPixels(100);//50);
         opts.setShowEventText(false);
         opts.setTheme(theme);
         opts.setEventSource(eventSource);
         opts.setDate(centerDate);
         this.setBandZoomIndex(opts, 2);
         this.setBandZoomSteps(opts, this.getZoomStepArray());

         fineBand = BandInfo.create(opts);
         fineBand.setDecorators(bandDecorators);
         bandInfos.add(fineBand);
         fineBand.setSyncWith(0);
      }
      
      // 5) HOUR Summary band (coarser-grain: Hour)
      int sumPriorHeights = 0;
      for(int percent : this.bandHeightsPercent){
         sumPriorHeights += percent;
      }
      int summaryHeight = this.getPercentHeightBand(ScenicTimelineRender.HOUR_SUMMARY_HEIGHT_PX);// 100 - sumPriorHeights;
      
      opts = BandOptions.create();
      opts.setWidth(summaryHeight + "%");
      opts.setTrackHeight(1.3f);
      opts.setTrackGap(0.1f);
      opts.setIntervalUnit(DateTime.HOUR());
      opts.setIntervalPixels(320);
      opts.setShowEventText(false);
      opts.setTheme(theme);
      opts.setEventSource(eventSource);
      opts.setDate(centerDate);
      opts.setZones(bandHotZones);
      opts.setTimeZone(0); // TODO: should this be -8??
      this.setBandOverview(opts);

      BandInfo summaryBand = BandInfo.create(opts);
      summaryBand.setDecorators(bandDecorators);
      bandInfos.add(summaryBand);
      summaryBand.setSyncWith(0);
      summaryBand.setHighlight(true);
      

      // 5) DAY Summary band (coarsest-grain: Day)
      int daySumHeight = this.getPercentHeightBand(ScenicTimelineRender.DAY_SUMMARY_HEIGHT_PX);// 100 - sumPriorHeights;
      
      opts = BandOptions.create();
      opts.setWidth(daySumHeight + "%");
      opts.setTrackHeight(1.3f);
      opts.setTrackGap(0.1f);
      opts.setIntervalUnit(DateTime.DAY());
      opts.setIntervalPixels(320);
      opts.setShowEventText(false);
      opts.setTheme(theme);
      opts.setEventSource(eventSource);
      opts.setDate(centerDate);
      opts.setZones(bandHotZones);
      opts.setTimeZone(0);// TODO: should this be -8??
      this.setBandOverview(opts);

      BandInfo daySumBand = BandInfo.create(opts);
      daySumBand.setDecorators(bandDecorators);
      bandInfos.add(daySumBand);
      daySumBand.setSyncWith(0);
      daySumBand.setHighlight(true);

      
   }
   
   
   /**
    * Sets the height of each band in pixels and as a percent.
    * @return
    */
   private void initHeights(){
      this.height                    = 0;
      List<ScenicEventResult> scenes = this.bundle.result.getScenes();
      
      if(null != scenes){
         this.numSceneBands = scenes.size();
         int numEventBands  = 1 + this.numSceneBands;
         int totalEvents    = 1 + this.numSceneBands;
         this.bandHeightsPx = new int[numEventBands];
         
         // First band is event + subsequences
         int numEvents         = 1 + this.numSceneBands;
         this.bandHeightsPx[0] = this.getPixelHeightBand(numEvents);
         
         // Other band heights are Scene + primitives
         for(int i = 0; i < scenes.size(); i++){
            ScenicEventResult r     = scenes.get(i);
            numEvents               = r.getTreeSize();
            totalEvents            += numEvents;
            this.bandHeightsPx[i+1] = this.getPixelHeightBand(numEvents);
         }
         this.height = totalEvents   * ScenicTimelineRender.TRACK_HEIGHT_PX        + 
                       numEventBands * ScenicTimelineRender.BAND_BASE_HEIGHT_PX    + 
                                       ScenicTimelineRender.HOUR_SUMMARY_HEIGHT_PX + 
                                       ScenicTimelineRender.DAY_SUMMARY_HEIGHT_PX;
         
         this.bandHeightsPercent    = new int[numEventBands];
         this.bandHeightsPercent[0] = this.getPercentHeightBand(this.bandHeightsPx[0]); 
         
         for(int i = 0; i < scenes.size(); i++){
            this.bandHeightsPercent[i+1] = this.getPercentHeightBand(this.bandHeightsPx[i+1]);
         }
      }
   }
   
   
   /**
    * Computes the height in pixels of a particular band as a function of 
    * the number of events it contains.
    */
   private int getPixelHeightBand(int numEvents){
      return numEvents * ScenicTimelineRender.TRACK_HEIGHT_PX + 
                         ScenicTimelineRender.BAND_BASE_HEIGHT_PX;
   }
   
   
   /**
    * Computes the height of a particular band as a percentage of the total
    * timeline height. 
    * @param numEvents
    * @param timelineHeightPx
    * @return
    */
   private int getPercentHeightBand(int bandHeightPx){
      return (int) Math.round(((double)bandHeightPx / (double)this.height) * 100.0);
   }
   
   
   //----------------------------------- NATIVE JS AND JSNI ------------------------------------
   
   
   /**
    * Gets a default zoom step array.
    * @return
    */
   private JavaScriptObject getZoomStepArray(){
      JavaScriptObject[] stepArray = new JavaScriptObject[10];
      
      stepArray[0] = ZoomStep.create(400,DateTime.MINUTE());
      stepArray[1] = ZoomStep.create(200,DateTime.MINUTE());
      stepArray[2] = ZoomStep.create(100,DateTime.MINUTE());
      stepArray[3] = ZoomStep.create(50 ,DateTime.MINUTE());
      stepArray[4] = ZoomStep.create(400,DateTime.HOUR()  );
      stepArray[5] = ZoomStep.create(200,DateTime.HOUR()  );
      stepArray[6] = ZoomStep.create(100,DateTime.HOUR()  );
      stepArray[7] = ZoomStep.create(50 ,DateTime.HOUR()  );
      stepArray[8] = ZoomStep.create(400,DateTime.DAY()   );
      stepArray[9] = ZoomStep.create(200,DateTime.DAY()   );
      
      JavaScriptObject stepArrayObj = JavaScriptObjectHelper.arrayConvert(stepArray);
      
      return stepArrayObj;
   }
   
   
   /**
    * Sets the mouseWheel attribute for the given theme to 'zoom'
    * @param theme
    */
   private native static void setMouseWheelZoom(Theme theme) /*-{
      theme.mouseWheel = 'zoom'; 
   }-*/;
   
   
   /**
    * Sets the imprecise opacity attribute of the theme, controls how
    * translucent or opaque the timeline tracks are.
    * @param theme
    */
   private native static void setImpreciseOpacity(Theme theme, int opacity) /*-{
      theme.event.instant.impreciseOpacity = opacity;
   }-*/;
   
   
   /**
    * Sets the overview option for a given band to be a summary.
    * @param opts
    */
   private void setBandOverview(BandOptions opts){
      JavaScriptObjectHelper.setAttribute(opts, "overview", true);
   }
   
   
   /**
    * Sets the zoom index option for a given band 
    * @param opts
    * @param value
    */
   private void setBandZoomIndex(BandOptions opts, int value){
      JavaScriptObjectHelper.setAttribute(opts, "zoomIndex", value);
   }
   
   
   /**
    * Sets the zoom steps option for a given band 
    * @param opts
    * @param value
    */
   private void setBandZoomSteps(BandOptions opts, JavaScriptObject stepArrayObj){
      JavaScriptObjectHelper.setAttribute(opts, "zoomSteps", stepArrayObj);
   }
}
