package edu.washington.cs.rfid.map.client.data;

import java.util.ArrayList;
import java.util.HashMap;

import com.google.gwt.maps.client.overlay.Marker;
import com.google.gwt.user.client.rpc.IsSerializable;

import edu.washington.cs.rfid.map.client.ui.overlay.MarkerFactory;

public class ParticleCloud implements IsSerializable {
   
   public String                    name;
   public String                    color;
   public ArrayList<Particle>       particles;
   
   //--//
   
   // A displacement for all particle markers, 
   // facilitates simultaneous display of overlapping clouds
   private double                   perturbAmt;
   
   private MarkerFactory            markerFactory;
   private HashMap<Particle,Marker> particleToMarkerMap;
   private HashMap<Marker,Particle> markerToParticleMap;
   
   // Lists of particle markers to be removed or added following an update 
   private ArrayList<Marker>        staleMarkers;
   private ArrayList<Marker>        freshMarkers;

   
   //--//
   
   
   /**
    * Creates a new ParticleCloud.
    */
   public ParticleCloud(){
      this.name                = null;
      this.color               = null;
      this.particles           = null;
      this.perturbAmt          = 0.0;
      this.markerFactory       = null;
      this.particleToMarkerMap = null;
      this.markerToParticleMap = null;
      this.staleMarkers        = null;
      this.freshMarkers        = null;
   }
   
   
   /**
    * Creates a new ParticleCloud.
    * @param name
    * @param particles
    */
   public ParticleCloud(String name, String color, MarkerFactory markerFactory){
      this.name                = name;
      this.color               = color;
      this.particles           = new ArrayList<Particle>();
      
      // Compute a perturbation amount based on hash code for the name
      this.perturbAmt          = 0.0;
      
      long hash    = Math.abs((long)this.name.hashCode());
      long divisor = 10L;
      while(hash / divisor != 0){
         divisor *= 10L;
      }
      // we want 0.0000005
      divisor                 *= 200000L;
      this.perturbAmt          = (double)hash/(double)divisor;
      
      this.markerFactory       = markerFactory;
      this.particleToMarkerMap = new HashMap<Particle,Marker>();
      this.markerToParticleMap = new HashMap<Marker,Particle>();
      this.staleMarkers        = new ArrayList<Marker>();
      this.freshMarkers        = new ArrayList<Marker>();
   }
   
   
   /**
    * Removes all particles that aren't in the update and 
    * includes all updates that aren't in the particle list.
    * 
    * @param updates
    */
   public void updateParticles(ArrayList<Particle> updates){
      
      // Make a local copy
      ArrayList<Particle> localUpdates = new ArrayList<Particle>();
      for(Particle p : updates){
         Particle pp = new Particle(this.color, 
                                    p.floor, 
                                    new LatLon(p.loc.lat + this.perturbAmt, p.loc.lon + this.perturbAmt), 
                                    p.prob);
         localUpdates.add(pp);
      }
      
      this.staleMarkers.clear();
      this.freshMarkers.clear();
      
      for(int i = 0; i < this.particles.size(); i++){
         Particle a     = this.particles.get(i);
         boolean  found = false;
         
         for(int j = 0; j < localUpdates.size(); j++){
            Particle b = localUpdates.get(j);
            if(a.equals(b)){
               found = true;
               localUpdates.remove(j);
               j--;
               break;
            }
         }
         if(!found){
            Particle stale = this.particles.remove(i);
            Marker   m     = this.particleToMarkerMap.remove(stale);
            this.markerToParticleMap.remove(m);
            this.staleMarkers.add(m);
            i--;
         }
      }
      for(int i = 0; i < localUpdates.size(); i++){
         Particle p = localUpdates.get(i);
         this.particles.add(p);
         Marker m = this.markerFactory.createParticleMarker(p);
         
         
         
         this.freshMarkers.add(m);
         this.particleToMarkerMap.put(p, m);
         this.markerToParticleMap.put(m, p);
      }
   }
   
   
   public Particle getParticleByMarker(Marker m){
      return this.markerToParticleMap.get(m);
   }
   
   
   public Marker getMarkerByParticle(Particle p){
      return this.particleToMarkerMap.get(p);
   }
   
   
   /**
    * Gets a list of stale markers since last update.
    * @return
    */
   public ArrayList<Marker> getStaleMarkers(){
      return this.staleMarkers;
   }
   
   
   /**
    * Gets a list of fresh markers since last update.
    * @return
    */
   public ArrayList<Marker> getFreshMarkers(){
      return this.freshMarkers;
   }
}
