package edu.washington.cs.rfid.map.client.data;

import com.google.gwt.maps.client.overlay.Marker;

import edu.washington.cs.rfid.map.client.ui.overlay.MarkerFactory;

/**
 * Class representing a tracked entity.  Used to display raw RFID data on the map.
 * @author evan
 */
public class TrackedEntity {
   
   public String     name;
   public String     type;
   public TagReading reading;
   
   //--//
   
   // A displacement for the entity marker, 
   // facilitates simultaneous display of overlapping entities
   private double                   perturbAmt;
   
   private MarkerFactory            markerFactory;
   private Marker                   marker;
   private Marker                   staleMarker;
   private Marker                   freshMarker;
   
   //--//
   
   /**
    * Creates a new TrackedEntity
    */
   public TrackedEntity(){
      this.name          = null;
      this.type          = null;
      this.reading       = null;
      this.perturbAmt    = 0.0;
      this.markerFactory = null;
      this.marker        = null;
   }
   
   
   /**
    * Creates a new TrackedEntity with the specified parameters
    * @param name The entity's name
    * @param floor The floor on which the entity currently resides
    * @param loc The location (lat,lng) at which the entity currently resides
    * @param markerFactory A marker factory to produce new markers for this entity
    */
   public TrackedEntity(String name, String type, MarkerFactory markerFactory){
      this.name    = name;
      this.type    = type;
      this.reading = null;
      
      // Compute a perturbation amount based on hash code for the name
      this.perturbAmt = 0.0;
      
      long hash    = Math.abs((long)this.name.hashCode());
      long divisor = 10L;
      while(hash / divisor != 0){
         divisor *= 10L;
      }
      // we want 0.0000005
      divisor           *= 200000L;
      this.perturbAmt    = (double)hash/(double)divisor;
      
      this.markerFactory = markerFactory;
      this.staleMarker   = null;
      this.marker        = null;
      this.freshMarker   = null;         
   }
   
   
   /**
    * Updates the current location of the entity to the specified location.
    * @param loc
    */
   public void updateLocation(TagReading r){
      
      if(null == r){
         if(null != this.reading){
            this.staleMarker = this.marker;
            this.marker      = null;
            this.freshMarker = null;
         } else {
            this.staleMarker = null;
            this.marker      = null;
            this.freshMarker = null;
         }
         this.reading = null;
      } else {
         boolean updated = false;
         
         if(null == this.reading                                                     ||
            (this.reading.loc.lat != r.loc.lat || this.reading.loc.lon != r.loc.lon) || 
            (null != this.reading.floor && !this.reading.floor.equals(r.floor))){
            this.reading = new TagReading(r.floor,new LatLon(r.loc.lat + this.perturbAmt, r.loc.lon + this.perturbAmt));
            updated      = true;
         }
         
         if(updated){
            this.staleMarker = this.marker;
            this.marker      = this.markerFactory.createTrackedEntityMarker(this);
            this.freshMarker = this.marker;
         } else {
            this.staleMarker = null;
            this.freshMarker = null;
         }
      }
   }
   
   
   /**
    * Returns the stale marker if there is one.
    * @return
    */
   public Marker getStaleMarker(){
      return this.staleMarker;
   }
   
   
   /**
    * Returns a fresh marker is there is one.
    * @return
    */
   public Marker getFreshMarker(){
      return this.freshMarker;
   }
}
