package edu.washington.cs.rfid.map.client.data.cache;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.google.gwt.http.client.Request;
import com.google.gwt.http.client.RequestBuilder;
import com.google.gwt.http.client.RequestCallback;
import com.google.gwt.http.client.RequestException;
import com.google.gwt.http.client.Response;
import com.google.gwt.user.client.Window;
import com.google.gwt.xml.client.Document;
import com.google.gwt.xml.client.NamedNodeMap;
import com.google.gwt.xml.client.NodeList;
import com.google.gwt.xml.client.XMLParser;

import edu.washington.cs.rfid.data.Edge;
import edu.washington.cs.rfid.data.Node;

public class ConnectivityGraph {

//   private static final String URL_DATA  = "http://rfid.cs.washington.edu/data/";
   private static final String URL_DATA  = "../data/";
   
   private static final String XML_NODES = "nodes.xml";
   private static final String XML_EDGES = "edges.xml";
   
   //--//
   
   private static ConnectivityGraph instance;
   
   //--//
   
   private List<Node>     nodes;
   private List<Edge>     edges;
   private Map<Long,Node> nidToNodeMap;
   
   private boolean        loaded;
   private boolean        nodesLoaded;
   private boolean        edgesLoaded;
   
   //--//
   
   /**
    * Creates a new ConnectivityGraph
    * @param nodes
    * @param edges
    */
   private ConnectivityGraph(){
      this.nodes        = new ArrayList<Node>();
      this.edges        = new ArrayList<Edge>();
      this.nidToNodeMap = new HashMap<Long,Node>();
      
      this.loaded       = false;
      this.nodesLoaded  = false;
      this.edgesLoaded  = false;
   }
   
   //--//
   
   /**
    * Static representation of the connectivity graph of the Allen
    * Center.  Removes the need to go to the database.
    * @return
    */
   public static ConnectivityGraph getInstance(){
      if(null == ConnectivityGraph.instance){
         ConnectivityGraph.instance = new ConnectivityGraph();
      }
      return ConnectivityGraph.instance;
   }
   
   //--//
   
   public void loadGraph(GraphLoadedCallback callback){
      if(this.loaded){
         callback.graphLoaded(this);
      } else {
         // Load data from XML files
         this.loadNodes(callback);
         this.loadEdges(callback);
      }
   }
   
   public List<Node> getNodes(){
      return this.nodes;
   }
   
   public List<Edge> getEdges(){
      return this.edges;
   }
   
   public Node getNode(long nid){
      return this.nidToNodeMap.get(nid);
   }
   
   public boolean getLoaded(){
      return this.loaded;
   }
   
   public boolean getNodesLoaded(){
      return this.nodesLoaded;
   }
   
   public boolean getEdgesLoaded(){
      return this.edgesLoaded;
   }
   
   //--//
   
   private void loadNodes(GraphLoadedCallback loadedCallback){
      if(!this.nodesLoaded){
         String         url     = ConnectivityGraph.URL_DATA + ConnectivityGraph.XML_NODES;
         
         RequestBuilder builder = new RequestBuilder(RequestBuilder.GET, url);
         
         try {
            GraphDataRequestCallback cb = new GraphDataRequestCallback(this,loadedCallback);
            builder.sendRequest(null, cb);
         } catch (RequestException e) {
            Window.alert("Error making request for node data");
         }
      }
   }
   
   
   private void loadEdges(GraphLoadedCallback loadedCallback){
      if(!this.edgesLoaded){
         String         url     = ConnectivityGraph.URL_DATA + ConnectivityGraph.XML_EDGES;
         RequestBuilder builder = new RequestBuilder(RequestBuilder.GET, url);
         
         try {
            GraphDataRequestCallback cb = new GraphDataRequestCallback(this,loadedCallback);
            builder.sendRequest(null, cb);
         } catch (RequestException e) {
            Window.alert("Error making request for edge data");
         }
      }
   }
   
      
   /**
    * Parses Nodes from an XML list of Node elements.
    * @param nodeXml
    */
   private void parseNodes(NodeList nodeXml){
      for(int i = 0; i < nodeXml.getLength(); i++){
         com.google.gwt.xml.client.Node n          = nodeXml.item(i);
         NamedNodeMap                   attributes = n.getAttributes();

         long   nid   = 0;
         double lat   = 0;
         double lon   = 0;
         double alt   = 0;
         int    floor = 0;
         String sid   = null;
         
         for(int j = 0; j < attributes.getLength(); j++){
            com.google.gwt.xml.client.Node attrib = attributes.item(j);
            String                         name   = attrib.getNodeName();
            String                         value  = attrib.getNodeValue();

            if(name.equals("nid")){
               nid = Long.parseLong(value);
            } else if(name.equals("lat")){
               lat = Double.parseDouble(value);
            } else if(name.equals("lon")){
               lon = Double.parseDouble(value);
            } else if(name.equals("alt")){
               alt = Double.parseDouble(value);
            } else if(name.equals("floor")){
               floor = Integer.parseInt(value);
            } else if(name.equals("sid")){
               sid = value;
            }
         }
         Node node = new Node(nid,lat,lon,alt,this.getFloor(alt),sid);
         this.nodes.add(node);
         this.nidToNodeMap.put(node.nid, node);
      }
      this.nodesLoaded = true;
      
      if(this.edgesLoaded){
         this.loaded = true;
      }
   }

   
   private int getFloor(double alt){
      int floor = 0;
      
      if(alt <= 0.0)      { floor = 1; }
      else if(alt <= 5.0 ){ floor = 2; }
      else if(alt <= 9.0 ){ floor = 3; }
      else if(alt <= 13.0){ floor = 4; }
      else if(alt <= 17.0){ floor = 5; }
      else if(alt <= 21.0){ floor = 6; }
      
      return floor;
   }

   
   /**
    * Parses Edges from an XML list of Edge elements.
    * @param edgeXml
    */
   private void parseEdges(NodeList edgeXml){
      for(int i = 0; i < edgeXml.getLength(); i++){
         com.google.gwt.xml.client.Node n          = edgeXml.item(i);
         NamedNodeMap                   attributes = n.getAttributes();

         long firstNodeID  = 0;
         long secondNodeID = 0;
         
         for(int j = 0; j < attributes.getLength(); j++){
            com.google.gwt.xml.client.Node attrib = attributes.item(j);
            String                         name   = attrib.getNodeName();
            String                         value  = attrib.getNodeValue();

            if(name.equals("firstNodeID")){
               firstNodeID = Long.parseLong(value);
            } else if(name.equals("secondNodeID")){
               secondNodeID = Long.parseLong(value);
            }
         }
         Edge edge = new Edge(firstNodeID,secondNodeID);
         this.edges.add(edge);
      }
      this.edgesLoaded = true;
      
      if(this.nodesLoaded){
         this.loaded = true;
      }
   }
   
   //--//
   
   private class GraphDataRequestCallback implements RequestCallback {
      
      private ConnectivityGraph   graph;
      private GraphLoadedCallback loadedCb;
      private boolean             didLoadCb;
      
      //--//
      
      public GraphDataRequestCallback(ConnectivityGraph   graph,
                                      GraphLoadedCallback loadedCb){
         this.graph     = graph;
         this.loadedCb  = loadedCb;
         this.didLoadCb = false;
      }
      
      public void onError(Request request, Throwable exception) {
         Window.alert("Error retrieving graph data");
      }
      
      public void onResponseReceived(Request request, Response response) {
         if(!this.graph.loaded){
            String  text    = response.getText();
            boolean isNodes = text.contains("<Node");

            if((isNodes  && !this.graph.getNodesLoaded()) ||
               (!isNodes && !this.graph.getEdgesLoaded())){
               Document                       doc  = XMLParser.parse(text);
               com.google.gwt.xml.client.Node root = doc.getChildNodes().item(0);
               XMLParser.removeWhitespace(root);
               
               NodeList children = root.getChildNodes();
               if(isNodes){
                  this.graph.parseNodes(children);
               } else {
                  this.graph.parseEdges(children);
               }
            }
         }
         
         if(!this.didLoadCb && this.graph.getLoaded()){
            this.loadedCb.graphLoaded(this.graph);
            this.didLoadCb = true;
         }
      }
   }

   //--//

   
//   /**
//    * Gets a Node in XML format:
//    * <Node nid="1" lat="47.00" lon="-122.00" alt="1" floor="1" sid="name"/>
//    * @param n
//    */
//   private String getNodeXml(Node n){
//      String xml = "<Node ";
//      xml       += "nid=\""   + n.nid   + "\" ";
//      xml       += "lat=\""   + n.lat   + "\" ";
//      xml       += "lon=\""   + n.lon   + "\" ";
//      xml       += "alt=\""   + n.alt   + "\" ";
//      xml       += "floor=\"" + n.floor + "\" ";
//      xml       += "sid=\""   + n.sid   + "\" ";
//      xml       += "/>";
//      return xml;
//   }
   
   
//   /**
//    * Gets an Edge in XML format:
//    * <Edge firstNodeID="1" secondNodeID="2"/>
//    * @param e
//    */
//   private String getEdgeXml(Edge e){
//      String xml = "<Edge ";
//      xml       += "firstNodeID=\""  + e.firstNodeID  + "\" ";
//      xml       += "secondNodeID=\"" + e.secondNodeID + "\" ";
//      xml       += "/>";
//      return xml;
//   }
   
   
//   /**
//    * Transform the coordinates in the graph to match those we use on the map!
//    */
//   public static void coordinateTransform(Graph g){
//      AltitudeSpecificMapping web2VizMap = AllenCenterMappingFactory.getACMapping_webToVisualizer();
//      
//      for (Iterator<Node> iter = g.nodes.iterator(); iter.hasNext();) {
//          Node      element = (Node)iter.next();
//          LatLonAlt point   = web2VizMap.mapPoint(new LatLonAlt(element.lat,element.lon,element.alt));
//          element.lat = point.getLat();
//          element.lon = point.getLon();
//          element.alt = point.getAlt();
//      }
//   }
}
