package edu.washington.cs.rfid.map.client.ui;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.google.gwt.event.dom.client.ChangeEvent;
import com.google.gwt.event.dom.client.ChangeHandler;
import com.google.gwt.maps.client.CopyrightCollection;
import com.google.gwt.maps.client.InfoWindow;
import com.google.gwt.maps.client.InfoWindowContent;
import com.google.gwt.maps.client.MapType;
import com.google.gwt.maps.client.MapWidget;
import com.google.gwt.maps.client.TileLayer;
import com.google.gwt.maps.client.control.Control;
import com.google.gwt.maps.client.control.ControlAnchor;
import com.google.gwt.maps.client.control.ControlPosition;
import com.google.gwt.maps.client.control.LargeMapControl;
import com.google.gwt.maps.client.control.Control.CustomControl;
import com.google.gwt.maps.client.event.MapClickHandler;
import com.google.gwt.maps.client.event.MapTypeChangedHandler;
import com.google.gwt.maps.client.event.MapZoomEndHandler;
import com.google.gwt.maps.client.event.MarkerClickHandler;
import com.google.gwt.maps.client.geom.LatLng;
import com.google.gwt.maps.client.geom.MercatorProjection;
import com.google.gwt.maps.client.geom.Projection;
import com.google.gwt.maps.client.overlay.Marker;
import com.google.gwt.maps.client.overlay.Overlay;
import com.google.gwt.maps.client.overlay.Polyline;
import com.google.gwt.maps.client.overlay.PolylineOptions;
import com.google.gwt.user.client.Command;
import com.google.gwt.user.client.DeferredCommand;
import com.google.gwt.user.client.ui.Composite;
import com.google.gwt.user.client.ui.ListBox;
import com.google.gwt.user.client.ui.Widget;

import edu.washington.cs.rfid.data.Ant;
import edu.washington.cs.rfid.data.Edge;
import edu.washington.cs.rfid.data.Rdr;
import edu.washington.cs.rfid.map.client.data.Particle;
import edu.washington.cs.rfid.map.client.data.ParticleCloud;
import edu.washington.cs.rfid.map.client.data.TrackedEntity;
import edu.washington.cs.rfid.map.client.data.cache.CachedAntRdr_7_19_08;
import edu.washington.cs.rfid.map.client.data.cache.ConnectivityGraph;
import edu.washington.cs.rfid.map.client.data.cache.GraphLoadedCallback;
import edu.washington.cs.rfid.map.client.event.AntClickEvent;
import edu.washington.cs.rfid.map.client.event.RdrClickEvent;
import edu.washington.cs.rfid.map.client.handler.AntClickHandler;
import edu.washington.cs.rfid.map.client.handler.RdrClickHandler;
import edu.washington.cs.rfid.map.client.ui.overlay.MarkerFactory;
import edu.washington.cs.rfid.map.client.ui.overlay.OverlayFilter;
import edu.washington.cs.rfid.map.client.ui.overlay.OverlayWrapper;

public class MercatorMapWidget extends Composite {

   public final static String[] DEFAULT_FLOORS = new String[]{"Floor B",
                                                              "Floor 1",
                                                              "Floor 2",  
                                                              "Floor 3",
                                                              "Floor 4",
                                                              "Floor 5",
                                                              "Floor 6"
                                                             };
   public final static int MAP_TYPE_NORMAL    = 1;
   public final static int MAP_TYPE_SATELLITE = 2;
   public final static int MAP_TYPE_HYBRID    = 3;

   //--//

	private static final String   CSE_COPYRIGHT  = "UW CSE";
	private static final String   TILE_BASE_URL  = "http://rfid.cs.washington.edu/tiles/";
	private static final String   ICON_BASE_URL  = "http://rfid.cs.washington.edu/map-icons/";
	private static final LatLng   CSE_LATLNG     = LatLng.newInstance(47.65325263709352, -122.305928170681);
	
   private final static String   DEFAULT_FLOOR  = MercatorMapWidget.DEFAULT_FLOORS[3];
   private final static int      DEFAULT_ZOOM   = 20;
   
   private final static double   POLYLINE_OPACITY = 0.3;
   private final static int      POLYLINE_WEIGHT  = 3;
   private final static String   POLYLINE_COLOR   = "#0000CC";
   
   //--//

	// The map
	private MapWidget 					 	     map;
   private int                              currentMapType;

   // Controls
   private Control                          largeMapControl;
   private Control                          floorControl;
	
	// Floors
	private String	                          currFloor;
	private boolean                          floorInitialized;
	
	private Map<Integer,Map<String,MapType>> floorMaps;     // Maps type number to Map of floor to MapType
	private Map<String,Set<OverlayWrapper>>  floorOverlays;
	
	// Overlays
	private Map<Overlay,OverlayWrapper>      overlayToWrapperMap;
	private MarkerFactory 					     markerFactory;
	private OverlayFilter                    filter;
	
	// Antenna Overlays
	private Map<Integer,Ant> 			        antMap;
	private Map<Marker,Ant> 			        antMarkerMap;
	private List<AntClickHandler>            antClickHandlers;
	
	// Reader Overlays
   private Map<Integer,Rdr>                 rdrMap;
   private Map<Marker,Rdr>                  rdrMarkerMap;
   private List<RdrClickHandler>            rdrClickHandlers;
   
   // Particle Overlays
   // private Map<String,ParticleCloud>     particleCloudMap;
   // private Map<Marker,ParticleCloud>     cloudMarkerMap;
   // Map<String,List<LatLon>> updates, Map<String,Integer> floorUpdates
   // String is a key - "alice-key" or "helmet-key"
   
   // CloudName, Data: List<(LatLng, Floor, Prob)>
   // Cloud has:
   // - a person/object name
   // - set of data List<(LatLng, Floor, Prob)>
   // - set of Markers: smart transitions between timesteps for efficiency
	
	//--//
	
   /**
    * Creates a new map widget with the default floor and zoom level.
    * @param width
    * @param height
    * @param showRfid
    */
	public MercatorMapWidget(String width, String height, boolean showRfid){
	   this(MercatorMapWidget.DEFAULT_FLOOR,
	        MercatorMapWidget.DEFAULT_ZOOM ,
	        width                          ,
	        height                         ,
	        showRfid);
	}
	
	
	
	
	/**
	 * Creates a map that is initialized to the current floor, 
	 * with the specified width and height, and with the RFID 
	 * antennas and readers showing (or not). 
	 * @param currFloor
	 * @param width
	 * @param height
	 * @param showRfid
	 */
	public MercatorMapWidget(String  currFloor    , 
	                         int     currZoomLevel,
	                         String  width        , 
	                         String  height       ,
	                         boolean showRfid) {

	   // Create map
      this.currentMapType = MercatorMapWidget.MAP_TYPE_HYBRID;
      this.map            = new MapWidget(MercatorMapWidget.CSE_LATLNG, 3);
	   
      this.initOverlays();
	   this.initFloors(currFloor);
      this.initMapTypeCollection();
      this.initControls();

      // Set map parameters
      this.map.setCenter(MercatorMapWidget.CSE_LATLNG);
      this.map.setSize(width, height);
      this.map.addControl(this.floorControl);
      this.map.addControl(this.largeMapControl);
      this.map.setDoubleClickZoom(false);
      this.map.setDraggable(true);
      this.map.setScrollWheelZoomEnabled(true);
      this.map.setZoomLevel(currZoomLevel);
      this.map.setPinchToZoom(false);
      this.map.setScrollWheelZoomEnabled(false);      
      
      OverlayFilter filter = OverlayFilter.createDefaultFilter();
      this.setFilter(filter, false);
      
      if(showRfid){
         filter.setAntVisible(true);
         filter.setRdrVisible(true);
         filter.setConnectivityGraphVisible(true);
         this.loadAntennas(CachedAntRdr_7_19_08.antVector);
         this.loadReaders (CachedAntRdr_7_19_08.rdrVector);
         this.loadConnectivityGraph(ConnectivityGraph.getInstance());
      }
      
      initWidget(this.map);      
      DeferredCommand.addCommand(new SetFloorCmd());
	}

		
	//-------------------- GET / SET METHODS ----------------------

   /**
    * Gets the antenna with the specified ID.
    * @param antID
    * @return
    */
   public Ant getAntenna(int antID) {
      return this.antMap.get(antID);
   }
	
	/**
	 * Gets the center coordinates of the map
	 * @return
	 */
	public LatLng getCenter(){
	   return this.map.getCenter();
	}

   /**
    * Get the current filter
    * @return
    */
   public OverlayFilter getFilter() {
      return this.filter;
   }
	
   /**
    * Gets the current floor.
    * @return
    */
   public String getFloor() {
      return this.currFloor;
   }
   
	/**
    * Get the single info window associated with this map
    * @return
    */
   public InfoWindow getInfoWindow() {
      return this.map.getInfoWindow();
   }

   /**
    * The raw gmap object, be careful what is done it may interrupt 
    * the prepared state that the mercator map widget attempts to control
    * @return
    */
   public MapWidget getRawMap() {
      return this.map;
   }
   
   /**
    * Gets the Rdr with the specified id
    * @param rdrID
    * @return
    */
   public Rdr getReader(int rdrID) {
      return this.rdrMap.get(rdrID);
   }
   
   /**
    * Gets the zoom level of the map
    * @return
    */
   public int getZoomLevel() {
      return this.map.getZoomLevel();
   }
   
   /**
    * Sets the map's center
    * @param center
    */
   public void setCenter(LatLng center) {
      this.map.setCenter(center);
   }
   
   /**
    * Centers the map on a saved position
    */
   public void returnToCenter() {
      this.map.returnToSavedPosition();
   }
      
  /**
   * Centers the map to the given point.  If the point is visible 
   * in the current map view, the map pans in a smooth animation.
   * 
   * @param center the new center
   */
   public void panTo(LatLng point) {
      this.map.panTo(point);
   }
   
   /**
    * Set this filter to use a new filter, and refresh the icons to this 
    * new filter if requested
    * @param newFilter
    */
   public void setFilter(OverlayFilter newFilter, boolean forceRefresh) {
      if( newFilter != null ) {
         this.filter = newFilter;
         this.filter.setMap(this);
         if(forceRefresh) {
            refreshOverlays();
         }
      }
   }
   
   /**
    * Change this floor number
    * @param floorNum
    */
   public void setFloor(String floor) {
      if(this.checkValidFloor(floor)){
         if(!this.floorInitialized || !this.currFloor.equals(floor)){
            this.currFloor = floor;
            this.map.setCurrentMapType(this.floorMaps.get(this.currentMapType).get(floor));
            this.refreshOverlays();
         }         
      } else {
         throw new IllegalArgumentException("Invalid floor: " + floor);
      }
   }
   
   /**
    * Set the maptype
    * @param mapType
    */
   public void setMapType(int mapType) {
      if(mapType != MAP_TYPE_NORMAL   ||
         mapType != MAP_TYPE_SATELLITE || 
         mapType != MAP_TYPE_HYBRID) {
         throw new IllegalArgumentException("Invalid MapType: " + mapType);
      }
      
      if(mapType != this.currentMapType) {
         this.currentMapType = mapType;
         this.map.setCurrentMapType(this.floorMaps.get(mapType).get(this.currFloor));
      }
   }
   
   /**
    * Sets the zoom level
    * @param zoomLevel
    */
   public void setZoomLevel(int zoomLevel) {
      this.map.setZoomLevel(zoomLevel);
   }
   
   private boolean checkValidFloor(String floor){
      boolean validFloor = false;
      for(String f : MercatorMapWidget.DEFAULT_FLOORS){
         if(f.equals(floor)){
            validFloor = true;
            break;
         }
      }
      return validFloor;
   }
   
   
   //-------------------- ADD / REMOVE METHODS ----------------------
   
   public void addAntClickHandler(AntClickHandler handler) {
      this.antClickHandlers.add(handler);
   }
   
   public void addRdrClickHandler(RdrClickHandler handler) {
      this.rdrClickHandlers.add(handler);
   }

   public void addClickHandler(MapClickHandler handler) {
      map.addMapClickHandler(handler);
   }

   public void addCustomControl(CustomControl control) {
      map.addControl(control);
   }

   /**
    * Add this overlay to this floor, if its the current floor, draw it
    * if this overlay already exists on the map, we do not add it.
    * 
    * @param overlay
    * @param floorNum
    */
   public void addOverlay(Overlay overlay, int overlayTypeID, String floor) {
      if(!this.overlayToWrapperMap.containsKey(overlay)) {
         OverlayWrapper wrapper = new OverlayWrapper(overlay, overlayTypeID, floor);
         this.overlayToWrapperMap.put(overlay, wrapper);
         this.floorOverlays.get(floor).add(wrapper);
         
         if(this.currFloor.equals(floor)) {
            if(this.filter.overlayAllowed(wrapper)) {
               this.map.addOverlay(overlay);
            }
         }
      }
   }
   
   public void removeAntClickHandler(AntClickHandler handler) {
      this.antClickHandlers.remove(handler);
   }
   
   public void removeRdrClickHandler(RdrClickHandler handler) {
      this.rdrClickHandlers.remove(handler);
   }

   public void removeClickHandler(MapClickHandler handler) {
      this.map.removeMapClickHandler(handler);
   }
   
   public void removeCustomControl(CustomControl control) {
      this.map.removeControl(control);
   }
   
   /**
    * Remove this overlay from this floor, if this floor is the current floor,
    * undraw it.
    * @param overlay
    * @param floorNum
    */
   public void removeOverlay(Overlay overlay) {
      OverlayWrapper wrapper = this.overlayToWrapperMap.remove(overlay);
      if(wrapper != null) {
         this.floorOverlays.get(wrapper.floor).remove(wrapper);
      }
      this.map.removeOverlay(overlay);
   }
   
   
   public void removeLargeMapControl(){
      this.getRawMap().removeControl(this.largeMapControl);
   }
   
   
   public void removeFloorControl(){
      this.getRawMap().removeControl(this.floorControl);
   }
   
   
   //------------- OPEN / CLOSE / CLEAR / REFRESH METHODS ---------------
   
   public void openInfoWindow(LatLng latlng, Widget widget){
      InfoWindow        w  = this.map.getInfoWindow();
      InfoWindowContent wc = new InfoWindowContent(widget);
      w.open(latlng, wc);
   }
   
   public void closeInfoWindow(){
      this.map.closeInfoWindow();
   }
   
   /**
    * Clear all overlays
    * note: this does not clear antennas or readers. To remove readers or antennas
    * access the getFilter()
    * 
    * {@link #getFilter()}
    */
   public void clearOverlays() {
      HashSet<Overlay> overlays = new HashSet<Overlay>(this.overlayToWrapperMap.keySet());
      for(Overlay o : overlays) {
         OverlayWrapper wrap = this.overlayToWrapperMap.get(o);
         if(wrap.typeID != OverlayWrapper.TYPE_ANTENNA && wrap.typeID != OverlayWrapper.TYPE_READER) {
            this.removeOverlay(o);
         }
      }
   }
   
   public void clearOverlayType(int typeId){
      HashSet<Overlay> overlays = new HashSet<Overlay>(this.overlayToWrapperMap.keySet());
      for(Overlay o : overlays) {
         OverlayWrapper wrap = this.overlayToWrapperMap.get(o);
         if(wrap.typeID == typeId) {
            this.removeOverlay(o);
         }
      }
   }
   
   /**
    * Repaint all the overlays on the currently selected floor
    */
   public void refreshOverlays() {
      this.map.clearOverlays(); 
      
      for(OverlayWrapper wrapper : this.floorOverlays.get(this.currFloor)) {
         // check to see if this overlay is allowed
         if(this.filter.overlayAllowed(wrapper)) {
            this.map.addOverlay(wrapper.overlay);
         }
      }
   }
   
   
   //------------- OTHER METHODS ---------------
   
   
   public void processParticleCloud(ParticleCloud cloud){
      ArrayList<Marker> stale = cloud.getStaleMarkers();
      for(Marker m : stale){
         this.removeOverlay(m);
      }
      
      ArrayList<Marker> fresh = cloud.getFreshMarkers();
      for(Marker m : fresh){
         Particle p = cloud.getParticleByMarker(m);
         this.addOverlay(m, OverlayWrapper.TYPE_PARTICLE, p.floor); // TODO: Change floors if necessary
      }
   }
   
   
   public void processTrackedEntity(TrackedEntity e){
      Marker stale = e.getStaleMarker();
      if(null != stale){
         this.removeOverlay(stale);
      }
      
      Marker fresh = e.getFreshMarker();
      if(null != fresh){
         this.addOverlay(fresh, OverlayWrapper.TYPE_ENTITY, e.reading.floor); // TODO: Change floors if necessary
      }
   }
   
   
   /**
    * Because of an issue with adding a map invisible, it needs to 
    * reinitialize its bounds to display properly
    */
   public void checkResize() {
      this.map.checkResize();
   }


   //--//
	
   
   //------------------------------- INIT METHODS ------------------------------
   
   private void initOverlays(){
      this.antMap              = new HashMap<Integer, Ant>();
      this.rdrMap              = new HashMap<Integer, Rdr>();
      this.antMarkerMap        = new HashMap<Marker, Ant>();
      this.rdrMarkerMap        = new HashMap<Marker, Rdr>();
      this.antClickHandlers    = new ArrayList<AntClickHandler>();
      this.rdrClickHandlers    = new ArrayList<RdrClickHandler>();
      this.overlayToWrapperMap = new HashMap<Overlay, OverlayWrapper>();
      this.markerFactory       = new MarkerFactory(MercatorMapWidget.ICON_BASE_URL);
   }
   
   private void initFloors(String currFloor){
      this.currFloor        = currFloor;
      this.floorInitialized = false;
   }
   
	
	/**
	 * Initializes controls.
	 */
	private void initControls(){
      this.largeMapControl = new LargeMapControl();
      this.floorControl    = new FloorControl(this.currFloor);
	}
	
	
	/**
	 * Initializes map type collection.
	 */
	private void initMapTypeCollection(){
	   CopyrightCollection collection     = new CopyrightCollection(MercatorMapWidget.CSE_COPYRIGHT);       
      Projection          projection     = new MercatorProjection(23);
      TileLayer           normalLayer    = MapType.getNormalMap().getTileLayers()[0];
      TileLayer           satelliteLayer = MapType.getSatelliteMap().getTileLayers()[0];
      TileLayer           hybridLayer0   = MapType.getHybridMap().getTileLayers()[0];
      TileLayer           hybridLayer1   = MapType.getHybridMap().getTileLayers()[1];
      TileLayer           blankLayer     = new BlankTileLayer(collection, 15, 23);
      
      this.floorMaps = new HashMap<Integer,Map<String,MapType>>();
      this.floorMaps.put(MercatorMapWidget.MAP_TYPE_NORMAL  , new HashMap<String,MapType>());   
      this.floorMaps.put(MercatorMapWidget.MAP_TYPE_SATELLITE, new HashMap<String,MapType>());
      this.floorMaps.put(MercatorMapWidget.MAP_TYPE_HYBRID  , new HashMap<String,MapType>());
      
      this.floorOverlays = new HashMap<String,Set<OverlayWrapper>>();
      
      for(String floor : MercatorMapWidget.DEFAULT_FLOORS){
         TileLayer overlayLayer = new MercatorTileLayer(collection, 12, 23, TILE_BASE_URL, floor);
         
         TileLayer[] tileLayer = new TileLayer[] {blankLayer, normalLayer, overlayLayer};
         this.floorMaps.get(MAP_TYPE_NORMAL).put(floor,new MapType(tileLayer, projection, floor));
         
         tileLayer = new TileLayer[]{blankLayer, satelliteLayer, overlayLayer};
         this.floorMaps.get(MAP_TYPE_SATELLITE).put(floor, new MapType(tileLayer,projection,floor));
         
         tileLayer = new TileLayer[] {blankLayer, hybridLayer0, hybridLayer1, overlayLayer};
         this.floorMaps.get(MAP_TYPE_HYBRID).put(floor, new MapType(tileLayer,projection,floor));
         
         this.map.addMapType(this.floorMaps.get(MAP_TYPE_NORMAL   ).get(floor));
         this.map.addMapType(this.floorMaps.get(MAP_TYPE_SATELLITE).get(floor));
         this.map.addMapType(this.floorMaps.get(MAP_TYPE_HYBRID   ).get(floor));
         
         this.floorOverlays.put(floor,new HashSet<OverlayWrapper>());
      }
	}
	
	
	//---------------- INFRASTRUCTURE MARKER METHODS ----------------------
	

	// ---- ANTENNAS ----
	
	/**
	 * Loads antennas
	 * @param antennas
	 */
	private void loadAntennas(List<Ant> antennas) {
      for(Ant ant : antennas) {
         this.antMap.put(ant.aid, ant);
         Marker m = this.markerFactory.createAntennaMarker(ant);
         this.antMarkerMap.put(m, ant);
         m.addMarkerClickHandler(this.antClickHandler);
         this.addOverlay(m, OverlayWrapper.TYPE_ANTENNA, MercatorMapWidget.DEFAULT_FLOORS[ant.floor]);
      }
   }

   /**
    * Forward these events to our listeners
    */
   private final MarkerClickHandler antClickHandler = new MarkerClickHandler() {
      public void onClick(MarkerClickEvent event) {
         fireAntClickHandlers(event.getSender());        
      }     
   };

   private void fireAntClickHandlers(Marker m) {
      Ant ant = antMarkerMap.get(m);
      if(ant != null) {
         for(AntClickHandler handler : this.antClickHandlers){
            handler.onClick(new AntClickEvent(ant,m));
         }
      }
   }

   
   // ---- READERS ----
   
	/**
	 * Loads readers
	 * @param readers
	 */
   private void loadReaders(List<Rdr> readers) {
      for(Rdr rdr : readers) {
         this.rdrMap.put(rdr.rid, rdr);
         Marker m = this.markerFactory.createReaderMarker(rdr);
         this.rdrMarkerMap.put(m, rdr);
         m.addMarkerClickHandler(this.rdrClickHandler);
         this.addOverlay(m, OverlayWrapper.TYPE_READER, MercatorMapWidget.DEFAULT_FLOORS[rdr.floor]);
      }
   }

	private final MarkerClickHandler rdrClickHandler = new MarkerClickHandler() {
		public void onClick(MarkerClickEvent event) {
			fireRdrClickHandlers(event.getSender());			
		}		
	};
	
	private void fireRdrClickHandlers(Marker m) {
		Rdr rdr = rdrMarkerMap.get(m);
		if(rdr != null) {
			for(RdrClickHandler handler : this.rdrClickHandlers){
			   handler.onClick(new RdrClickEvent(rdr,m));
			}
		}
	}
	
	
   // ---- CONNECTIVITY GRAPH ----
	
	private void loadConnectivityGraph(ConnectivityGraph graph){
	   
	   graph.loadGraph(new GraphLoadedCallback(){
	      public void graphLoaded(ConnectivityGraph graph){
	         for(Edge e : graph.getEdges()){
	            int    floor = graph.getNode(e.firstNodeID).floor;
	            double lat1  = graph.getNode(e.firstNodeID).lat;
	            double lon1  = graph.getNode(e.firstNodeID).lon;
	            double lat2  = graph.getNode(e.secondNodeID).lat;
	            double lon2  = graph.getNode(e.secondNodeID).lon;
	            
	            PolylineOptions options  = PolylineOptions.newInstance(false,false);
	            LatLng[]        points   = new LatLng[]{LatLng.newInstance(lat1,lon1), 
	                                                    LatLng.newInstance(lat2,lon2)};
	            Polyline        edgeLine = new Polyline(points                            , 
	                                                    MercatorMapWidget.POLYLINE_COLOR  ,
	                                                    MercatorMapWidget.POLYLINE_WEIGHT ,
	                                                    MercatorMapWidget.POLYLINE_OPACITY, 
	                                                    options);
	            // TODO: ADD OVERLAYS IN BATCH FOR FAST REFRESH!
	            addOverlay(edgeLine,
	                       OverlayWrapper.TYPE_CONNECTIVITY_EDGE, 
	                       MercatorMapWidget.DEFAULT_FLOORS[floor]);
	         }	         
	      }
	   });
	}
	
	//--//
	
	/**
	 * When we zoom, lets figure out what needs to change (allowing icons to be drawn, etc)
	 */
	private final MapZoomEndHandler zoomHandler = new MapZoomEndHandler() {
		public void onZoomEnd(MapZoomEndEvent event) {
			//if(event.getNewZoomLevel() > ICON_ZOOM_TOLERANCE)
		}
	};
	
	//--//

	private class SetFloorCmd implements Command {
      public void execute(){
        setFloor(currFloor);
        setZoomLevel(20);
      }
   }
	
	//--//
	
	/**
	 * Floor control widget to be embedded on the map under the zoom control
	 * @author gbc3
	 *
	 */
	private class FloorControl extends CustomControl {				
		
	   private ListBox lb;
		private String  startFloor;
		
		public FloorControl(String startFloor) {
			super(new ControlPosition(ControlAnchor.TOP_LEFT, 77, 7));
			this.startFloor = startFloor;
		}
		
		@Override
		protected Widget initialize(MapWidget map) {
			this.lb = new ListBox();
			for(String floor : MercatorMapWidget.DEFAULT_FLOORS){
			   this.lb.addItem(floor, floor);
			}
			this.setSelectedIndex(this.startFloor);
			this.lb.addChangeHandler(new ChangeHandler(){
				public void onChange(ChangeEvent e){
					setFloor(lb.getValue(lb.getSelectedIndex()));
				}
			});
			map.addMapTypeChangedHandler(new MapTypeChangedHandler() {
				public void onTypeChanged(MapTypeChangedEvent event) {
				   setSelectedIndex(currFloor);
				}
			});
			return this.lb;
		}
		
		private void setSelectedIndex(String floor){
		   for(int i = 0; i < this.lb.getItemCount(); i++){
            if(this.lb.getValue(i).equals(currFloor)){
               this.lb.setSelectedIndex(i);
            }
         }
		}

		@Override
		public boolean isSelectable() {
			return true;
		}
	}
}
