package edu.washington.cs.rfid.map.client.ui.overlay;

import java.util.HashMap;
import java.util.Map;

import com.google.gwt.maps.client.geom.LatLng;
import com.google.gwt.maps.client.geom.Point;
import com.google.gwt.maps.client.geom.Size;
import com.google.gwt.maps.client.overlay.Icon;
import com.google.gwt.maps.client.overlay.Marker;
import com.google.gwt.maps.client.overlay.MarkerOptions;

import edu.washington.cs.rfid.data.Ant;
import edu.washington.cs.rfid.data.Rdr;
import edu.washington.cs.rfid.map.client.data.Particle;
import edu.washington.cs.rfid.map.client.data.TrackedEntity;
import edu.washington.cs.rfid.map.client.data.cache.DefaultMapImages;

public class MarkerFactory {
 
	private static final String ICON_URL_ANTENNA = "ant_red.gif";
	private static final String ICON_URL_READER  = "rdr_orange.gif";
	
	//--//
	
	private String                      iconBaseUrl;
	
	private MarkerOptions               readerOptions;
	private MarkerOptions               antennaOptions;
	private Map<String,MarkerOptions[]> colorToMarkerOptions;
	private Map<String,MarkerOptions>   entityTypeToMarkerOptions;
	
	private Marker                      blankMarker;
   
	//--//
	
	public MarkerFactory(String iconBaseURL) {
	   
	   this.initDefaultIcons(iconBaseURL);
	   
	   this.blankMarker = null;
	}

	
	public Marker getBlankMarker(){
	   return this.blankMarker;
	}
	
	
	/**
	 * Generate a marker at this point with these marker options
	 * 
	 * @param p
	 * @param o
	 * @return
	 */
	public Marker createMarker(LatLng point, MarkerOptions options) {
		return new Marker(point, options);
	}
	
	
	public Marker createAntennaMarker(Ant ant) {		
		return createMarker(LatLng.newInstance(ant.lat, ant.lon), this.antennaOptions);
	}
	
	
	public Marker createReaderMarker(Rdr rdr) {
		return createMarker(LatLng.newInstance(rdr.lat, rdr.lon), this.readerOptions);
	}
	
	
	/**
	 * Creates and returns a marker for the specified TrackedEntity.
	 */
	public Marker createTrackedEntityMarker(TrackedEntity e){
	   return this.createMarker(LatLng.newInstance(e.reading.loc.lat,e.reading.loc.lon), 
	                            this.entityTypeToMarkerOptions.get(e.type));
	}
	
	
	/**
	 * Creates and return a marker for the specified Particle.
	 * @param p
	 * @return
	 */
	public Marker createParticleMarker(Particle p){
	   // 10, 30, 50, 70, 90, 100 -- transparency setting for probability level is arbitrary (based on inspection of data)
	   int probIndex = 0;
	   if(p.prob < 0.1){
	      probIndex = 0;
	   } else if(p.prob < 0.2){
	      probIndex = 1;
	   } else if(p.prob < 0.3){
         probIndex = 2;
      } else if(p.prob < 0.4){
         probIndex = 3;
      } else if(p.prob < 0.5){
         probIndex = 4;
      } else if(p.prob >= 0.5){
         probIndex = 5;
      }
	   return this.createMarker(LatLng.newInstance(p.loc.lat,p.loc.lon),
	                            this.colorToMarkerOptions.get(p.color)[probIndex]);
	}
	
	
// public Marker createDataMarker(String key, LatLon point){
// Marker        m = null;
// MarkerOptions o = this.keyToOptionsMap.get(key + "-data");
// if(null != o){
//    Double perturbD = this.keyPerturbations.get(key);
//    double perturb  = 0;
//    if(null == perturbD){
//       // Add small perturbation based on key hashcode
//       long hash   = (long)key.hashCode();
//       long divisor = 10;
//       while(hash / divisor != 0){
//          divisor *= 10;
//       }
//       // we want: 0.000001;
//       divisor *= 100000;
//       
//       perturb = (double)hash/(double)divisor;
//       perturbD = new Double(perturb);
//       this.keyPerturbations.put(key, perturbD);
//    }
//    LatLng latlng = LatLng.newInstance(point.lat + perturb,point.lon + perturb); 
//    m             = this.createMarker(latlng, o);         
// }
// return m;
//}
	
	
// public Marker createModelMarker(String key, LatLon point){
// Marker        m = null;
// MarkerOptions o = this.keyToOptionsMap.get(key + "-model");
// if(null != o){
//    Double perturbD = this.keyPerturbations.get(key);
//    double perturb  = 0;
//    if(null == perturbD){
//       // Add small perturbation based on key hashcode
//       long hash   = (long)key.hashCode();
//       long divisor = 10;
//       while(hash / divisor != 0){
//          divisor *= 10;
//       }
//       // we want: 0.000001;
//       divisor *= 100000;
//       
//       perturb = (double)hash/(double)divisor;
//       perturbD = new Double(perturb);
//       this.keyPerturbations.put(key, perturbD);
//    }
//    
//    LatLng latlng = LatLng.newInstance(point.lat,point.lon);
//    m             = this.createMarker(latlng, o);
// }
// return m;
//}
	
	
	//--//
	
	
	private void initDefaultIcons(String iconBaseUrl){
	   this.iconBaseUrl = iconBaseUrl;

	   // Init antenna and reader icons
	   Icon icon = Icon.newInstance();
	   icon.setIconSize(Size.newInstance(13, 13));
	   icon.setShadowSize(Size.newInstance(22, 20));
	   icon.setIconAnchor(Point.newInstance(0, 0));
	   icon.setInfoWindowAnchor(Point.newInstance(5, 1));

	   Icon antennaIcon = Icon.newInstance(icon);
	   antennaIcon.setImageURL(this.iconBaseUrl + "/" + ICON_URL_ANTENNA);
	   this.antennaOptions = MarkerOptions.newInstance(antennaIcon);

	   Icon readerIcon = Icon.newInstance(icon);
	   readerIcon.setImageURL(this.iconBaseUrl + "/" + ICON_URL_READER);
	   this.readerOptions = MarkerOptions.newInstance(readerIcon);
	   
      this.colorToMarkerOptions      = this.getColorToMarkerOptionsMap(this.iconBaseUrl);
      this.entityTypeToMarkerOptions = this.getEntityTypeToMarkerOptions(this.iconBaseUrl);
	}
	
	
	/**
    * Creates and returns a map of particle color names to marker options.
    * @param iconBaseUrl
    * @return
    */
   private HashMap<String,MarkerOptions[]> getColorToMarkerOptionsMap(String iconBaseUrl){
      HashMap<String,MarkerOptions[]> colorToMarkerOptions = new HashMap<String,MarkerOptions[]>();
      
      colorToMarkerOptions.put(DefaultMapImages.BLUE_PARTICLE      , new MarkerOptions[6]);
      colorToMarkerOptions.put(DefaultMapImages.BLUE_GREEN_PARTICLE, new MarkerOptions[6]);
      colorToMarkerOptions.put(DefaultMapImages.ORANGE_PARTICLE    , new MarkerOptions[6]);
      colorToMarkerOptions.put(DefaultMapImages.PINK_PARTICLE      , new MarkerOptions[6]);
      colorToMarkerOptions.put(DefaultMapImages.PURPLE_PARTICLE    , new MarkerOptions[6]);
      colorToMarkerOptions.put(DefaultMapImages.RED_PARTICLE       , new MarkerOptions[6]);
      colorToMarkerOptions.put(DefaultMapImages.YELLOW_PARTICLE    , new MarkerOptions[6]);
      
      Icon picon = Icon.newInstance();
      picon.setIconSize        (Size.newInstance (8 , 8 ));
      picon.setShadowSize      (Size.newInstance (17, 15));
      picon.setIconAnchor      (Point.newInstance(0 , 0 ));
      picon.setInfoWindowAnchor(Point.newInstance(4 , 1 ));

      for(int i = 0; i < 6; i++){
         Icon particleIcon = Icon.newInstance(picon);
         particleIcon.setImageURL(iconBaseUrl + "/" + DefaultMapImages.ICONS_URL_BLUE_PARTICLES[i]);
         colorToMarkerOptions.get(DefaultMapImages.BLUE_PARTICLE)[i] = MarkerOptions.newInstance(particleIcon);
         
         particleIcon = Icon.newInstance(picon);
         particleIcon.setImageURL(iconBaseUrl + "/" + DefaultMapImages.ICONS_URL_BLUE_GREEN_PARTICLES[i]);
         colorToMarkerOptions.get(DefaultMapImages.BLUE_GREEN_PARTICLE)[i] = MarkerOptions.newInstance(particleIcon);
         
         particleIcon = Icon.newInstance(picon);
         particleIcon.setImageURL(iconBaseUrl + "/" + DefaultMapImages.ICONS_URL_ORANGE_PARTICLES[i]);
         colorToMarkerOptions.get(DefaultMapImages.ORANGE_PARTICLE)[i] = MarkerOptions.newInstance(particleIcon);
         
         particleIcon = Icon.newInstance(picon);
         particleIcon.setImageURL(iconBaseUrl + "/" + DefaultMapImages.ICONS_URL_PINK_PARTICLES[i]);
         colorToMarkerOptions.get(DefaultMapImages.PINK_PARTICLE)[i] = MarkerOptions.newInstance(particleIcon);
         
         particleIcon = Icon.newInstance(picon);
         particleIcon.setImageURL(iconBaseUrl + "/" + DefaultMapImages.ICONS_URL_PURPLE_PARTICLES[i]);
         colorToMarkerOptions.get(DefaultMapImages.PURPLE_PARTICLE)[i] = MarkerOptions.newInstance(particleIcon);
         
         particleIcon = Icon.newInstance(picon);
         particleIcon.setImageURL(iconBaseUrl + "/" + DefaultMapImages.ICONS_URL_RED_PARTICLES[i]);
         colorToMarkerOptions.get(DefaultMapImages.RED_PARTICLE)[i] = MarkerOptions.newInstance(particleIcon);
         
         particleIcon = Icon.newInstance(picon);
         particleIcon.setImageURL(iconBaseUrl + "/" + DefaultMapImages.ICONS_URL_YELLOW_PARTICLES[i]);
         colorToMarkerOptions.get(DefaultMapImages.YELLOW_PARTICLE)[i] = MarkerOptions.newInstance(particleIcon);
      }
      return colorToMarkerOptions;
   }
	
	
	/**
    * Creates and returns a map of entity names to marker options.
    * @param iconBaseUrl
    * @return
    */
   private HashMap<String,MarkerOptions> getEntityTypeToMarkerOptions(String iconBaseUrl){
      HashMap<String,MarkerOptions> entityTypeToMarkerOptions = new HashMap<String,MarkerOptions>();
      
      Icon entityIcon = this.getIconByType(DefaultMapImages.BLUE_PERSON);
      entityIcon.setImageURL(iconBaseUrl + "/" + DefaultMapImages.ICON_BLUE_PERSON);
      entityTypeToMarkerOptions.put(DefaultMapImages.BLUE_PERSON, MarkerOptions.newInstance(entityIcon));
      
      entityIcon = this.getIconByType(DefaultMapImages.BLUE_GREEN_PERSON);
      entityIcon.setImageURL(iconBaseUrl + "/" + DefaultMapImages.ICON_BLUE_GREEN_PERSON);
      entityTypeToMarkerOptions.put(DefaultMapImages.BLUE_GREEN_PERSON, MarkerOptions.newInstance(entityIcon));
      
      entityIcon = this.getIconByType(DefaultMapImages.ORANGE_PERSON);
      entityIcon.setImageURL(iconBaseUrl + "/" + DefaultMapImages.ICON_ORANGE_PERSON);
      entityTypeToMarkerOptions.put(DefaultMapImages.ORANGE_PERSON, MarkerOptions.newInstance(entityIcon));
      
      entityIcon = this.getIconByType(DefaultMapImages.PINK_PERSON);
      entityIcon.setImageURL(iconBaseUrl + "/" + DefaultMapImages.ICON_PINK_PERSON);
      entityTypeToMarkerOptions.put(DefaultMapImages.PINK_PERSON, MarkerOptions.newInstance(entityIcon));
      
      entityIcon = this.getIconByType(DefaultMapImages.PURPLE_PERSON);
      entityIcon.setImageURL(iconBaseUrl + "/" + DefaultMapImages.ICON_PURPLE_PERSON);
      entityTypeToMarkerOptions.put(DefaultMapImages.PURPLE_PERSON, MarkerOptions.newInstance(entityIcon));
      
      entityIcon = this.getIconByType(DefaultMapImages.RED_PERSON);
      entityIcon.setImageURL(iconBaseUrl + "/" + DefaultMapImages.ICON_RED_PERSON);
      entityTypeToMarkerOptions.put(DefaultMapImages.RED_PERSON, MarkerOptions.newInstance(entityIcon));
      
      entityIcon = this.getIconByType(DefaultMapImages.YELLOW_PERSON);
      entityIcon.setImageURL(iconBaseUrl + "/" + DefaultMapImages.ICON_YELLOW_PERSON);
      entityTypeToMarkerOptions.put(DefaultMapImages.YELLOW_PERSON, MarkerOptions.newInstance(entityIcon));
      
      entityIcon = this.getIconByType(DefaultMapImages.BACKPACK);
      entityIcon.setImageURL(iconBaseUrl + "/" + DefaultMapImages.ICON_BACKPACK);
      entityTypeToMarkerOptions.put(DefaultMapImages.BACKPACK, MarkerOptions.newInstance(entityIcon));
      
      entityIcon = this.getIconByType(DefaultMapImages.BALL);
      entityIcon.setImageURL(iconBaseUrl + "/" + DefaultMapImages.ICON_BALL);
      entityTypeToMarkerOptions.put(DefaultMapImages.BALL, MarkerOptions.newInstance(entityIcon));
      
      entityIcon = this.getIconByType(DefaultMapImages.BIKE_HELMET);
      entityIcon.setImageURL(iconBaseUrl + "/" + DefaultMapImages.ICON_BIKE_HELMET);
      entityTypeToMarkerOptions.put(DefaultMapImages.BIKE_HELMET, MarkerOptions.newInstance(entityIcon));
      
      entityIcon = this.getIconByType(DefaultMapImages.BINDER);
      entityIcon.setImageURL(iconBaseUrl + "/" + DefaultMapImages.ICON_BINDER);
      entityTypeToMarkerOptions.put(DefaultMapImages.BINDER, MarkerOptions.newInstance(entityIcon));
      
      entityIcon = this.getIconByType(DefaultMapImages.BOOK);
      entityIcon.setImageURL(iconBaseUrl + "/" + DefaultMapImages.ICON_BOOK);
      entityTypeToMarkerOptions.put(DefaultMapImages.BOOK, MarkerOptions.newInstance(entityIcon));
      
      entityIcon = this.getIconByType(DefaultMapImages.CAMCORDER);
      entityIcon.setImageURL(iconBaseUrl + "/" + DefaultMapImages.ICON_CAMCORDER);
      entityTypeToMarkerOptions.put(DefaultMapImages.CAMCORDER, MarkerOptions.newInstance(entityIcon));
      
      entityIcon = this.getIconByType(DefaultMapImages.CAMERA);
      entityIcon.setImageURL(iconBaseUrl + "/" + DefaultMapImages.ICON_CAMERA);
      entityTypeToMarkerOptions.put(DefaultMapImages.CAMERA, MarkerOptions.newInstance(entityIcon));
      
      entityIcon = this.getIconByType(DefaultMapImages.MOP);
      entityIcon.setImageURL(iconBaseUrl + "/" + DefaultMapImages.ICON_MOP);
      entityTypeToMarkerOptions.put(DefaultMapImages.MOP, MarkerOptions.newInstance(entityIcon));
      
      entityIcon = this.getIconByType(DefaultMapImages.MOUSE);
      entityIcon.setImageURL(iconBaseUrl + "/" + DefaultMapImages.ICON_MOUSE);
      entityTypeToMarkerOptions.put(DefaultMapImages.MOUSE, MarkerOptions.newInstance(entityIcon));
      
      entityIcon = this.getIconByType(DefaultMapImages.MUG);
      entityIcon.setImageURL(iconBaseUrl + "/" + DefaultMapImages.ICON_MUG);
      entityTypeToMarkerOptions.put(DefaultMapImages.MUG, MarkerOptions.newInstance(entityIcon));
      
      entityIcon = this.getIconByType(DefaultMapImages.NOTEBOOK);
      entityIcon.setImageURL(iconBaseUrl + "/" + DefaultMapImages.ICON_NOTEBOOK);
      entityTypeToMarkerOptions.put(DefaultMapImages.NOTEBOOK, MarkerOptions.newInstance(entityIcon));
      
      return entityTypeToMarkerOptions;
   }
   

   
   
   /**
    * Gets the icon for the specified entity.
    * @param entityType
    * @return
    */
   private Icon getIconByType(String entityType){
      Icon icon = Icon.newInstance();
      
      // Mop and persons are the only special cases
      if(entityType.contains("person")){
         icon.setIconSize        (Size.newInstance (20,29));// 11,16
         icon.setShadowSize      (Size.newInstance (20,29));
         icon.setIconAnchor      (Point.newInstance(0 , 0));
         icon.setInfoWindowAnchor(Point.newInstance(10, 1));
      } else if(entityType.equals(DefaultMapImages.MOP)) {
         icon.setIconSize        (Size.newInstance (14,20));
         icon.setShadowSize      (Size.newInstance (14,20));
         icon.setIconAnchor      (Point.newInstance(0 , 0));
         icon.setInfoWindowAnchor(Point.newInstance(7 , 1));
      } else {
         icon.setIconSize        (Size.newInstance (20,20));
         icon.setShadowSize      (Size.newInstance (20,20));
         icon.setIconAnchor      (Point.newInstance(0 , 0));
         icon.setInfoWindowAnchor(Point.newInstance(10, 1));
      }
      return icon;
   }
}
